/* File:      tp_index.P
** Author(s): Saumya Debray, Terrance Swift, Kostis Sagonas
** Contact:   xsb-contact@cs.sunysb.edu
** 
** Copyright (C) The Research Foundation of SUNY, 1986, 1993-1998
** Copyright (C) ECRC, Germany, 1990
** 
** XSB is free software; you can redistribute it and/or modify it under the
** terms of the GNU Library General Public License as published by the Free
** Software Foundation; either version 2 of the License, or (at your option)
** any later version.
** 
** XSB is distributed in the hope that it will be useful, but WITHOUT ANY
** WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
** FOR A PARTICULAR PURPOSE.  See the GNU Library General Public License for
** more details.
** 
** You should have received a copy of the GNU Library General Public License
** along with XSB; if not, write to the Free Software Foundation,
** Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
**
** $Id: tp_index.P,v 1.12 2006/03/20 20:41:00 kostis Exp $
** 
*/

/*======================================================================*/
/*  pred_indexarg(+Pred, -IndxArg)    	  	  	 		*/
/*  	Returns the argument on which the predicate will be indexed on.	*/
/*======================================================================*/

:- mode pred_indexarg(+,-).

pred_indexarg(Pred, IndxArg) :-
	Pred = pred(Sym,Clauses,Prag),
	sym_name(Sym, _, Arity),
	( Arity =:= 0 -> IndxArg = 0
	; %% index argument is already specified
	    member2(index(IndxArg), Prag), nonvar(IndxArg) -> true 
	; %% no real reason to index these
	    Clauses = [_] -> IndxArg = 0
	; %% index argument is now computed
	    extract_heads(Clauses, Heads), index_euristic(Heads, IndxArg),
	    indexing_message(Sym, IndxArg)
	).

/*======================================================================*/

:- mode extract_heads(+,-).

extract_heads([], []).
extract_heads([clause(HeadArgs,_,_)|Cls], [HeadArgs|HArgs]) :-
	extract_heads(Cls, HArgs).

/*======================================================================*/

:- import message/1 from auxlry.
:- import sym_has_prop/2 from symtab.

indexing_message(Sym,_IndxArg) :- sym_has_prop(Sym, internal), !.
indexing_message(Sym, IndxArg) :- 
	( IndxArg > 1 -> %% =\= 1 ->
	    sym_name(Sym,N,A),
	    % message(('% Indexing clauses of ', Sym, ' on argument ', IndxArg))
	    message(('% Indexing clauses of ', N,'/',A, ' on argument ', IndxArg))
	; true
        ).

/*======================================================================*/
/*  makeindex(+Pred, -Pil, -PilR, +Tabled)				*/
/* 	Generates try/retry/trust and index instructions for clauses	*/
/*	of the predicate.						*/
/*======================================================================*/

:- ti makeindex/4.

/*----------------------------------------------------------------------*/
/* single clause predicate cases: no try/index instruction is needed.	*/
/*----------------------------------------------------------------------*/

makeindex(pred(Sym,[clause(_,_,crec(_,Label))],_), Pil, Pilr, Tabled) :- !,
	( Tabled =:= 0 -> Pil = Pilr
	; sym_type(Sym, tabled(TabInd,Arity)),
	  Pil = [tabletrysingle(Arity,Label,TabInd)|Pilr]
	).

/*----------------------------------------------------------------------*/
/* 0-ary predicates generate try instructions only and are NOT indexed.	*/
/*----------------------------------------------------------------------*/

makeindex(pred(Sym,Clauses,_Prag), Pil, Pilr, Tabled) :-
	sym_name(Sym, _, Arity),
	Arity =:= 0,
	!,
	makeNOindex(Sym, Clauses, Arity, Pil, Pilr, Tabled).

/*----------------------------------------------------------------------*/
/* all other cases                                                      */
/*----------------------------------------------------------------------*/

makeindex(pred(Sym,Clauses,Prag), Pil, Pilr, Tabled) :-
	sym_name(Sym, _, Arity),
	pred_indexarg(pred(Sym,Clauses,Prag), Narg),
	( Narg =:= 0 ->
		makeNOindex(Sym, Clauses, Arity, Pil, Pilr, Tabled)
	; index_argpats(Clauses, Narg, ArgList),
	  index_split(ArgList, ArgListGroup),
	  ( ArgListGroup = [_One] -> NeedLabel = 0, Pil2 = Pil
	  ; index_extract_label_list(ArgListGroup, LabelList),
	    index_gentrys(Sym, Arity, LabelList, Pil, Pil2, Tabled),
	    NeedLabel = 1
	  ),
	  index_groups(ArgListGroup, Narg, Pil2, Pilr,
		       NeedLabel, Sym, Arity, Tabled)
	).

/*======================================================================*/
/* This predicate generates try-retry-trust chains for predicates that	*/
/* are not indexed: 0-ary predicates, predicates with an index on arg 0	*/
/* directive, or predicates for which indexing will be ineffective.	*/
/*======================================================================*/

makeNOindex(Sym, Clauses, Arity, Pil, Pilr, Tabled) :-
	index_label_list(Clauses, LabelList),
	index_gentrys(Sym, Arity, LabelList, Pil, Pilr, Tabled).

/*======================================================================*/
/* index_groups(+ArgListGroup, +ArgNum, -Pil, -PilRest, +NeedLabel,	*/
/*		+Sym, +Arity, +Tabled)			                */
/*======================================================================*/

index_groups([], _, Pil, Pil, _, _, _, _).
index_groups([arglabel(T,Group,L)|Rest], Narg, Pil, Pilr,
	     NeedLabel, Sym, Arity, Tabled) :-
	( T == g ->
	   ( NeedLabel =:= 0 -> Pil = Pil1
	   ; gennum(N), L = (inde,N), Pil = [label(inde,L)|Pil1]
           ),
	   index_group(Group, Narg, Pil1, Pil2, Sym, Arity, Tabled)
	; Pil = Pil2
	),
	index_groups(Rest, Narg, Pil2, Pilr, NeedLabel, Sym, Arity, Tabled).

/*======================================================================*/
/* index_group(+Group, +Narg, -Pil, -PilRest, +Sym, +Arity, +Tabled)  	*/
/*	Generates index and try/retry/trust instructions for a group.	*/
/*======================================================================*/

index_group([arglabel(T1,_,L1),arglabel(T2,_,L2)],
	    Narg, Pil, Pilr, Sym, Arity, Tabled) :-
        ( T1 == l, T2 == l -> Pil = Pil1
	; index_sot_inst(T1,T2,Narg,L1,L2,Inst) -> Pil = [Inst|Pil1]
	),
	!,
	index_gentrys(Sym, Arity, [L1,L2], Pil1, Pilr, Tabled).
index_group(Group, Narg, Pil, Pilr, Sym, Arity, Tabled) :-
	( all_args_same(Group, LabelList) -> % indexing makes no sense here
	    sym_name(Sym, P, A),
	    warning(('Indexing of ', P,'/',A, ' on arg ', Narg,
		     ' will be ineffective; ',
		     'consider rewrite or use of the "ti" compiler directive')),
	    Pil3 = Pil
	; gennum(N), L = (inde,N),
	  Pil = [cases(L,Arity,Size,_)|Pil1],
	  index_genarglabel(Group, Pil1, Pil2, LabelList, 0, Size),
	  index_genhashinst(Group, Narg, L, Size, Inst),
	  jit_index(Sym, Narg, Arity, Size, JITI_Insts),
	  append([Inst|JITI_Insts], Pil3, Pil2)
	),
	index_gentrys(Sym, Arity, LabelList, Pil3, Pilr, Tabled).

all_args_same([arglabel(_,_,L)], [L]) :- !.
all_args_same([arglabel(T1,A1,L1),arglabel(T2,A2,L2)|ArgLabs], [L1|Labs]) :-
	T1 = T2, A1 == A2,
	all_args_same([arglabel(T2,A2,L2)|ArgLabs], Labs).

index_genarglabel([], Pil, Pil, [], N, N).
index_genarglabel([ArgLabel|Rest], [ArgLabel|Pil], Pilr, [Lab|Labs], M, N) :-
	ArgLabel = arglabel(_,_,Lab),
	N1 is M+1,
	index_genarglabel(Rest, Pil, Pilr, Labs, N1, N).

:- mode index_genhashinst(+,+,+,+,-).

index_genhashinst(Group, Narg,L,Size, Inst) :-
	( all_atomics(Group) -> Inst = hash_on_atomic(Narg,L,Size)
	; all_structs(Group) -> Inst = hash_on_struct(Narg,L,Size)
	;                       Inst = hash_on_bound(Narg,L,Size)
	).

all_atomics([]).
all_atomics([arglabel(T,_,_)|ArgLabs]) :-
	is_atomic_type(T), all_atomics(ArgLabs).

all_structs([]).
all_structs([arglabel(T,_,_)|ArgLabs]) :- T == s, all_structs(ArgLabs).

is_atomic_type(c).
is_atomic_type(i).
is_atomic_type(n).

/*======================================================================*/

:- mode jit_index(+,++,++,++,-).

jit_index(Sym, Narg, Arity, Sz, Insts) :-
	( jiti_me(Sym, Narg, Arity, Sz) ->
	    sym_name(Sym, Name, _),
	    message(('% JITI opportunity on predicate ', Name, '/', Arity)),
	    jiti_instructions(Narg, Arity, Sz, Insts)
	; Insts = []
	).

jiti_me(Sym, Narg, Arity, Sz) :-
	fail_if(option(jiti_off)),
	sym_has_prop(Sym, edb),		% is an EDB predicate
	fail_if(sym_has_prop(Sym, ti)),	% is not declared as ":- ti"
	Narg < Arity,			% is not indexed on the last arg
	Sz > 1.				% and has multiple (many) clauses

jiti_instructions(Narg, Arity, Size, Insts) :-
	( Narg >= Arity -> Insts = []
        ; NextNarg is Narg+1,
	  Insts = [jiti_on_atomic(NextNarg,Arity,Size)|NextInsts],
	  jiti_instructions(NextNarg, Arity, Size, NextInsts)
        ).

/*======================================================================*/
/* index_gentrys(+Sym, +Arity, +LabelList, -Pil, -PilR, +Tabled)      	*/
/*======================================================================*/

index_gentrys(Sym, Arity, [L1|Ls], Pil, Pilr, Tabled) :-
	( Tabled =:= 0 -> Pil = [try(Arity,L1)|Pilm]
	; sym_type(Sym, tabled(TabInd,Arity)),
	  Pil = [tabletry(Arity,L1,TabInd)|Pilm]
	),
	Ls = [L2|Labs],
	index_genretrys(Labs, Arity, L2, Pilm, Pilr, Tabled).

index_genretrys([], Arity, L, Pil, Pilr, Tabled) :-
	( Tabled =:= 0 -> Pil = [trust(Arity,L)|Pilr]
	;                 Pil = [tabletrust(Arity,L)|Pilr]
	).
index_genretrys([L2|Ls], Arity, L1, Pil, Pilr, Tabled) :-
	( Tabled =:= 0 -> Pil = [retry(Arity,L1)|Pilm]
	;                 Pil = [tableretry(Arity,L1)|Pilm]
	),
	index_genretrys(Ls, Arity, L2, Pilm, Pilr, Tabled).

/*======================================================================*/
/* special switchonterm cases: two clauses, one is atomic type and the	*/
/* other is list or structure type.					*/
/*======================================================================*/

:- ti index_sot_inst/6.

index_sot_inst(i,l,N,L1,L2,switchonlist(N,L1,L2)).
index_sot_inst(i,s,N,L1,L2,switchonterm(N,L1,L2)).
index_sot_inst(n,l,N,L1,L2,switchonlist(N,L1,L2)).
index_sot_inst(n,s,N,L1,L2,switchonterm(N,L1,L2)).
index_sot_inst(c,l,N,L1,L2,switchonlist(N,L1,L2)).
index_sot_inst(c,s,N,L1,L2,switchonterm(N,L1,L2)).
index_sot_inst(l,i,N,L1,L2,switchonlist(N,L2,L1)).
index_sot_inst(l,n,N,L1,L2,switchonlist(N,L2,L1)).
index_sot_inst(l,c,N,L1,L2,switchonlist(N,L2,L1)).
index_sot_inst(s,n,N,L1,L2,switchonterm(N,L2,L1)).
index_sot_inst(s,i,N,L1,L2,switchonterm(N,L2,L1)).
index_sot_inst(s,c,N,L1,L2,switchonterm(N,L2,L1)).

/*======================================================================*/
/* index_label_list(+Clauses, -LabelList)				*/
/*	Obtain the list of labels of all clauses. The labels are	*/
/*	generated (later) by compile_clause/5.				*/
/*======================================================================*/

index_label_list([],[]).
index_label_list([clause(_,_,crec(_,Label))|ClRest],[Label|LRest]) :-
	index_label_list(ClRest,LRest).

/*======================================================================*/
/* index_extract_label_list(+ArgPatterns, -LabelList)			*/
/*	Same as index_label_list/2 but the input is a list of arg	*/
/*	patterns.							*/
/*======================================================================*/

index_extract_label_list([], []).
index_extract_label_list([arglabel(_,_,Label)|Rest], [Label|LRest]) :-
	index_extract_label_list(Rest, LRest).

/*======================================================================*/
/* index_split(+ArgPatterns, +ArgGroups)				*/
/*	Split the argpatterns into a list of arg patterns (ArgGroup)	*/
/*	so that each group contains either a "v" arg or a list of	*/
/*	non-"v" (bound) args.						*/
/*	Precisely, ArgGroups is a list of arglabel(T,Value,Label) 	*/
/*	and when T = g, Value is again a list of arg patterns.		*/
/*======================================================================*/
/* index_split/2 is also used in module tab_CP.				*/
/*======================================================================*/

index_split(ArgPatterns, ArgGroups) :-
	index_split0(ArgPatterns, ArgGroups0),
	index_simplify_singleton(ArgGroups0, ArgGroups).

index_split0([], []).
index_split0([One|Rest], [arglabel(g,Group,_L)|GRest]) :-
	( One = arglabel(v, _, _) -> Group = [One], NRest = Rest
	; index_split1([One|Rest], NRest, Group)
	),
	index_split0(NRest, GRest).

index_split1([], [], []).
index_split1([One|Rest], NRest, NA) :-
	( One = arglabel(v, _, _) -> NRest = [One|Rest], NA = []
	; NA = [One|NNA], index_split1(Rest, NRest, NNA)
	).

index_simplify_singleton([], []).
index_simplify_singleton([Group|Rest], [NGroup|NRest]) :-
	Group = arglabel(g, Patterns, _L),
	( Patterns = [One] -> NGroup = One ; NGroup = Group ),
	index_simplify_singleton(Rest, NRest).

/*======================================================================*/
/* index_argpats(+Clauses, +IndexArg, -ArgPatList)			*/
/*	IndexArg: the argument number to be indexed			*/
/*	ArgPatList: a list of arglabel(Pattern, HashVal, Label)		*/
/*	  where Label is the label for the clause			*/
/*		Pattern 	HashVal					*/
/*		n (nil symbol)	[]					*/
/*		i (integer)	an integer				*/
/*		c (constant)	a Symbol entry				*/
/*		v (variable)	v					*/
/*		l (".")		l					*/
/*		s (structure)	a Symbol entry				*/
/*======================================================================*/

:- mode index_argpats(+,++,-).

index_argpats([], _, []).
index_argpats([clause(Arglist,_,crec(_,L))|CRest], N, 
	      [arglabel(Pattern,NewArg,L)|APatternRest]) :-
        ith(N, Arglist, Arg),
	index_argpat(Arg, Pattern, NewArg),
	index_argpats(CRest, N, APatternRest).

:- mode index_argpat(++,-,-).

index_argpat(varocc(_), v, v).
index_argpat(integer(X), i, X).
index_argpat(real(X), i, X).
index_argpat(constant(Sym), T, Arg) :-
        sym_name(Sym, Name, Arity),
        ( Name == [], Arity =:= 0 -> T = n, Arg = [] ; T = c, Arg = Sym ).
index_argpat(structure(Sym,_), T, Arg) :-
        sym_name(Sym, Name, Arity),
        ( Name == '.', Arity =:= 2 -> T = l, Arg = l ; T = s, Arg = Sym ).

/*======================================================================*/

force_tp_index_loaded.
