/* File:      biassert.c
** Author(s): David S. Warren, Jiyang Xu
** Contact:   xsb-contact@cs.sunysb.edu
** 
** Copyright (C) The Research Foundation of SUNY, 1986, 1993-1998
** Copyright (C) ECRC, Germany, 1990
** 
** XSB is free software; you can redistribute it and/or modify it under the
** terms of the GNU Library General Public License as published by the Free
** Software Foundation; either version 2 of the License, or (at your option)
** any later version.
** 
** XSB is distributed in the hope that it will be useful, but WITHOUT ANY
** WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
** FOR A PARTICULAR PURPOSE.  See the GNU Library General Public License for
** more details.
** 
** You should have received a copy of the GNU Library General Public License
** along with XSB; if not, write to the Free Software Foundation,
** Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
**
** $Id: biassert.c,v 1.49 2006/08/23 15:44:34 kostis Exp $
*/


#include "xxx_config.h"

#include <stdio.h>
#include <errno.h>
#include <string.h>
#include <stdlib.h>
#include <setjmp.h>

#include "auxlry.h"
#include "cell_xsb.h"
#include "error_xsb.h"
#include "memory_xsb.h"
#include "psc_xsb.h"
#include "cinterf.h"
#include "heap_xsb.h"
#include "flags_xsb.h"
#include "deref.h"
#include "inst_xsb.h"
#include "token_xsb.h"
#include "macro_xsb.h"
#include "choice.h"
#include "register.h"
#include "debug.h"
#include "loader_xsb.h"
#include "tr_utils.h"
#include "builtin.h"
#include "biassert.h"

#include "dis.h"
#include "jiti.h"

#include "xxx_inst_sizes.h"

#undef ASSERT_DEBUG
#undef JITI_DEBUG
#undef TRACE_DEBUG
#undef SERIOUS_DEBUG
#undef THREADED_TEAM_DEBUG
#undef RETRACT_GC_DEBUG

#define JIT_INST


#ifdef TRACE_DEBUG
#define DBG_TRACE(x)   fprintf(stderr,">>>> %s\n",x);
#else
#define DBG_TRACE(x)
#endif


#define IS_HEADCLAUSE_INST(x) ((x==getnumcon) ||  \
                               (x==getcon) ||     \
                               (x==getstr) ||     \
                               (x==uninumcon) ||  \
                               (x==unicon) ||     \
                               (x==unitval) ||     \
                               (x==test_heap) ||  \
                               (x==noop) ||       \
                               (x==getlist_tvar_tvar) ||       \
                               (x==getlist) ||       \
                               (x==getstr))

/*======================================================================*/
/* dbgen_inst: Generate an instruction in the buffer.			*/
/*======================================================================*/

#define write_word(Buff,Loc,w) {\
   *(CPtr)((byte *)Buff + *(Loc)) = (Cell)(w); *(Loc) += 4; \
   pad64bits(Loc); \
}
#define write_byte(Buff,Loc,w) {\
   *(byte *)((byte *)Buff + *(Loc)) = (byte)(w); *(Loc) += 1; \
}

#ifdef BITS64
#define pad64bits(Loc)	{ *(Loc) += 4; }
#else
#define pad64bits(Loc)	{}
#endif

#ifdef ASSERT_DEBUG
static void dbgen_printinst3(void * pAddress, byte Opcode, Cell Arg1, Cell Arg2, Cell Arg3)
{
  switch (Opcode) {
  case getlist_tvar_tvar:
    xxx_dbgmsg("0x%x   getlist_tvar_tvar - %ld %ld %ld", (Cell)pAddress, Arg1, Arg2, Arg3);
    break;
  case hash_on_bound:
    xxx_dbgmsg("0x%x   hash_on_bound - %ld 0x%x %ld", (Cell)pAddress, Arg1, Arg2, Arg3);
    break;
  case switchon3bound:
    xxx_dbgmsg("0x%x   switchon3bound - %ld %ld %ld", (Cell)pAddress, Arg1, Arg2, Arg3);
    break;
  case jiti_on_bound:
    xxx_dbgmsg("0x%x   jiti_on_bound - %ld %ld 0x%x\n", (Cell)pAddress, Arg1, Arg2, Arg3); 
    break;
  default: xxx_dbgmsg("0x%x   Unknown instruction in assert %d", (Cell)pAddress, Opcode);
  }
}

static void dbgen_printinst(void * pAddress, byte Opcode, int Arg1, int Arg2)
{
  fprintf(stderr,"0x%x   ", (unsigned int) pAddress); 

  switch (Opcode) {
  case getpvar:	/* PRV */
    fprintf(stderr,"getpvar - %d %d\n", Arg1, Arg2); break;
  case getpval:	/* PRV */
    fprintf(stderr,"getpval - %d %d\n", Arg1, Arg2); break;
  case putpvar:	/* PRV */
    fprintf(stderr,"putpvar - %d %d\n", Arg1, Arg2); break;
  case putpval:	/* PRV */
    fprintf(stderr,"putpval - %d %d\n", Arg1, Arg2); break;
  case gettval:	/* PRR */
    fprintf(stderr,"gettval - %d %d\n", Arg1, Arg2); break;
  case puttvar:	/* PRR */
    fprintf(stderr,"puttvar - %d %d\n", Arg1, Arg2); break;
  case movreg:	/* PRR */
    fprintf(stderr,"movreg - %d %d\n", Arg1, Arg2); break;
  case unipvar:	/* PPV */
    fprintf(stderr,"unipvar - - %d\n", Arg1); break;
  case unipval:	/* PPV */
    fprintf(stderr,"unipval - - %d\n", Arg1); break;
  case bldpvar:	/* PPV */
    fprintf(stderr,"bldpvar - - %d\n", Arg1); break;
  case bldpval:	/* PPV */
    fprintf(stderr,"bldpval - - %d\n", Arg1); break;
  case unitvar:	/* PPR */
    fprintf(stderr,"unitvar - - %d\n", Arg1); break;
  case unitval:	/* PPR */
    fprintf(stderr,"unitval - - %d\n", Arg1); break;
  case bldtvar:	/* PPR */
    fprintf(stderr,"bldtvar - - %d\n", Arg1); break;
  case bldtval:	/* PPR */
    fprintf(stderr,"bldtval - - %d\n", Arg1); break;
  case putlist:	/* PPR */
    fprintf(stderr,"putlist - - %d\n", Arg1); break;
  case getlist:	/* PPR */
    fprintf(stderr,"getlist - - %d\n", Arg1); break;
#ifdef ATTR_VARS
  case getattv: /* PPR */
    fprintf(stderr,"getattv - - %d\n", Arg1); break;
  case putattv: /* PPR */
    fprintf(stderr,"putattv - - %d\n", Arg1); break;
#endif
  case putcon:
    fprintf(stderr,"putcon - - %d 0x%x\n", Arg1, Arg2); break;
  case putnumcon:
    fprintf(stderr,"putnumcon - - %d 0x%x\n", Arg1, int_val(Arg2)); break;
  case putfloat:
    fprintf(stderr,"putfloat - - %d %f (0x%x)\n", Arg1, (Float)Arg2, Arg2); break;
  case getcon:
    fprintf(stderr,"getcon - - %d 0x%x\n", Arg1, Arg2); break;
  case getnumcon:
    fprintf(stderr,"getnumcon - - %d 0x%x\n", Arg1, int_val(Arg2)); break;
  case getfloat:
    fprintf(stderr,"getfloat - - %d %f (0x%x)\n", Arg1, (Float)Arg2, Arg2); break;
  case putstr:
    fprintf(stderr,"putstr - - %d 0x%x\n", Arg1, Arg2); break;
  case getstr:
    fprintf(stderr,"getstr - - %d 0x%x\n", Arg1, Arg2); break;
  case putnil:
    fprintf(stderr,"putnil - - %d\n", Arg1); break;
  case getnil:
    fprintf(stderr,"getnil - - %d\n", Arg1); break;
  case bldcon:
    fprintf(stderr,"bldcon - - - 0x%x\n", Arg1); break;
  case bldnumcon:
    fprintf(stderr,"bldnumcon - - - 0x%x\n", Arg1); break;
  case bldfloat:
    fprintf(stderr,"bldfloat - - - 0x%x\n", Arg1); break;
  case unicon:
    fprintf(stderr,"unicon - - - 0x%x\n", Arg1); break;
  case uninumcon:
    fprintf(stderr,"uninumcon - - - 0x%x\n", Arg1); break;
  case unifloat:
    fprintf(stderr,"unifloat - - - 0x%x\n", Arg1); break;
  case execute:
    fprintf(stderr,"execute - - - 0x%x\n", Arg1); break;
  case bldnil:
    fprintf(stderr,"bldnil - - -\n"); break;
  case uninil:
    fprintf(stderr,"uninil - - -\n"); break;
  case proceed:
    fprintf(stderr,"proceed - - -\n"); break;
  case noop:
    fprintf(stderr,"noop - - -\n"); break;
  case jumptbreg:
    fprintf(stderr,"jumptbreg - - %d 0x%x\n", Arg1, Arg2); break;
  case test_heap:
    fprintf(stderr,"test_heap - - %d %d\n", Arg1, Arg2); break;
  case dyntrustmeelsefail:
    fprintf(stderr,"dyntrustmeelsefail - - %d 0x%x\n", Arg1, Arg2); break;
  case retrymeelse:
    fprintf(stderr,"retrymeelse - - %d 0x%x\n", Arg1, Arg2); break;
  case trymeelse:
    fprintf(stderr,"trymeelse - - %d 0x%x\n", Arg1, Arg2); break;
  case jump:
    fprintf(stderr,"jump - - - 0x%x\n", Arg1); break;
  case fail:
    fprintf(stderr,"fail - - -\n"); break;
  case jiti_on_bound:
    fprintf(stderr,"jiti_on_bound - %d 0x%x \n", Arg1, Arg2); break;
  default: xxx_dbgmsg("Unknown instruction in assert %d",
		      Opcode);
  }
}

#define dbgen_printinst3_macro(Address, Opcode, Arg1, Arg2, Arg3) \
	dbgen_printinst3(Address, Opcode, (int)Arg1, (int)Arg2, (int)Arg3)

#define dbgen_printinst_macro(Address, Opcode, Arg1, Arg2) \
	dbgen_printinst(Address, Opcode, (int)Arg1, (int)Arg2)

#else  /* ASSERT_DEBUG */

#define dbgen_printinst3_macro(Address, Opcode, Arg1, Arg2, Arg3)
#define dbgen_printinst_macro(Address, Opcode, Arg1, Arg2)

#endif /* ASSERT_DEBUG */

#ifdef THREADED
#define write_threaded_inst(Buff,L,Op) write_word(Buff,L,instr_label_tab[Op])
#else
#define write_threaded_inst(Buff,L,Op) {}
#endif

#define dbgen_inst3_tv(Opcode,Arg1,Arg2,Arg3,Buff,Loc) {	\
  dbgen_printinst3_macro((void *)((byte *)Buff + *(Loc)),Opcode, Arg1, Arg2, Arg3);		\
  write_threaded_inst(Buff,Loc,Opcode);				\
  write_byte(Buff,Loc,Opcode); write_byte(Buff,Loc,Arg1);	\
  write_byte(Buff,Loc,Arg2); write_byte(Buff,Loc,Arg3);		\
  pad64bits(Loc);						\
}

#define dbgen_inst3_sob(Opcode,Arg1,Arg2,Arg3,Buff,Loc) {	\
  dbgen_printinst3_macro((void *)((byte *)Buff + *(Loc)),Opcode, Arg1, Arg2, Arg3);		\
  write_threaded_inst(Buff,Loc,Opcode);				\
  write_byte(Buff,Loc,Opcode); write_byte(Buff,Loc,Arg1>>16);	\
  write_byte(Buff,Loc,Arg1>>8); write_byte(Buff,Loc,Arg1);	\
  pad64bits(Loc);						\
  write_word(Buff,Loc,Arg2); write_word(Buff,Loc,Arg3);		\
}

#define dbgen_inst_pvv(Opcode,Arg1,Arg2,Buff,Loc) {	\
  dbgen_printinst_macro((void *)((byte *)Buff + *(Loc)), Opcode, Arg1, Arg2);		\
  write_threaded_inst(Buff,Loc,Opcode);			\
  write_byte(Buff,Loc,Opcode); write_byte(Buff,Loc,0);	\
  write_byte(Buff,Loc,Arg1); write_byte(Buff,Loc,Arg2);	\
  pad64bits(Loc);					\
}

#define dbgen_inst_ppv(Opcode,Arg1,Buff,Loc) {		\
  dbgen_printinst_macro((void *)((byte *)Buff + *(Loc)), Opcode, Arg1, 0);		\
  write_threaded_inst(Buff,Loc,Opcode);			\
  write_byte(Buff,Loc,Opcode); write_byte(Buff,Loc,0);	\
  write_byte(Buff,Loc,0); write_byte(Buff,Loc,Arg1);	\
  pad64bits(Loc);					\
}

#define dbgen_inst_ppvw(Opcode,Arg1,Arg2,Buff,Loc) {	\
  dbgen_printinst_macro((void *)((byte *)Buff + *(Loc)), Opcode, Arg1, Arg2);		\
  write_threaded_inst(Buff,Loc,Opcode);			\
  write_byte(Buff,Loc,Opcode); write_byte(Buff,Loc,0);	\
  write_byte(Buff,Loc,0); write_byte(Buff,Loc,Arg1);	\
  pad64bits(Loc);					\
  write_word(Buff,Loc,Arg2);				\
}

#define dbgen_inst_ppvww(Opcode,Arg1,Arg2,Arg3,Buff,Loc) {	\
  dbgen_printinst_macro((void *)((byte *)Buff + *(Loc)), Opcode, Arg1, Arg2);			\
  write_threaded_inst(Buff,Loc,Opcode);				\
  write_byte(Buff,Loc,Opcode); write_byte(Buff,Loc,0);		\
  write_byte(Buff,Loc,0); write_byte(Buff,Loc,Arg1);		\
  pad64bits(Loc);						\
  write_word(Buff,Loc,Arg2);					\
  write_word(Buff,Loc,Arg3);					\
}

#define dbgen_inst_pppw(Opcode,Arg1,Buff,Loc) {		\
  dbgen_printinst_macro((void *)((byte *)Buff + *(Loc)), Opcode, Arg1, 0);		\
  write_threaded_inst(Buff,Loc,Opcode);			\
  write_byte(Buff,Loc,Opcode); write_byte(Buff,Loc,0);	\
  write_byte(Buff,Loc,0); write_byte(Buff,Loc,0);	\
  pad64bits(Loc);					\
  write_word(Buff,Loc,Arg1);				\
}

#define dbgen_inst_ppp(Opcode,Buff,Loc) {		\
  dbgen_printinst_macro((void *)((byte *)Buff + *(Loc)), Opcode, 0, 0);			\
  write_threaded_inst(Buff,Loc,Opcode);			\
  write_byte(Buff,Loc,Opcode); write_byte(Buff,Loc,0);	\
  write_byte(Buff,Loc,0); write_byte(Buff,Loc,0);	\
  pad64bits(Loc);					\
}

/*
 * Macros for adding instructions to a buffer below.
 */

#define dbgen_instB3_tv(Opcode,Arg1,Arg2,Arg3,IS) {		\
  if (*Loc >= BLim) Buff = buff_realloc(IS);			\
  dbgen_printinst3_macro((void *)((byte *)Buff + *(Loc)), Opcode, Arg1, Arg2, Arg3);	\
  write_threaded_inst(Buff,Loc,Opcode);				\
  write_byte(Buff,Loc,Opcode); write_byte(Buff,Loc,Arg1);	\
  write_byte(Buff,Loc,Arg2); write_byte(Buff,Loc,Arg3);		\
  pad64bits(Loc);						\
}

#define dbgen_instB3_sob(Opcode,Arg1,Arg2,Arg3,IS) {		\
  if (*Loc >= BLim) Buff = buff_realloc(IS);			\
  dbgen_printinst3_macro((void *)((byte *)Buff + *(Loc)), Opcode, Arg1, Arg2, Arg3);		\
  write_threaded_inst(Buff,Loc,Opcode);				\
  write_byte(Buff,Loc,Opcode); write_byte(Buff,Loc,Arg1>>16);	\
  write_byte(Buff,Loc,Arg1>>8); write_byte(Buff,Loc,Arg1);	\
  pad64bits(Loc);						\
  write_word(Buff,Loc,Arg2); write_word(Buff,Loc,Arg3);		\
}

#define dbgen_instB_pvv(Opcode,Arg1,Arg2,IS) {		\
  if (*Loc >= BLim) Buff = buff_realloc(IS);		\
  dbgen_printinst_macro((void *)((byte *)Buff + *(Loc)), Opcode, Arg1, Arg2);		\
  write_threaded_inst(Buff,Loc,Opcode);			\
  write_byte(Buff,Loc,Opcode); write_byte(Buff,Loc,0);	\
  write_byte(Buff,Loc,Arg1); write_byte(Buff,Loc,Arg2);	\
  pad64bits(Loc);					\
}

#define dbgen_instB_ppv(Opcode,Arg1,IS) {	       	\
  if (*Loc >= BLim) Buff = buff_realloc(IS);		\
  dbgen_printinst_macro((void *)((byte *)Buff + *(Loc)), Opcode, Arg1, 0);		\
  write_threaded_inst(Buff,Loc,Opcode);			\
  write_byte(Buff,Loc,Opcode); write_byte(Buff,Loc,0);	\
  write_byte(Buff,Loc,0); write_byte(Buff,Loc,Arg1);	\
  pad64bits(Loc);					\
}

#define dbgen_instB_ppvw(Opcode,Arg1,Arg2,IS) {		\
  if (*Loc >= BLim) Buff = buff_realloc(IS);		\
  dbgen_printinst_macro((void *)((byte *)Buff + *(Loc)), Opcode, Arg1, Arg2);		\
  write_threaded_inst(Buff,Loc,Opcode);			\
  write_byte(Buff,Loc,Opcode); write_byte(Buff,Loc,0);	\
  write_byte(Buff,Loc,0); write_byte(Buff,Loc,Arg1);	\
  pad64bits(Loc);					\
  write_word(Buff,Loc,Arg2);				\
}

#define dbgen_instB_pppw(Opcode,Arg1,IS) {	       	\
  if (*Loc >= BLim) Buff = buff_realloc(IS);		\
  dbgen_printinst_macro((void *)((byte *)Buff + *(Loc)), Opcode, Arg1, 0);		\
  write_threaded_inst(Buff,Loc,Opcode);			\
  write_byte(Buff,Loc,Opcode); write_byte(Buff,Loc,0);	\
  write_byte(Buff,Loc,0); write_byte(Buff,Loc,0);	\
  pad64bits(Loc);					\
  write_word(Buff,Loc,Arg1);				\
}

#define dbgen_instB_ppp(Opcode,IS) {			\
  if (*Loc >= BLim) Buff = buff_realloc(IS);		\
  dbgen_printinst_macro((void *)((byte *)Buff + *(Loc)), Opcode, 0, 0);			\
  write_threaded_inst(Buff,Loc,Opcode);			\
  write_byte(Buff,Loc,Opcode); write_byte(Buff,Loc,0);	\
  write_byte(Buff,Loc,0); write_byte(Buff,Loc,0);	\
  pad64bits(Loc);					\
}


/*======================================================================*/
/* db_cmpl(+Clause, +Buffer, +Index, -Size)                             */
/*      Clause is a fact or rule.                                       */
/*      Buffer is the buffer where the code is put.                     */
/*      Index is the argument to index on (0 if none).                  */
/*      Size is the size of the compiled code                           */
/* The predicate will generate code for the given clause in the Buffer. */
/* The first 8 bytes are reserved for the general chain.  If index is   */
/* requested, the 2nd 8 bytes are used for the bucket chain.  See Code  */
/* below.                                                               */
/*======================================================================*/

static jmp_buf assertcmp_env;

struct flatten_elt {
  union {
    prolog_term term;
    Cell opcode;
  } v;
  int reg;
};

struct instruction {
  byte opcode;
  Cell arg1;
  Cell arg2;
};

#define INST_QUEUE_SIZE	16384	/* was 1024 (which was too low) */
#define FLATTEN_STACK_SIZE 512	

static struct flatten_elt flatten_stack[FLATTEN_STACK_SIZE];
static struct instruction inst_queue[INST_QUEUE_SIZE];
static int flatten_stack_top;
static int inst_queue_top;
static int inst_queue_bottom;

static void assertcmp_throw(int num)
{
  longjmp(assertcmp_env, num);
}


#define ERR_FUNCTOR	1
#define ERR_REGISTER	2

inline static int arity(prolog_term T0)
{
  if (isstruct(T0)) return p2c_arity(T0);
  else if (islist(T0)) return 2;
  else if (isatom(T0)) return 0;
  else assertcmp_throw(ERR_FUNCTOR);
  return -1;
}

static void assertcmp_printerror(int num)
{
  switch (num) {
    case ERR_FUNCTOR:
      xxx_abort("Assert: functor expected");
      break;
    case ERR_REGISTER:
      xxx_abort("Assert: need too many registers");
      break;
    default: 
      xxx_abort("Assert: error occured in assert_cmp");
  }
}

static Integer p2c_float_as_int(prolog_term T0)
{
  union float_conv {
    Float f;
    Integer i;
  } float_conv;
  float_conv.f = float_val(T0);
  return float_conv.i;
}

inline static int is_frozen_var(prolog_term T0)
{
  if (isstruct(T0) && strcmp(p2c_functor(T0), "$assertVAR")==0 &&
      p2c_arity(T0) == 1) {
    T0 = p2p_arg(T0, 1);
    return int_val(T0);
  } else return 0;
}

inline static void flatten_stack_init(void)
{
  flatten_stack_top = 0;
}

inline static int flatten_stack_size(void)
{
  return flatten_stack_top;
}

inline static void flatten_stack_push(struct flatten_elt *flatten_stack,
				      byte bld_opcode, Cell term)
{
  flatten_stack[flatten_stack_top].reg = bld_opcode;
  flatten_stack[flatten_stack_top].v.term = term;
  flatten_stack_top++;
  if (flatten_stack_top >= FLATTEN_STACK_SIZE)
    xxx_abort("flatten_stack overflow in assert");
}

inline static void flatten_stack_pop(struct flatten_elt *flatten_stack,
				     byte *argnop, Cell *termp)
{
  flatten_stack_top--;
  *argnop = flatten_stack[flatten_stack_top].reg;
  *termp = flatten_stack[flatten_stack_top].v.term;
}

inline static void inst_queue_init(void)
{
  inst_queue_top = 0;
  inst_queue_bottom = 0;
}

inline static int inst_queue_empty(void)
{
  return (inst_queue_top == inst_queue_bottom);
}

inline static void inst_queue_push(struct instruction *inst_queue,
				   byte opcode, Cell arg1, Cell arg2)
{
  inst_queue[inst_queue_top].opcode = opcode;
  inst_queue[inst_queue_top].arg1 = arg1;
  inst_queue[inst_queue_top].arg2 = arg2;
  inst_queue_top++;
  if (inst_queue_top >= INST_QUEUE_SIZE)
    xxx_abort("instruction queue overflow in assert");
}

inline static void inst_queue_pop(struct instruction *inst_queue,
				  byte *opcodep, Cell *arg1p, Cell *arg2p)
{
  inst_queue_top--;
  *opcodep = inst_queue[inst_queue_top].opcode;
  *arg1p = inst_queue[inst_queue_top].arg1;
  *arg2p = inst_queue[inst_queue_top].arg2;
}

inline static void inst_queue_rem(struct instruction *inst_queue,
				  byte *opcodep, Cell *arg1p, Cell *arg2p)
{
  *opcodep = inst_queue[inst_queue_bottom].opcode;
  *arg1p = inst_queue[inst_queue_bottom].arg1;
  *arg2p = inst_queue[inst_queue_bottom].arg2;
  inst_queue_bottom++;
}

typedef int *RegStat;
	/* 0 - all rest registers are free */
	/* >0 - next free register */
	/* -1 used for real var */
	/* -2 used for introduced var */

#define RVAR -1
#define TVAR -2

static int RegArray[MAX_REGS];
static int RegArrayInit[MAX_REGS];
static int FreeReg;

static RegStat reg_init(int Size)
{
  int i;
  
  FreeReg = Size+1;
  for (i=0; i<FreeReg; i++) RegArray[i] = RVAR;
  RegArray[FreeReg] = 0;
  return RegArray;
}

/* Type: RVAR=-1 - used for real var; TVAR=-2 - used for introduced var */
inline static int reg_get(int Type)
{
  int new_reg;

  new_reg = FreeReg;
  if (RegArray[FreeReg]==0) {
    FreeReg++;
    if (FreeReg >= MAX_REGS) {
      assertcmp_throw(ERR_REGISTER);
    }
    RegArray[FreeReg] = 0;
  } else FreeReg = RegArray[FreeReg];
  RegArray[new_reg] = Type;
  RegArrayInit[new_reg] = 0;	/* register is not initialized */
  return new_reg;
}

inline static void reg_release(int R0)
{
  if (RegArray[R0]==TVAR) {
    RegArray[R0] = FreeReg;
    FreeReg = R0;
  }
}

/*----------------------------------------------------------------------*/

static char *Buff = NULL;
static int Buff_size = 512;
static int *Loc;
static int BLim = 0;
static int Size;

static char *buff_realloc(int inst_size)
{
  /*  xxx_dbgmsg("Enter buff_realloc(%d) %X", Buff_size,Buff); */
  Buff_size = Buff_size + Buff_size;
  if (Buff == NULL) Buff = (char *)malloc(Buff_size);
  else Buff = (char *)realloc(Buff,Buff_size);
  BLim = Buff_size-inst_size;
  /*  xxx_dbgmsg("Leave buff_realloc(%d) %X", Buff_size,Buff); */
  return Buff;
}

/*----------------------------------------------------------------------*/

static void db_geninst(prolog_term Sub, struct instruction *inst_queue)
{
  int Rt;
  
  if (isinteger(Sub)) {
    dbgen_instB_pppw(uninumcon, Sub, uninumcon_sizeof);
  } else if (isatom(Sub)) {
    dbgen_instB_pppw(unicon, Sub, unicon_sizeof);
  } else if (isnil(Sub)) {
    dbgen_instB_ppp(uninil, uninil_sizeof);
  } else if (isfloat(Sub)) {
    dbgen_instB_pppw(unifloat, p2c_float_as_int(Sub), unifloat_sizeof);
  } else if (isref(Sub)) {
    c2p_functor("$assertVAR", 1, Sub);
    Sub = p2p_arg(Sub, 1);
    Rt = reg_get(RVAR);
    c2p_int(Rt, Sub);
    dbgen_instB_ppv(unitvar, Rt, unitvar_sizeof);
    RegArrayInit[Rt] = 1;  /* reg is inited */
  } else if ((Rt = is_frozen_var(Sub))) {
    dbgen_instB_ppv(unitval, Rt, unitval_sizeof);
  } else
#ifdef ATTR_VARS
         if (isattv(Sub)) {
    /*
     * An ATTV is treated as a real variable, so that the register will
     * never be released.
     */
    Rt = reg_get(RVAR);
    dbgen_instB_ppv(unitvar, Rt, unitvar_sizeof);
    RegArrayInit[Rt] = 1;  /* reg is inited */
    inst_queue_push(inst_queue, Rt, Sub, 0);

    Sub = p2p_arg(Sub, 0);		/* the VAR part of the attv */
    c2p_functor("$assertVAR", 1, Sub);
    Sub = p2p_arg(Sub, 1);
    c2p_int(Rt, Sub);
  } else
#endif
  {
    Rt = reg_get(TVAR);
    dbgen_instB_ppv(unitvar, Rt, unitvar_sizeof);
    RegArrayInit[Rt] = 1;  /* reg is inited */
    inst_queue_push(inst_queue, Rt, Sub, 0);
  }
} /* db_geninst() */

/*----------------------------------------------------------------------*/

static void db_genterms(struct instruction *inst_queue)
{
  byte Argno;
  Cell T0, T1;
  
  while (!inst_queue_empty()) {
    inst_queue_pop(inst_queue, &Argno, &T0, &T1);
    RegArrayInit[Argno] = 1;	/* Reg is initted */
    if (islist(T0)) {
      Cell T2;

      T1 = p2p_car(T0);
      T2 = p2p_cdr(T0);
      if (isref(T1) && isref(T2) && T1!=T2 /* not same var */) {
	int Rt1, Rt2;
	c2p_functor("$assertVAR", 1, T1);
	T1 = p2p_arg(T1, 1);
	Rt1 = reg_get(RVAR);
	c2p_int(Rt1, T1);
	c2p_functor("$assertVAR", 1, T2);
	T2 = p2p_arg(T2, 1);
	Rt2 = reg_get(RVAR);
	c2p_int(Rt2, T2);
	dbgen_instB3_tv(getlist_tvar_tvar, Argno, Rt1, Rt2,
			getlist_tvar_tvar_sizeof);
	RegArrayInit[Rt1] = 1;	/* Reg is initted */
	RegArrayInit[Rt2] = 1;	/* Reg is initted */
	reg_release(Argno);
      } else {
	dbgen_instB_ppv(getlist, Argno, getlist_sizeof);
	reg_release(Argno);
	db_geninst(p2p_car(T0), inst_queue);
	db_geninst(p2p_cdr(T0), inst_queue);
      }
    } else if (isstruct(T0)) {
      dbgen_instB_ppvw(getstr, Argno, get_str_psc(T0), getstr_sizeof);
      reg_release(Argno);
      for (Argno=1; Argno <= get_arity(get_str_psc(T0)); Argno++) {
	db_geninst(p2p_arg(T0,Argno), inst_queue);
      }
    }
#ifdef ATTR_VARS
    else { /* isattv(T0) */
      T1 = cell(clref_val(T0) + 1);	/* the ATTR part of the attv */
      deref(T1);
      dbgen_instB_ppv(getattv, Argno, getattv_sizeof);
      /* The register for a new attv CANNOT be released ! */
      /* reg_release(Argno); */
      db_geninst(T1, inst_queue);
    }
#endif
  }
} /* db_genterms() */

/*----------------------------------------------------------------------*/

static void db_gentopinst(prolog_term T0, int Argno)
{
  int Rt;
  
  if (isinteger(T0)) {
    dbgen_instB_ppvw(getnumcon, Argno, T0, getnumcon_sizeof);
  } else if (isatom(T0)) {
    dbgen_instB_ppvw(getcon, Argno, T0, getcon_sizeof);
  } else if (isfloat(T0)) {
    dbgen_instB_ppvw(getfloat, Argno, p2c_float_as_int(T0), getfloat_sizeof);
  } else if (isref(T0)) {
    c2p_functor("$assertVAR", 1, T0);
    T0 = p2p_arg(T0, 1);
    c2p_int(Argno, T0);
    RegArrayInit[Argno] = 1;	/* Reg is initted */
  } else if (isnil(T0)) {
    dbgen_instB_ppv(getnil, Argno, getnil_sizeof);
  } else if ((Rt = is_frozen_var(T0))) {
    dbgen_instB_pvv(gettval, Rt, Argno, gettval_sizeof);
  } else {
    inst_queue_init();
    inst_queue_push(inst_queue, Argno, T0, 0);
#ifdef ATTR_VARS
    if (isattv(T0)) {
      T0 = p2p_arg(T0, 0);		/* the VAR part of the attv */
      c2p_functor("$assertVAR", 1, T0);
      T0 = p2p_arg(T0, 1);
      c2p_int(Argno, T0);
      RegArrayInit[Argno] = 1;		/* Reg is initted */
    }
#endif
    db_genterms(inst_queue);
  }
} /* db_gentopinst() */

/*----------------------------------------------------------------------*/
/*  Function prototypes.						*/
/*----------------------------------------------------------------------*/

static void db_bldsubs(prolog_term, RegStat, struct flatten_elt *);

/*----------------------------------------------------------------------*/

static void db_putterm(int Rt, prolog_term T0, RegStat Reg)
{
  byte BldOpcode;
  Cell Arg1;
  int  stack_size;
  
  stack_size = flatten_stack_size();
  if (islist(T0)) {
    db_bldsubs(p2p_cdr(T0),Reg,flatten_stack);
    db_bldsubs(p2p_car(T0),Reg,flatten_stack);
    dbgen_instB_ppv(putlist, Rt, putlist_sizeof);
  } else if (isstruct(T0)) {
    int argno;
    for (argno = get_arity(get_str_psc(T0)); argno >= 1; argno--)
      db_bldsubs(p2p_arg(T0,argno),Reg,flatten_stack);
    dbgen_instB_ppvw(putstr, Rt, get_str_psc(T0), putstr_sizeof);
  }
#ifdef ATTR_VARS
  else {			/* is attv */
    db_bldsubs(cell(clref_val(T0)+1), Reg, flatten_stack);
    dbgen_instB_ppv(putattv, Rt, putattv_sizeof);
  }
#endif
  RegArrayInit[Rt] = 1;	/* in any case, reg is inited */
  while (flatten_stack_size() > stack_size) {
    flatten_stack_pop(flatten_stack, &BldOpcode, &Arg1);	
    /* be careful about order!!*/
    switch (BldOpcode) {
    case bldpvar:
      if (RegArrayInit[Arg1]) {
	dbgen_instB_ppv(bldpval, Arg1, bldpval_sizeof);
      } else {
	RegArrayInit[Arg1] = 1;
	dbgen_instB_ppv(bldpvar, Arg1, bldpvar_sizeof);
      }
      break;
    case bldtvar:
      if (RegArrayInit[Arg1]) {
	dbgen_instB_ppv(bldtval, Arg1, bldtval_sizeof);
      } else {
	RegArrayInit[Arg1] = 1;
	dbgen_instB_ppv(bldtvar, Arg1, bldtvar_sizeof);
      }
      break;
    case bldcon:
      dbgen_instB_pppw(bldcon, Arg1, bldcon_sizeof);
      break;
    case bldnumcon:
      dbgen_instB_pppw(bldnumcon, Arg1, bldnumcon_sizeof);
      break;
    case bldfloat:
      dbgen_instB_pppw(bldfloat, Arg1, bldfloat_sizeof);
      break;
    case bldnil:
      dbgen_instB_ppp(bldnil, bldnil_sizeof);
      break;
    default: xxx_dbgmsg("Incorrect bld instruction in assert %d", BldOpcode);
    }
  }
}

/*----------------------------------------------------------------------*/

static void db_bldsubs(prolog_term Sub, RegStat Reg,
		       struct flatten_elt *flatten_stack)
{
  int Rt;
  
  if (isnil(Sub)) {
    flatten_stack_push(flatten_stack, bldnil, 0);
  } else if (isatom(Sub)) {
    flatten_stack_push(flatten_stack, bldcon, Sub);
  } else if (isinteger(Sub)) {
    flatten_stack_push(flatten_stack, bldnumcon, Sub);
  } else if (isfloat(Sub)) {
    flatten_stack_push(flatten_stack, bldfloat, p2c_float_as_int(Sub));
  } else if (isref(Sub)) {
    c2p_functor("$assertVAR", 1, Sub);
    Sub = p2p_arg(Sub, 1);
    Rt = reg_get(RVAR);
    c2p_int(Rt, Sub);
    flatten_stack_push(flatten_stack, bldtvar, Rt);
  } else if ((Rt = is_frozen_var(Sub))) {
    flatten_stack_push(flatten_stack, bldtvar, Rt);
  } else
#ifdef ATTR_VARS
         if (isattv(Sub)) {
    prolog_term T1;

    Rt = reg_get(RVAR);
    flatten_stack_push(flatten_stack, bldtvar, Rt);

    T1 = p2p_arg(Sub, 0);	/* the VAR part of the attv */
    c2p_functor("$assertVAR", 1, T1);
    T1 = p2p_arg(T1, 1);
    c2p_int(Rt, T1);

    /* RegArrayInit[Rt] will be set to 1 in db_putterm() */

    db_putterm(Rt, Sub, Reg);
  } else
#endif
	 {
    Rt = reg_get(TVAR);
    flatten_stack_push(flatten_stack, bldtvar, Rt);
    db_putterm(Rt,Sub,Reg);
  }
} /* db_bldsubs() */

/*----------------------------------------------------------------------*/

static void db_genaput(prolog_term T0, int Argno,
		       struct instruction *inst_queue,
		       RegStat Reg)
{
  int Rt;

  if (isref(T0)) {
    c2p_functor("$assertVAR", 1, T0);
    T0 = p2p_arg(T0, 1);
    Rt = reg_get(RVAR);
    c2p_int(Rt, T0);  /* used to be TempVar???? */
    dbgen_instB_pvv(puttvar, Rt, Rt, puttvar_sizeof);
    RegArrayInit[Rt] = 1;  /* reg is inited */
    inst_queue_push(inst_queue, movreg, Rt, Argno);
  } else if ((Rt = is_frozen_var(T0))) {
    inst_queue_push(inst_queue, movreg, Rt, Argno);
  } else if (isinteger(T0)) {
    inst_queue_push(inst_queue, putnumcon, T0, Argno);
  } else if (isfloat(T0)) {
    inst_queue_push(inst_queue, putfloat, p2c_float_as_int(T0), Argno);
  } else if (isnil(T0)) {
    inst_queue_push(inst_queue, putnil, 0, Argno);
  } else if (isatom(T0)) {
    inst_queue_push(inst_queue, putcon, T0, Argno);
  } else
#ifdef ATTR_VARS
         if (isattv(T0)) {
    prolog_term T1;
    
    Rt = reg_get(RVAR);
    inst_queue_push(inst_queue, movreg, Rt, Argno);
    flatten_stack_init();

    T1 = p2p_arg(T0, 0);		/* the VAR part of the attv */
    c2p_functor("$assertVAR", 1, T1);
    T1 = p2p_arg(T1, 1);
    c2p_int(Rt, T1);
    RegArrayInit[Rt] = 1;		/* Reg is initted */

    db_putterm(Rt,T0,Reg);    
  } else
#endif
         {  /* structure */
    Rt = reg_get(TVAR);
    inst_queue_push(inst_queue, movreg, Rt, Argno);
    flatten_stack_init();
    db_putterm(Rt,T0,Reg);
  }
}

/*----------------------------------------------------------------------*/

inline static bool target_is_not_source(byte Reg)
{
  int i;
  
  for (i=inst_queue_bottom; i<inst_queue_top; i++) {
    if (inst_queue[i].opcode==movreg && inst_queue[i].arg1 == Reg)
      return FALSE;
  }
  return TRUE;
}

inline static bool source_is_not_target(byte Reg)
{
  int i;
  
  for (i=inst_queue_bottom; i<inst_queue_top; i++) {
    if (inst_queue[i].arg2 == Reg) return FALSE;
  }
  return TRUE;
}

/*----------------------------------------------------------------------*/
/* this is a simple routine to generate a series of instructions to
   load a series of registers with constants or from other registers.
   It is given a list of Source,Target pairs.  Target is always a
   register number.  Source may be a putcon(con), putnumcon(num),
   puttvar(reg), puttvar(Var), or movreg(reg).  The registers can
   overlap in any way.  db_genmvs tries to generate a reasonably
   efficient series of instructions to load the indicated registers
   with the indicated values.  */

static void db_genmvs(struct instruction *inst_queue)
{
  byte Opcode;
  Cell Arg, T0;
  
  /* pay attention to the ordering, must be a QUEUE !!!!! */
  while (!inst_queue_empty()) {
    inst_queue_rem(inst_queue, &Opcode, &Arg, &T0);	/* T0: target reg */
    switch (Opcode) {
    case puttvar:  
      dbgen_instB_pvv(puttvar, Arg, T0, puttvar_sizeof);
      break;
    case putnil:
      if (target_is_not_source(T0)) {
	dbgen_instB_ppv(putnil, T0, putnil_sizeof);
      }
      else inst_queue_push(inst_queue, Opcode, Arg, T0);
      break;
    case putcon:
      if (target_is_not_source(T0)) {
	dbgen_instB_ppvw(putcon, T0, Arg, putcon_sizeof);
      }
      else inst_queue_push(inst_queue, Opcode, Arg, T0);
      break;
    case putnumcon:
      if (target_is_not_source(T0)) {
	dbgen_instB_ppvw(putnumcon, T0, Arg, putnumcon_sizeof);
      }
      else inst_queue_push(inst_queue, Opcode, Arg, T0);
      break;
    case movreg:
      if (Arg==T0) break;
      else if (target_is_not_source(T0)) {
	dbgen_instB_pvv(movreg, Arg, T0, movreg_sizeof);
	reg_release(Arg);
      } else if (source_is_not_target(Arg)) /* assume target is source */
	inst_queue_push(inst_queue, movreg, Arg, T0);
      /* delay the instruction at the end */
      /* else if (Arg>T0) dbgen_instB_pvv(movreg,Arg,T0, movreg_sizeof); */
      else {
	byte R0 = reg_get(TVAR);
	dbgen_instB_pvv(movreg, Arg, R0, movreg_sizeof);
	reg_release(Arg);
	inst_queue_push(inst_queue, movreg, R0, T0);
	/* dbgen_instB_pvv(movreg, R0, T0, movreg_sizeof); */
      }
      break;
    }
  }
}

/*======================================================================*/
/*  The following code compiles a clause into a local buffer.  It	*/
/*  treats all rules as though they had a single literal on their	*/
/*  right-hand-side.  Thus it compiles a clause with more than one 	*/
/*  literal on the right-hand-side as a call to the predicate ,/2.	*/
/*======================================================================*/

bool db_assert_code_to_buff(Cell Clause)
{
  prolog_term Head, Body;
  int Location;
  int Loc_size;
  RegStat Reg;
  int Arity;
  int has_body;
  int Argno;

DBG_TRACE("db_assert_code_to_buff");
#ifdef ASSERT_DEBUG
  fprintf(stderr, "%s: ", __FUNCTION__);
  printterm(Clause,1,9);
  fprintf(stderr, "\n");
#endif

  /* set catcher */
  if ((Argno = setjmp(assertcmp_env))) {
    assertcmp_printerror(Argno);
    return FALSE;
  }
  /* if (isstruct(Clause) && strcmp(p2c_functor(Clause),":-")==0 &&
     get_arity(get_str_psc(Clause))==2) { */
  if (isstruct(Clause) && get_str_psc(Clause)==if_psc) {
    Head = p2p_arg(Clause, 1);
    Body = p2p_arg(Clause, 2);
    has_body = 1;
    if (isatom(Body)) {
      if (atom_val(Body) == true_sym) has_body = 0;
      else {
	bool is_new;
	Pair sym = insert(atom_val(Body),0,(Psc)flags[CURRENT_MODULE],&is_new);
	Body = makestruct(hreg);
	new_heap_functor(hreg,sym->psc_ptr);
      }
    }
  } else {
    Head = Clause;
    Body = (prolog_term) NULL;
    has_body = 0;
  }
  Arity = arity(Head);
  Location = 0;
  Loc = &Location;
  /* The 3rd argument below (heap_need) will be backpatched */
  dbgen_instB_ppvw(test_heap, Arity, 0, test_heap_sizeof);
  Loc_size = *Loc - sizeof(Cell);
  if (has_body) Reg = reg_init(max(Arity, get_arity(get_str_psc(Body))));
  else Reg = reg_init(Arity);
  for (Argno = 1; Argno <= Arity; Argno++) {
    db_gentopinst(p2p_arg(Head,Argno),Argno);
  }
  if (has_body) {
    inst_queue_init();
    for (Argno = 1; Argno<=arity(Body); Argno++) {
      db_genaput(p2p_arg(Body,Argno),Argno,inst_queue,Reg);
    }
    db_genmvs(inst_queue);
    dbgen_instB_pppw(execute, get_str_psc(Body), execute_sizeof);
  } else dbgen_instB_ppp(proceed, proceed_sizeof);
  Size = *Loc;
  /* backpatch max heap_need */
  write_word(Buff,&Loc_size,(Size/sizeof(Cell)));
#ifdef ASSERT_DEBUG
   fprintf(stderr, "Backpathcing test_heap to %d\n", Size/sizeof(Cell));
  fprintf(stderr, "... leaving %s() \n", __FUNCTION__);
#endif
  return TRUE;
}

/*
** =====================================================================
** The following byte offsets are valid for 32 bit architectures and
** for not directly threaded code. For 64 bit architectures multiply
** everything by 2.
** =====================================================================
*/

/*======================================================================*/
/* assert_buff_to_clref(+Head,+Arg,+Arity,+Prref,+AZ,+Index,+HashTabSize)	*/
/*	allocates a Clref, copies the byte-code for the clause from	*/
/*	an internal buffer into it, and adds to to the chains.		*/
/*	The arguments are:						*/
/*	Head:  The head of the clause being asserted.			*/
/*	Arg:   The argument value of the indexed arg (ignored if no ind)*/
/*	Arity: the number of registers to save in a choice point.	*/
/*		Note the Arity is one more than the original arity, to  */
/*		hold the cut address.					*/
/*	Prref: predicate reference to which to add the asserted fact	*/
/*	AZ:   0 - inserted as the first clause; 1 - as the last clause	*/
/*	Index:  0 if no index is to be built, or n if an index		*/
/*		on the nth argument of the fact is to be used		*/
/*	HashTabSize:  The size of the hash table to create if one must	*/
/*		be created for this clause (the SOB record)		*/
/*======================================================================*/

/*======================================================================*/
/* Formats for dynamic code:						*/
/* PSC rec point to:							*/
/*	PrRef:								*/
/*		 0: BC instruction: fail (if empty),			*/
/*			jump and save breg (if nonempty)		*/
/*		 4: Addr of first Clref on ALL chain			*/
/*		 8: Addr of last Clref on ALL chain			*/
/*									*/
/* PrRef's point to chain of clRef's (one of 2 types):			*/
/* (the -8 location stores length of buff + flag indicating ClRef type	*/
/*	ClRef0 (for unindexed asserted code):				*/
/*		-8: length of buffer (+0)				*/
/*		-4: Addr of previous ClRef (or PrRef)			*/
/*		 0: Try-type instruction, for chain			*/
/*		 4: (cont) Addr of next ClRef on chain			*/
/*		 8+: BC for asserted clause				*/
/*	ClRef1 (for group of indexed clauses, aka SOB record):		*/
/*		-8: length of buffer (+1)				*/
/*		-4: Addr of previous ClRef (or PrRef)			*/
/*		 0: Try-type instruction, for chain			*/
/*		 4: (cont) Addr of next ClRef on chain			*/
/*		 8: BC switch-on-bound instruction (drop thru if var)	*/
/*		11: (cont) arg to index on				*/
/*		12: (cont) address of Hash Table			*/
/*		16: (cont) size of Hash Table				*/
/*		20: BC jump to	(or fail if empty)			*/
/*		24: (cont) Addr of first ClRefI on all subchain		*/
/*		    or to ClRef1 for next index				*/
/*		28: Addr of last ClRefI on all subchain			*/
/*              32: Number of clauses accessible thru this hash table   */
/*		36+: Hash Table						*/
/*									*/
/* ClRef1's point to indexed clauses, each represented by a ClRefI:	*/
/*	ClRefI (for an indexed clause):					*/
/*		-8: length of buffer (+3)				*/
/*		-4: Addr of previous ClRefI on all chain		*/
/*		 0: Try-type instruction, for all subchain		*/
/*		 4: (cont) Addr of next ClRefI on all subchain		*/
/*            For each index we have the following four fields:         */
/*		 8: BC noop(14) to skip next NI*8-2 bytes		*/
/*		12: Addr of previous ClRefI on bucket chain		*/
/*		16: Try-type instruction, for hash bucket subchain	*/
/*		20: (cont) Addr of next ClRefI in bucket		*/
/*		24: BC noop(6) to skip next (NI-1)*8-2 bytes		*/
/*		28: Addr of previous ClRefI on bucket chain		*/
/*		32: Try-type instruction, for hash bucket subchain	*/
/*		36: (cont) Addr of next ClRefI in bucket		*/
/*	   NI*16+8: BC for asserted code				*/
/*									*/
/*======================================================================*/

/* Predicate References */

typedef struct {

#ifdef THREADED
  Cell  threaded_inst;  /* must be the first field of the struct */
#endif
  Cell	inst_opcode;    /* must appear immediately afterwards */
  ClRef first_clref;
  ClRef last_clref;
} *PrRef, PrRefData;

#define prref_inst_opcode(PRREF) get_opcode((PRREF))
#define set_prref_inst(PRREF,OP) set_opcode((PRREF),(OP))
#define prref_first_clref(PRREF) ((PrRef)(PRREF))->first_clref
#define prref_last_clref(PRREF)  ((PrRef)(PRREF))->last_clref

/* Clause References */

#define ClRefAddr(Cl)		((CPtr)((ClRef)(Cl)-1))
#define ClRefSize(Cl)		(((ClRef)(Cl))[-1].buflen & ~0x3)
#define ClRefType(Cl)		(((ClRef)(Cl))[-1].buflen & 0x3)
#define SetClRefSize(Cl,len)	(((ClRef)(Cl))[-1].buflen |= \
		(((ClRef)(Cl))[-1].buflen & 0x3) | ((len) & ~0x3))
#define SetClRefType(Cl,type)	(((ClRef)(Cl))[-1].buflen = \
		(((ClRef)(Cl))[-1].buflen & ~0x3) | ((type) & 0x3))
#define ClRefPrev(Cl)		(((ClRef)(Cl))[-1].prev)
#define SetClRefPrev(Cl,Prv)	(((ClRef)(Cl))[-1].prev = (ClRef)(Prv))

/* Clause types */

#define UNINDEXED_CL	0
#define SOB_RECORD	1
#define TRASSERTED_CL	2
#define INDEXED_CL	3

/*
** sz below is rounded up tp the nearest 8 bytes;
** type is stored in buflen's two lower bits.
*/
#define MakeClRef(ptr,Type,NCells)\
{   long sz = (((NCells)*sizeof(Cell)+sizeof(ClRefHdr) + 7) & ~0x7);\
    (ptr) = (ClRef)mem_alloc(sz);\
    (ptr)->buflen = ((Type)&3)+(sz&~3);\
    (ptr)++;\
}

#define ClRefWord(Cl,pos)	(((CPtr)(Cl))[(pos)])

/* Clause common fields */

#define ClRefTryInstr(Cl)	(ClRefWord((Cl),0))
#define ClRefTryOpCode(Cl)	(get_opcode(&ClRefTryInstr(Cl)))

#define NEXT_CLREF_OFFSET       (try_arg_2/sizeof(word))
#define ClRefNext(Cl)		(ClRefWord((Cl),NEXT_CLREF_OFFSET))
#define SetClRefNext(Cl,Nxt)	(ClRefNext((Cl)) = (Cell)(Nxt))

/*
** First byte code in clause - Jump/SOB/etc
** The macro below assumes that try_sizeof == jump_sizeof
*/
#define ENTRY_POINT_OFFSET	(try_sizeof/sizeof(word))  /* WAS 2 */
#define ClRefEntryPoint(Cl)	(&ClRefWord((Cl),ENTRY_POINT_OFFSET))

#define HASH_SIZE_OFFSET	(ENTRY_POINT_OFFSET+(switchon3bound_arg_5/sizeof(word)))

#ifdef JIT_INST
#define JUMP_SEG_JITI_OFFSET  (ENTRY_POINT_OFFSET+switchon3bound_wordsof)
#define JUMP_INSTR_OFFSET  (JUMP_SEG_JITI_OFFSET+jump_wordsof)

#else
#define JUMP_INSTR_OFFSET  (ENTRY_POINT_OFFSET+switchon3bound_wordsof)

#endif


#define FIRST_INDEX_OFFSET	(JUMP_INSTR_OFFSET+(jump_arg_1/sizeof(word)))
#define LAST_INDEX_OFFSET	(JUMP_INSTR_OFFSET+jump_wordsof)

/*
#define FIRST_INDEX_OFFSET  (JUMP_INSTR_OFFSET+jump_sizeof)
#define LAST_INDEX_OFFSET (FIRST_INDEX_OFFSET+1)
*/
#define NUM_CLAUSES_OFFSET	(LAST_INDEX_OFFSET+1)
#define HASH_TABLE_OFFSET	(NUM_CLAUSES_OFFSET+1)

/* For indexed clause group (SOBblock) */

#define ClRefSOBInstr(Cl)	(ClRefWord((Cl),ENTRY_POINT_OFFSET))
#define ClRefHashSize(Cl)	(ClRefWord((Cl),HASH_SIZE_OFFSET))
#ifdef JIT_INST
#define ClRefJumpJiti(Cl) (ClRefWord((Cl),JUMP_SEG_JITI_OFFSET))
#endif
#define ClRefJumpInstr(Cl)	(ClRefWord((Cl),JUMP_INSTR_OFFSET))
#define ClRefFirstIndex(Cl)	(ClRefWord((Cl),FIRST_INDEX_OFFSET))
#define ClRefLastIndex(Cl)	(ClRefWord((Cl),LAST_INDEX_OFFSET))
#define ClRefNumClauses(Cl)	(ClRefWord((Cl),NUM_CLAUSES_OFFSET))
#define ClRefHashTable(Cl)	(&ClRefWord((Cl),HASH_TABLE_OFFSET))

#define ClRefSOBArg(Cl,n)	(inst_operandn(&ClRefSOBInstr(Cl),(n)))

/* Get the PrRef field of a SOB */
#define ClRefPrRef(Cl)		((PrRef)&ClRefJumpInstr(Cl))

/* Get the ClRef containing the PrRef */
#ifdef JIT_INST
#ifdef THREADED
#define WEIRD_OFFSET 10
#else
#define WEIRD_OFFSET 7
#endif
#else  /* Without JIT_INST */
#ifdef THREADED
#define WEIRD_OFFSET 7
#else
#define WEIRD_OFFSET 5
#endif
#endif
#define PrRefClRef(Pr)		((ClRef)((CPtr)(Pr)-WEIRD_OFFSET))

#define ClRefUpSOB(Cl)		(PrRefClRef((ClRefNext(Cl))))

/*
** A predicate may have multiple indexes, e.g. an index on arg 1 and 2
** and 3. So a clause for such a predicate needs 3 pieces of code, one
** for each index. Each index requires words to contain:
**
**   a. a noop (really skip) instruction to skip over the rest of this
**      entry if it's not to be used for a call.
**   b. the address of the previous ClRefI on the chain linking
**      clauses in a single bucket together
**   c. a try-type instruction for the hash bucket chain
**   d. rest of the try-type instruction: address of next ClRef in
**      the hash bucket chain.
**
** So the first define below denotes the number of words needed to
** implement a single index, and the second the offset where the first
** index segment comes in a ClRefI: immediately after bypassing the
** try-type instruction.
*/
#define WORDS_PER_INDEX		(noop_wordsof+1+try_wordsof)
#define FIRST_INDEX_SEG_OFFSET	try_wordsof
#define IC_CELLS(NI)		(WORDS_PER_INDEX*(NI)+FIRST_INDEX_SEG_OFFSET)

/* For Indexed clause index table */

#define ClRefNumInds(Cl)\
	 ( (inst_operand3(&ClRefSOBInstr(Cl))/(sizeof(Cell)/2) + noop_wordsof)/WORDS_PER_INDEX )
#define ClRefIndPtr(Cl,Ind)	(&ClRefWord((Cl),WORDS_PER_INDEX*(Ind)))

#define IndPtrClRef(IP,Ind)	((ClRef)((CPtr)(IP)-WORDS_PER_INDEX*(Ind)))
/* #define IndRefNoop(IndPtr)	((IndPtr)[-(noop_wordsof+1)]) */
#define IndRefPrev(IndPtr)	(((CPtr *)(IndPtr))[-1])
#define IndRefTryInstr(IndPtr)	((IndPtr)[0])
#define IndRefNext(IndPtr)	(((CPtr *)(IndPtr))[NEXT_CLREF_OFFSET])

#define ClRefIEntryPoint(Cl,NI)	(&ClRefWord((Cl),IC_CELLS(NI)))

/* First word of code in an (un)indexed clause */
#define ClRefEntryAny(Cl) 						\
	((CPtr)								\
		((ClRefType(Cl)&1) == UNINDEXED_CL ? ClRefEntryPoint(Cl) :\
				ClRefIEntryPoint((Cl),ClRefNumInds(Cl))	\
	))

#define ClRefNotRetracted(Cl) (get_opcode(ClRefEntryAny(Cl)) != fail || \
                               inst_operand1(ClRefEntryAny(Cl)) != 66)

/*======================================================================*/

static byte * MakeJitiSegm(byte *ptrJumpAddress,int iArity,byte *ptrPredicate)
{ 
  byte *pResult; 
  long lIndex;
  long lLocation=0;
  long szSize;

DBG_TRACE("MakeJitiSegm");
#ifdef JITI_DEBUG
  fprintf(stderr, "MakeJitiSegm 0x%x %d 0x%x\n", (unsigned int)ptrJumpAddress, iArity, (unsigned int)ptrPredicate);
#endif
  /* create segment of memory */
  szSize = (jiti_on_bound_sizeof * (iArity-2)) + jump_sizeof;
  pResult = mem_alloc(szSize);
  /* fill the new segments with the jiti_on_bound instractions */
  for(lIndex = 2; lIndex <= iArity-1; lIndex++)
  {
    dbgen_inst_ppvww(jiti_on_bound, lIndex,(byte *)ptrPredicate, iArity, pResult, &lLocation)
  }
   /* last instraction */
  dbgen_inst_pppw(jump, ptrJumpAddress, pResult, &lLocation);

  return pResult;
}
/*======================================================================*/
static void InvalidateJitiSegm(byte *ptrJitiSegm,int iArity,byte *ptrPredicate)
{
  long lIndex;
  byte opCode;

DBG_TRACE(__FUNCTION__);
  /* iterate instructions */
  for(lIndex = 2; lIndex <= iArity-1; lIndex++)
  {
    opCode = get_opcode(ptrJitiSegm);
#ifdef JITI_DEBUG
    fprint_inst(stderr, ptrJitiSegm);
#endif
    if( (opCode == hash_on_bound) || (opCode == noop))
    { 
      long lLocation = 0;
      DBG_TRACE("release memory of hash_table");
      dbgen_inst_ppvww(jiti_on_bound, lIndex, (byte *)ptrPredicate, iArity, ptrJitiSegm, &lLocation)
    }

    /* this code assumes jiti_on_bound and hash_on_bound have the same size */
    ptrJitiSegm += jiti_on_bound_sizeof;
  } 
}


/*======================================================================*/

/* add NewClause to beginning of try-retry chain beginning with FirstClause */
static void prefix_to_chain(byte arity, ClRef FirstClause, ClRef NewClause)
{
  int Loc = 0;

DBG_TRACE("prefix_to_chain");
  if (ClRefTryOpCode(FirstClause) == noop)
  {  dbgen_inst_ppvw(dyntrustmeelsefail,arity,ClRefNext(FirstClause),
		     FirstClause,&Loc); }
  else if (ClRefTryOpCode(FirstClause) == trymeelse)
  {  dbgen_inst_ppvw(retrymeelse,arity,ClRefNext(FirstClause),
		     FirstClause,&Loc);}
  else xxx_dbgmsg("***Error 1 in assert: 0x%x", ClRefTryOpCode(FirstClause));

  ClRefPrev(NewClause)   = ClRefPrev(FirstClause);
  ClRefPrev(FirstClause) = NewClause;

  Loc = 0;
  dbgen_inst_ppvw(trymeelse,arity,FirstClause,NewClause,&Loc);
}

/* add NewClause after LastClause on try-retry chain */
static void append_to_chain(byte arity, ClRef LastClause, ClRef NewClause)
{
  int Loc = 0;

DBG_TRACE("append_to_chain");

  dbgen_inst_ppvw(dyntrustmeelsefail,arity,ClRefNext(LastClause),
		  NewClause,&Loc);
  Loc = 0;
  if (ClRefTryOpCode(LastClause) == noop) {
    dbgen_inst_ppvw(trymeelse,arity,NewClause,LastClause,&Loc);
  }
  else
    if (ClRefTryOpCode(LastClause) == dyntrustmeelsefail) {
      dbgen_inst_ppvw(retrymeelse,arity,NewClause,LastClause,&Loc);
    }
    else xxx_dbgmsg("***Error 2 in assert: 0x%x", ClRefTryOpCode(LastClause));

  SetClRefPrev(NewClause, LastClause);
}

static void db_addbuff(byte arity, ClRef Clause, PrRef Pred, int AZ, int Inum) 
{
  byte inst;

DBG_TRACE("db_addbuff");

  inst = prref_inst_opcode(Pred);

#ifdef THREADED_TEAM_DEBUG
  fprintf(stderr, "THREADED-TEAM DEBUG: db_addbuff .. \n");
#endif
  if (inst == fail) {
    int Loc = 0;
#ifdef THREADED_TEAM_DEBUG
    fprintf(stderr, "THREADED-TEAM DEBUG: PredOpCode == fail .. \n");
#endif
    if (Inum > 1) {dbgen_inst_pppw(jump,Clause,Pred,&Loc);}
    else {dbgen_inst_ppvw(jumptbreg,arity,Clause,Pred,&Loc);}
    prref_last_clref(Pred) = Clause ;
    SetClRefPrev(Clause, Pred) ;
    Loc = 0;
    dbgen_inst_ppv(noop,sizeof(Cell)/2,Clause,&Loc);
    SetClRefNext(Clause, Pred) ;
  } else if (inst == jumptbreg || inst == jump) {
#ifdef THREADED_TEAM_DEBUG
    fprintf(stderr, "THREADED-TEAM DEBUG: PredOpCode == jump .. \n");
#endif
    if (AZ == 0) {
      prefix_to_chain(arity, prref_first_clref(Pred), Clause);
      prref_first_clref(Pred) = Clause ;
    } else {
      ClRef LastClause = prref_last_clref(Pred) ;
      append_to_chain(arity,LastClause,Clause);
      prref_last_clref(Pred) = Clause ;
    }
  } else xxx_dbgmsg("***Error 3 in assert");

#ifdef THREADED_TEAM_DEBUG
  fprintf(stderr, "THREADED-TEAM DEBUG: leaving db_addbuff ! \n");
#endif
}

/*======================================================================*/

/* add Clause to end of pred's prref */
static unsigned int hash_resize(PrRef prref, ClRef SOBrec,
				const unsigned int OldTabSize)
{
DBG_TRACE("hash_resize");
/* xxx_dbgmsg("SOB - %p, with %d cls", SOBrec, ClRefNumClauses(SOBrec) ) ; */
   /* Compute number of clauses */
  if (prref_inst_opcode(prref) != fail && ClRefType(SOBrec) == SOB_RECORD )
    {
      unsigned int ThisTabSize = ClRefHashSize(SOBrec) ;
      if (ClRefNumClauses(SOBrec)+ClRefNumClauses(SOBrec)/2 >= ThisTabSize)
	ThisTabSize = 2*ThisTabSize+1 ;
      return max(ThisTabSize, OldTabSize) ;
    }
  else return OldTabSize ;
}

static bool can_hash(int Ind, Cell head)
{
DBG_TRACE("can_hash");
  if (Ind < 256) {  /* handle usual case specially */
    Cell arg = p2p_arg(head, Ind);
    return (!isref(arg) && !isattv(arg));
  } else {
    int i, j;
    for (i = 2; i >= 0; i--) {
      j = (Ind >> (i*8)) & 0xff;
      if (j > 0) {
	Cell arg = p2p_arg(head,j);
	if (isref(arg) || isattv(arg)) return FALSE;
      }
    }
  }
  return TRUE;
}

inline static int hash_val(int Ind, Cell head, int TabSize)
/* assumes we can hash to this Ind */
{
  register int Hashval = 0;

DBG_TRACE("hash_val");

  if (Ind < 256) {  /* handle usual case specially */
    Cell arg = p2p_arg(head,Ind) ;
    Hashval = ihash(val_to_hash(arg), TabSize);
  } else {   /* handle joint indexes */
    int i, j;
    for (i = 2; i >= 0; i--) {
      j = (Ind >> (i*8)) & 0xff;
      if (j > 0) {
        Cell arg = p2p_arg(head,j);
        Hashval += Hashval + ihash(val_to_hash(arg), TabSize);
      }
    }
    Hashval %= TabSize;
  }
#ifdef JITI_DEBUG
  fprintf(stderr,"HASH_VAL %d 0x%x %d\n",Ind,head,TabSize);
#endif

  return Hashval;
}

static ClRef new_SOBblock(int ThisTabSize, int Ind)
{
  int   i, Loc ;
  ClRef NewSOB ;

DBG_TRACE("new_SOBblock");

  /* get NEW SOB block */
  MakeClRef(NewSOB,SOB_RECORD,HASH_TABLE_OFFSET+ThisTabSize);
  /*   xxx_dbgmsg("New SOB %p, size = %d", NewSOB, ThisTabSize); */
  Loc = 0 ;
  dbgen_inst3_sob( Ind>255 ? switchon3bound : hash_on_bound,
		   Ind,ClRefHashTable(NewSOB),ThisTabSize,
		   &ClRefSOBInstr(NewSOB),&Loc);

#ifdef JIT_INST
  Loc = 0;
  dbgen_inst_ppv(noop,sizeof(Cell)/2,&ClRefJumpJiti(NewSOB),&Loc);
  write_word(&ClRefJumpJiti(NewSOB),&Loc,0);
#endif
  /* set the PrRef inside SOB */
  Loc = 0 ;
  dbgen_inst_ppp(fail,&ClRefJumpInstr(NewSOB),&Loc);
  ClRefFirstIndex(NewSOB) = (Cell)&ClRefJumpInstr(NewSOB) ;
  ClRefLastIndex( NewSOB) = (Cell)&ClRefJumpInstr(NewSOB) ;
  ClRefNumClauses(NewSOB) = 0 ;
  
  /* Initialize hash table */
  for (i = 0; i < ThisTabSize; i++)
    ClRefHashTable(NewSOB)[i] = (Cell)&fail_inst ;
  
  return NewSOB ;
}

static void addto_hashchain(const int AZ, int Hashval, ClRef SOBrec,
			    CPtr NewInd, const int arity)
{


  CPtr *Bucketaddr = (CPtr *) (ClRefHashTable(SOBrec) + Hashval);
  CPtr OldInd = *Bucketaddr ;
  int Loc ;
DBG_TRACE("addto_hashchain");

  if ((byte *)OldInd == (byte *)&fail_inst) {
    /* empty bucket, add first clause */
    *Bucketaddr = NewInd ;
    IndRefPrev(NewInd) = (CPtr) Bucketaddr ;
    IndRefNext(NewInd) = (CPtr) SOBrec ;
  } else if (AZ == 0) { /* add at beginning */
    *Bucketaddr = NewInd ;
    IndRefPrev(NewInd) = (CPtr) Bucketaddr ;
    Loc = 0;
    if (get_opcode(OldInd) == noop)
      {  dbgen_inst_ppvw(dyntrustmeelsefail,arity,IndRefNext(OldInd),
			 OldInd,&Loc);
      }
    else
      {  dbgen_inst_ppvw(retrymeelse,arity,IndRefNext(OldInd),
			 OldInd,&Loc);
      }
    Loc = 0;
    dbgen_inst_ppvw(trymeelse,arity,OldInd,NewInd,&Loc);
    IndRefPrev(OldInd) = NewInd;
  } else { /* AZ == 1 add at end */
    Loc = 0;
    if (get_opcode(OldInd) == noop)
      {  dbgen_inst_ppvw(trymeelse,arity,NewInd,OldInd,&Loc); }
    else {
      while (get_opcode(OldInd) != dyntrustmeelsefail)
	OldInd = IndRefNext(OldInd);
      dbgen_inst_ppvw(retrymeelse,arity,NewInd,OldInd,&Loc);
    }
    Loc = 0;
    dbgen_inst_ppvw(dyntrustmeelsefail,arity,SOBrec,NewInd,&Loc);
    IndRefPrev(NewInd) = OldInd ;
  }
}

static void addto_allchain(int AZ, ClRef Clause, ClRef SOBrec, byte arity)
{
  ClRef Last, First ;
  PrRef prref;
DBG_TRACE("addto_allchain");
  prref = ClRefPrRef(SOBrec);

  /* add code buff to all chain */
  if (prref_inst_opcode(prref) == fail) { /* insert first clrefI into SOB buff */
    int Loc = 0 ;
    Loc=0;

/*    dbgen_inst_pppw(jump,Clause,(byte *)(&ClRefJumpInstr(SOBrec)),&Loc);*/

    dbgen_inst_pppw(jump,Clause,ClRefPrRef(SOBrec),&Loc);
    ClRefLastIndex(SOBrec) = (Cell) Clause ;
 /*   ClRefFirstIndex(SOBrec) = (Cell) Clause ; */
    ClRefPrev(Clause) = SOBrec ;
    Loc = 0;
    dbgen_inst_ppv(noop,sizeof(Cell)/2,Clause,&Loc);
    SetClRefNext(Clause, SOBrec);
  } else if (AZ == 0) {  /* add at beginning */
    First = (ClRef) ClRefFirstIndex(SOBrec);
    prefix_to_chain(arity,First,Clause);
    ClRefPrev(First) = Clause;
    ClRefFirstIndex(SOBrec) = (Cell) Clause;
  } else {  /* add at end */
    Last = (ClRef) ClRefLastIndex(SOBrec);
    append_to_chain(arity, Last, Clause);
    ClRefPrev(Clause) = Last ;
    ClRefLastIndex(SOBrec) = (Cell) Clause;
  }
}

#ifdef SERIOUS_DEBUG
#define THREADED_TEAM_DEBUG
#endif

/* adds an indexed buffer to an index chain */
static void db_addbuff_i(byte arity, ClRef Clause, PrRef Pred, int AZ,
			 int *Index, int num_indx, Cell Head, int HashTabSize)
{
  ClRef SOBbuff ;

  int Inum, Ind;
  unsigned int ThisTabSize;
  int Hashval;
  PrRef pMainPredicate=0;
  byte *pJitiSegmentJump;

DBG_TRACE("db_addbuff_i");

  pMainPredicate = Pred;

#ifdef ASSERT_DEBUG
  fprintf(stderr, "%s: ", __FUNCTION__);
  fprintf(stderr, "INDEX: %d \n", num_indx);
  printterm(Head,1,10);
  fprintf(stderr, " .. HashSz = %d\n", HashTabSize);
#endif
  SOBbuff = AZ == 0 ? prref_first_clref(Pred) : prref_last_clref(Pred) ;
#ifdef THREADED_TEAM_DEBUG
  fprintf(stderr, "THREADED-TEAM DEBUG: hash_resize ... ");
#endif
  HashTabSize = ThisTabSize = hash_resize(Pred, SOBbuff, HashTabSize);
#ifdef THREADED_TEAM_DEBUG
  fprintf(stderr, ".. done (new HashSz = %d) \n", HashTabSize);
#endif
  for (Inum = 1; Inum <= num_indx; Inum++) {
    SOBbuff = AZ == 0 ? prref_first_clref(Pred) : prref_last_clref(Pred) ;
    Ind = Index[Inum];
    if ( can_hash(Ind, Head) )
      Hashval = hash_val(Ind, Head, ThisTabSize) ;
    else { Hashval = 0; ThisTabSize = 1; }
    if (prref_inst_opcode(Pred) == fail || ClRefType(SOBbuff) != SOB_RECORD
	|| ClRefHashSize(SOBbuff) != ThisTabSize
	|| ClRefSOBArg(SOBbuff,1) != (byte)(Ind>>16)  /* for byte-back */
	|| ClRefSOBArg(SOBbuff,2) != (byte)(Ind>>8)
	|| ClRefSOBArg(SOBbuff,3) != (byte)Ind) {
#ifdef THREADED_TEAM_DEBUG
      fprintf(stderr, "THREADED-TEAM DEBUG: creating SOBblock .. \n");
#endif
      SOBbuff = new_SOBblock(ThisTabSize,Ind);

#ifdef JIT_INST
      if((AZ == 0) && (num_indx == 1))
      {
        /* remove JITI SEGMENT in Pred (if there is any SOB)*/
        byte inst = prref_inst_opcode(pMainPredicate);
        if (inst != fail) 
        {
          int Loc = 0;
          ClRef firstSOB = prref_first_clref(pMainPredicate);
          DBG_TRACE("##############  remove JITI SEGMENT in Pred (if there is any SOB)");

          /* invalidate hash_on_bound instructions */
          pJitiSegmentJump = &ClRefJumpJiti(firstSOB);
          pJitiSegmentJump = *(CPtr)(pJitiSegmentJump+jump_arg_1);
          InvalidateJitiSegm(pJitiSegmentJump, arity, 0);
          /* remove jiti segment */
          free(pJitiSegmentJump);

          dbgen_inst_ppv(noop, sizeof(Cell)/2, &ClRefJumpJiti(firstSOB), &Loc);
          write_word(&ClRefJumpJiti(firstSOB), &Loc, 0);
        }
      }
#endif
      /* add new SOB block */
      db_addbuff(arity,SOBbuff,Pred,AZ,Inum);
    }
#ifdef THREADED_TEAM_DEBUG
    fprintf(stderr, "THREADED-TEAM DEBUG: get Pred from SOBblock .. \n");
#endif
    ClRefNumClauses(SOBbuff)++ ;
    Pred = ClRefPrRef(SOBbuff) ;
#ifdef THREADED_TEAM_DEBUG
    fprintf(stderr, "THREADED-TEAM DEBUG: addto_hashchain .. \n");
#endif
#ifdef JITI_DEBUG
    fprintf(stderr, "SOB 0x%x\n",(unsigned int) SOBbuff);
#endif

    addto_hashchain(AZ, Hashval, SOBbuff, ClRefIndPtr(Clause,Inum), arity);
  }
#ifdef THREADED_TEAM_DEBUG
  fprintf(stderr, "THREADED-TEAM DEBUG: addto_allchain .. \n");
#endif

  addto_allchain(AZ, Clause, SOBbuff, arity) ;

#ifdef JIT_INST 
  if (num_indx==1) 
  {
    ClRef firstSOB=prref_first_clref(pMainPredicate);
    if(get_opcode(&ClRefJumpJiti(firstSOB))==noop)
    {
      int Loc=0;
      DBG_TRACE("##############  add JITI_SEGMENT in SOBBlock");
      pJitiSegmentJump=MakeJitiSegm((byte *)&ClRefJumpInstr(firstSOB),arity,(byte *)pMainPredicate);
      dbgen_inst_pppw(jump,pJitiSegmentJump,&ClRefJumpJiti(firstSOB),&Loc);
    }
    else
    {
      DBG_TRACE("##############  invalidate JITI_SEGMENT of first SOBBlock");
      pJitiSegmentJump=&ClRefJumpJiti(firstSOB);
      pJitiSegmentJump=*(CPtr)(pJitiSegmentJump+jump_arg_1);
      InvalidateJitiSegm(pJitiSegmentJump,arity,(byte *)pMainPredicate);
    }
  }
#endif
#ifdef ASSERT_DEBUG
  fprintf(stderr, "... leaving %s()\n", __FUNCTION__);
#endif


}

/*======================================================================*/

#include "ptoc_tag_xsb_i.h"
#include "term_psc_xsb_i.h"

/*======================================================================*/
/* Some auxilliary stuff to allow asserting tabled predicates.          */
/*======================================================================*/

/*----------------------------------------------------------------------
  The entry point clause for all dynamic tabled predicates starts with
  the following fixed sequence of instructions implicitly shown below.
  ----------------------------------------------------------------------*/
#define FIXED_BLOCK_SIZE_FOR_TABLED_PRED \
		( tabletrysingle_sizeof	 \
		+ allocate_gc_sizeof	 \
		+ getVn_sizeof		 \
		+ calld_sizeof		 \
		+ new_answer_sizeof )
/* Hence the following macro (size is in number of cells). */
#define PRREF_OFFSET_FOR_DYN_TABLED_PRED \
	       (( tabletrysingle_sizeof  \
		+ allocate_gc_sizeof	 \
		+ getVn_sizeof		 \
		+ calld_arg_2 ) / sizeof(Cell))

inline static PrRef get_real_prref(Psc psc)
{
DBG_TRACE("get_real_prref");
  if (get_tip(psc)) /* is a tabled predicate */
    return (PrRef)*((CPtr *)get_ep(psc)+PRREF_OFFSET_FOR_DYN_TABLED_PRED);
  else return (PrRef)get_ep(psc);
}

static PrRef build_and_return_prref(Psc psc, bool tabled)
{
  PrRef p;
  int Loc;
DBG_TRACE("build_and_return_prref");
  set_type(psc, T_DYNA) ;
  set_env(psc, T_VISIBLE) ;
  p = (PrRef)mem_alloc(sizeof(PrRefData)) ;
  Loc = 0 ;

  dbgen_inst_ppp(fail,p,&Loc) ;
  prref_first_clref(p) = prref_last_clref(p) = (ClRef)p ;
  if ( tabled )
    {
      CPtr tp;
      byte arity;
      TIFptr tip;

      New_TIF(tip,psc);
      tp  = (CPtr)mem_alloc(FIXED_BLOCK_SIZE_FOR_TABLED_PRED) ;
      Loc = 0 ;
      arity = get_arity(psc) ;
      dbgen_inst_ppvww(tabletrysingle,arity,
		       ((byte *)tp+tabletrysingle_sizeof),tip,tp,&Loc) ;
      dbgen_inst_pvv(allocate_gc,3,3,tp,&Loc) ;
      dbgen_inst_ppv(getVn,2,tp,&Loc) ;
      /* p is *(tp+PRREF_OFFSET_FOR_DYN_TABLED_PRED) below; see remove_prref */
      dbgen_inst_ppvw(calld,3,p,tp,&Loc) ;
      dbgen_inst_pvv(new_answer,arity,2,tp,&Loc) ;
      set_ep(psc, (byte *)tp) ;
    }
  else set_ep(psc, (byte *)p) ;

  return p ;
}

/*======================================================================*/

/* Used by assert & retract to get through the SOBs */

static int Index[20];

static int get_indices(prolog_term prolog_ind)
{
  int num_indx = 0;
DBG_TRACE("get_indices");
  if (isinteger(prolog_ind)) {
    Index[1] = int_val(prolog_ind);
    if (Index[1] != 0) num_indx = 1;
  } else {
    for ( ; !isnil(prolog_ind); prolog_ind = p2p_cdr(prolog_ind)) {
      num_indx++;
      Index[num_indx] = int_val(p2p_car(prolog_ind));
    }
  }
#ifdef SERIOUS_DEBUG
  { int i;
  for (i = 0; i <= num_indx; i++)
    fprintf(stderr, "Index[%d] = %d\n", i, Index[i]);
  fprintf(stderr, "\n");
  }
#endif
  return num_indx;
}

Integer get_clauseno(byte *prref)
{
  Integer lResult = 0;
  ClRef IterSOBbuff;
  PrRef prrefMain = (PrRef)prref;

  IterSOBbuff = prref_first_clref(prrefMain);
  /* get num clauses */
  while(IterSOBbuff)
  {
    if(ClRefType(IterSOBbuff)==SOB_RECORD)
    {
      lResult += (unsigned int)ClRefNumClauses(IterSOBbuff);
    }
    /* next SOB */
    if(IterSOBbuff!=prref_last_clref(prrefMain))
    {
      IterSOBbuff = (ClRef)ClRefNext(IterSOBbuff);
    }
    else
    {
      IterSOBbuff = 0;
    }
  }
  return lResult;
}

byte *jiti_create_htable_asserts(byte *prref, Integer iArg, Integer iArity)
{
  byte * pResult;
  Integer iClauseNo;
  ClRef IterClause ;  
  ClRef IterSOBbuff;
  struct hash_space * pHashSpace;
  PrRef prrefMain;
  byte *pInstrAddr;

  pResult = 0;

  prrefMain = (PrRef)prref;

#ifdef ASSERT_DEBUG
  fprintf(stderr, "############ PRED(0x%x) prref_first_clref (0x%x)  prref_last_clref (0x%x)\n",
                  (unsigned int)prrefMain ,
                  (unsigned int)prref_first_clref(prrefMain),
                  (unsigned int)prref_last_clref(prrefMain) );
#endif

  iClauseNo = get_clauseno((byte *)prrefMain);

  /* create a hash space to build the index */
  pHashSpace = alloc_hspace(iClauseNo);

  IterSOBbuff = prref_first_clref(prrefMain);
  while(IterSOBbuff)
  {
#ifdef ASSERT_DEBUG
    fprintf(stderr,"SOB (0x%x) TYPE SOB [%d] CHAIN FIRST (0x%x) LAST (0x%x) NUMC(%d)\n", 
	    (unsigned int)IterSOBbuff,
	    (int)ClRefType(IterSOBbuff),
	    (unsigned int)ClRefFirstIndex(IterSOBbuff), 
	    (unsigned int)ClRefLastIndex(IterSOBbuff),
	    (int)ClRefNumClauses(IterSOBbuff));
#endif
    if(ClRefType(IterSOBbuff) == SOB_RECORD)
    {
      long bIterated = 0;
      IterClause=(ClRef)ClRefLastIndex(IterSOBbuff);
      do
      {
        byte opCode;
        byte *pClause;
#ifdef JITI_DEBUG
        fprintf(stderr,"CLAUSE 0x%x\n", (unsigned int)IterClause);
#endif
        pInstrAddr = (byte *)IterClause;
        pInstrAddr += inst_sizeof_in_O_file[try];
#ifdef THREADED
        pInstrAddr += sizeof(void *);
#endif

        opCode = get_opcode(pInstrAddr);
        while(opCode == noop)
        {
            pInstrAddr += inst_sizeof_in_O_file[noop] + (inst_operand3(pInstrAddr)*2);
#ifdef THREADED
            pInstrAddr += sizeof(void *);
#endif
            opCode = get_opcode(pInstrAddr);
        }
        pClause = pInstrAddr;

        while(IS_HEADCLAUSE_INST(opCode))
        {
          Integer iRegNum;
          Integer iValue;
#ifdef JITI_DEBUG
          fprint_inst(stderr, pInstrAddr);
#endif
          if(opCode == noop)
          {
            pInstrAddr += inst_sizeof_in_O_file[noop]+(inst_operand3(pInstrAddr)*2);
#ifdef THREADED
            pInstrAddr += sizeof(void *);
#endif
          }
          else
          {
            if((opCode == getnumcon) || (opCode == getcon) || (opCode == getstr)
                || (opCode == getlist) || (opCode == getlist_tvar_tvar) )
            {
              /* get number register of get instruction */
              if(opCode == getlist_tvar_tvar)
              {
                iRegNum = *(pInstrAddr + getlist_tvar_tvar_arg_1);
              }
              else
              {
                iRegNum = *(pInstrAddr + getnumcon_arg_1);
              }
              /* look only at get instructions that matchs with register number */
              if(iRegNum == iArg)
              {
                if((opCode == getlist) || (opCode == getlist_tvar_tvar))
                {
                  iValue = (Cell)(list_str);
                }
                else
                {
                  iValue = *(CPtr)(pInstrAddr + getnumcon_arg_2);
                }
                /* insert value and address in hash table */
                inserth(pHashSpace, (CPtr)pClause, iValue);
#ifdef JITI_DEBUG
		fprintf(stderr,"MATCH GET_INST %p %d %d\n",
			(void *)pClause, iRegNum, iValue); 
#endif
              }
            }
            else
	    {
#ifdef JITI_DEBUG
	      fprintf(stderr,"IS NOT GET_ INSTRUCTION\n");
#endif
            }
            /* next instruction */
            pInstrAddr += inst_sizeof_in_O_file[opCode]; /*size instruction */
#ifdef THREADED
            pInstrAddr += sizeof(void *);
#endif
          }
          opCode = get_opcode(pInstrAddr);
        }
        /* print proceed or fail*/
#ifdef JITI_DEBUG
        fprintf(stderr,"IS NOT HEADCLAUSE INSTRUCTION\n");
        fprint_inst(stderr, pInstrAddr); 
#endif
        if(IterClause == (ClRef)ClRefFirstIndex(IterSOBbuff))
        {
          bIterated = TRUE;
        }
        else
        {
          /* next clause */
          IterClause = ClRefPrev(IterClause);
        }
      }while(!bIterated); 
    }
    /* next SOB */
    if(IterSOBbuff != prref_last_clref(prrefMain))
    {
      IterSOBbuff = (ClRef)ClRefNext(IterSOBbuff);
    }
    else
    {
      IterSOBbuff = 0;
    }
  }

  if(hashing_is_discriminatory(pHashSpace))
  {
    /* generate code to the hash table */
    gen_index(iClauseNo, (void *)(&pResult), iArity, pHashSpace);
#ifdef JITI_DEBUG
    fprint_hashtable(stderr, (byte *)pResult, pHashSpace->size);
#endif
  }
  else
  {
#ifdef JITI_DEBUG
    fprintf(stderr, "Hashing is \"NOT\" discriminatory\n");
#endif
  }

  /* release resources */
  free_hspace(pHashSpace);
  return pResult;
}

bool db_assert_buff_to_clref(Cell Head, int AZ, Cell Indices, int HashTabSize)
{
  Psc   psc;
  byte  arity;
  PrRef Prref;
  ClRef Clause;
  int   Location, *Loc, Inum, num_indx;
DBG_TRACE("db_assert_buff_to_clref");
#ifdef THREADED_TEAM_DEBUG
  fprintf(stderr, "THREADED-TEAM DEBUG: db_assert_buff_to_clref .. \n");
#endif
  psc = term_psc(Head);
  if (get_type(psc) == T_DYNA) { 
    Prref = get_real_prref(psc);
  } else goto weird_error;
  arity = get_arity(psc)+1;  /* +1 to possibly hold the cut address */

#ifdef THREADED_TEAM_DEBUG
  fprintf(stderr, "THREADED-TEAM DEBUG: getting indices.. \n"); 
#endif
  num_indx = get_indices(Indices);

#ifdef THREADED_TEAM_DEBUG
  fprintf(stderr, "THREADED-TEAM DEBUG: creating clref.. \n"); 
#endif

  MakeClRef( Clause,
	     (num_indx > 0) ? INDEXED_CL : UNINDEXED_CL,
	     IC_CELLS(num_indx) + ((Size+0xf)&~0x7)/sizeof(Cell) ) ;

#ifdef THREADED_TEAM_DEBUG
  fprintf(stderr, "THREADED-TEAM DEBUG: preparing clref.. \n"); 
#endif
  Location = 0; Loc = &Location;
  dbgen_inst_ppv(noop,sizeof(Cell)/2,Clause,Loc);    /* will become try */
  write_word(Clause,Loc,0);
  for (Inum = num_indx; Inum > 0; Inum--) {
    /* put template code for chaining buffers from hash tables  */
    int bytes_to_skip = (WORDS_PER_INDEX*Inum-noop_wordsof)*sizeof(Cell)/2;
    dbgen_inst_ppv(noop,bytes_to_skip,Clause,Loc);
#ifdef SERIOUS_DEBUG
    fprintf(stderr, "Generating noop %d ... \n", bytes_to_skip);
#endif
    write_word(Clause,Loc,0);
    dbgen_inst_ppv(noop,sizeof(Cell)/2,Clause,Loc);  /* noop(2) */
    write_word(Clause,Loc,0);
  }

/* Buff is a global variable used to communicate from db_assert_code_to_buff
   to db_assert_buff_to_clref through PROLOG calls */

#ifdef THREADED_TEAM_DEBUG
  fprintf(stderr, "THREADED-TEAM DEBUG: moving buffer.. \n"); 
#endif
  /* fill in clause with code from Buff */
  memmove(((byte *)Clause)+Location,Buff,Size);
  
#ifdef JITI_DEBUG
{
  byte * pAddress=((byte *)Clause)+Location;
  byte * pEndAddress=pAddress+Size;
  while(pAddress<pEndAddress)
  {
    pAddress=fprint_inst(stderr,pAddress);
  } 
}
#endif

  /* What happens down here ? */
#ifdef THREADED_TEAM_DEBUG
  fprintf(stderr, "THREADED-TEAM DEBUG: addbuffing.. \n"); 
#endif
  if (num_indx <= 0) db_addbuff(arity,Clause,Prref,AZ,1);
  else db_addbuff_i(arity,Clause,Prref,AZ,Index,num_indx,Head,HashTabSize);

#ifdef THREADED_TEAM_DEBUG
  fprintf(stderr, "THREADED-TEAM DEBUG: leaving db_assert_buff_to_clref! \n");
#endif
  return TRUE;

 weird_error:
  xxx_abort("%s: Predicate %s/%d is not dynamic",
	    __FUNCTION__, get_name(psc), get_arity(psc));
  return FALSE;
}

/*======================================================================*/

/** The following macros traverse the SOB chains/trees
 ** and pick the first (next) clause since a given sob
 **/

/* Check if a clause with head H is in the hash table of a SOB */
/* The indexing Level is used to adjust the returned clause    */
/* pointer to the beginning of the clause		       */

#define CheckSOBClause(H, Ind, SOB, Level)			\
{    /* DBG_TRACE ("CheckSOBClause"); */                        \
     int h, t ;							\
     ClRef cl ;				    			\
     t = ClRefHashSize(SOB); 					\
     h = hash_val( (Ind), (H), t ) ;				\
     cl = (ClRef) ClRefHashTable(SOB)[h] ;			\
     if ((byte *)cl != (byte *)&fail_inst)			\
	return IndPtrClRef(cl,Level) ;				\
}

/* This macro finds the next SOB to search in the indexing tree */
/* the ordering is down, left, (up+)left, the down and left	*/
/* parts being performed by the next macro.			*/
/* It returns 0 if the root (prref) is ever reached.		*/
/* Compiled clauses may be intermixed with dynamic ones, so 	*/
/* that possibility must be checked.				*/

#define NextSOB(s,cur_level)					\
{ DBG_TRACE ("NextSOB");    \
  while (ClRefTryOpCode(s) == dyntrustmeelsefail ||		\
	   ClRefTryOpCode(s) == noop) /* end of sob chain */	\
  {\
       /*  TOREMOVE */\
    /*fprint_inst(stderr,s);*/\
    if (cur_level-- == 1)   /* root of sob tree */		\
    {\
       /*  TOREMOVE */\
      /* xxx_dbgmsg("NextSOB(1) :: index %d Next SOB %p",cur_level,0); */ \
      return 0 ;					\
    }\
    else \
    {\
      s = ClRefUpSOB(s); /* go up */			\
    }\
       /*  TOREMOVE */\
      /* fprint_inst(stderr,s); */ \
  } \
  s = (ClRef)ClRefNext(s) ;   /* follow sob chain */		\
  if (ClRefType(s) != SOB_RECORD)\
  {\
       /*  TOREMOVE */\
      /*xxx_dbgmsg("NextSOB(2) :: index %d Next SOB %p",cur_level,0);*/ \
      return s;			\
  }\
}

/* s points to first SOB of the index chain			*/
/* look for Head/Ind in all SOB chains for this index level	*/
/* if needed go up to look in next sob chain(s)			*/

#define FirstClauseSOB(s,i,l,Head,Ind)				\
{ /* DBG_TRACE("FirstClauseSOB"); */                            \
  for (;;)							\
	if (i < l) /* sob node */				\
	{   PrRef pr = ClRefPrRef(s);                           \
            s = prref_first_clref(pr); /* go down */    	\
	    i++ ;						\
	}							\
	else /* i == l -> sob leaf */				\
	{   CheckSOBClause(Head,Ind,s,i) ;			\
	    NextSOB(s,i) ;					\
	}							\
}

static ClRef first_clref(PrRef Pred, Cell Head, int *ILevel, int *Index)
{
  ClRef s ;	/* working SOB */
  int i, l ;    /* index depth */
  int Ind = 0;  /* indexing sob instruction argument */

DBG_TRACE("first_clref");
  if (prref_inst_opcode(Pred) == fail)
    return 0 ;

  /* first findout what index shall we use */
  *Index = *ILevel = 0 ;
  for (s = prref_first_clref(Pred), i = 1;
       ClRefType(s) == SOB_RECORD;
       Pred = ClRefPrRef(s), s = prref_first_clref(Pred), i++)
    {
      Ind = ((ClRefSOBArg(s,1) << 8) | ClRefSOBArg(s,2) ) << 8 |
	ClRefSOBArg(s,3) ;
      if( can_hash(Ind,Head) )
	{
	  *Index = Ind ; *ILevel = i ;
	  /* This is the index we're going to use */
	  break ;
	}
    }
  if (*ILevel == 0 )	/* It's not indexable, so s points to first clause */
    return s ;	/* in all chain of first SOB at lowest level */
  else
    {
      i = l = *ILevel ;
      FirstClauseSOB(s,i,l,Head,Ind) ;
    }
}

static ClRef next_clref(ClRef Clause, Cell Head, int IndexLevel, int Ind)
{
DBG_TRACE("next_clref");
  if (ClRefType(Clause) != INDEXED_CL) 
  {	/* mixed clause types */
DBG_TRACE("next_clref:NO indexed_cl");
    if (ClRefTryOpCode(Clause) == dyntrustmeelsefail ||
          ClRefTryOpCode(Clause) == noop)
      return 0 ;
    else if (ClRefType(((ClRef)ClRefNext(Clause))) != SOB_RECORD)
      return (ClRef)ClRefNext(Clause) ;
    else /* should do as in cl_ref_first -- to index */
    {
      ClRef s = (ClRef)ClRefNext(Clause) ;
      if (IndexLevel == 0) /* goto first cl in all chain */
      {
        while (ClRefType(s) == SOB_RECORD) {
          PrRef pr = ClRefPrRef(s) ;
          s = prref_first_clref(pr) ;
        }
        return s ;
      }
      else
      {
        int i;
        for (i = 1 ; i < IndexLevel ; i++) {
          PrRef pr = ClRefPrRef(s) ;
          s = prref_first_clref(pr) ; /* all the way down */
        }
        CheckSOBClause(Head,Ind,s,IndexLevel) ;
        NextSOB(s,i) ;
        FirstClauseSOB(s,i,IndexLevel,Head,Ind) ;
      }
    }
  }
  else
  {
DBG_TRACE("next_clref:indexed_cl");

    if (IndexLevel == 0) 
    { /* look in all chain */
      if (ClRefTryOpCode(Clause) == trymeelse || /* mid chain */
           ClRefTryOpCode(Clause) == retrymeelse ) 
           return (ClRef)ClRefNext(Clause) ;
      else /* INDEXED_CL, look on next SOB */
      {
        int i, ni;	/* number of indexes */
        ClRef s;
        PrRef pr;
        
        ni = i = ClRefNumInds(Clause);
        /* all chain is on lowest index chain */
        s = (ClRef)ClRefNext(Clause);    /* s = current SOB */
        NextSOB(s,i);
        /* all leaf SOBs have non empty all chains */
        while (i++ < ni) {
          pr = ClRefPrRef(s) ;
          s = prref_first_clref(pr) ;
        }
        pr = ClRefPrRef(s) ;
        return prref_first_clref(pr) ;
      }
    }
    else	/* look in appropriate hash chain */
    {
      CPtr PI = ClRefIndPtr(Clause,IndexLevel) ;
      if (get_opcode(PI) == trymeelse || /* mid chain */
          get_opcode(PI) == retrymeelse) 
        return IndPtrClRef(IndRefNext(PI),IndexLevel) ;
      else /* end of chain */
      {
        int i = IndexLevel ;
        ClRef s = (ClRef)IndRefNext(PI) ; /* s = current SOB */
        NextSOB(s,i) ;
        FirstClauseSOB(s,i,IndexLevel,Head,Ind) ;
      }
    }
  }
  DBG_TRACE("RETURN :next_clref");
}

/*
#define RETRACT_DEBUG
#define RETRACT_GC_DEBUG
*/
/* Generic macro that deletes an element from a chain		*
 * Made possible because of the design of all chains containing	*
 * the three words:						*
 * 		(-1)Prev - (0)TryInstr - (1)Next		*
 * Index Rec Macros are used, although any kind of chain can be *
 * handled							*
 * Args are PC    - pointer to Chain element			*
 *          Displ - value to use as noop arg for begging of BC  *
 * return position of element just deleted		        */

#define delete_from_chain(c, PC, Displ)                         \
{   switch( c )                                                 \
    { case noop: /* uniq */                                     \
        break ;                                                 \
      case trymeelse: /* first */                               \
        IndRefPrev(IndRefNext(PC)) = IndRefPrev(PC) ;           \
        if (get_opcode(IndRefNext(PC)) == retrymeelse) {        \
	  set_opcode(IndRefNext(PC),trymeelse) ;                \
        }\
        else /* dyntrustme */                                   \
        { set_opcode(IndRefNext(PC),noop) ;                     \
          inst_operand3(IndRefNext(PC)) = (Displ) ;             \
        }                                                       \
        break ;                                                 \
      case retrymeelse: /* mid */                               \
        IndRefPrev(IndRefNext(PC)) = IndRefPrev(PC) ;           \
        IndRefNext(IndRefPrev(PC)) = IndRefNext(PC) ;           \
        break ;                                                 \
      case dyntrustmeelsefail: /* last */                       \
        IndRefNext(IndRefPrev(PC)) = IndRefNext(PC) ;           \
        if (get_opcode(IndRefPrev(PC)) == retrymeelse) {        \
          set_opcode(IndRefPrev(PC),dyntrustmeelsefail) ;       \
        }\
        else /* trymeelse */                                    \
        { set_opcode(IndRefPrev(PC),noop) ;                     \
          inst_operand3(IndRefPrev(PC)) = (Displ) ;             \
        }                                                       \
        break ;                                                 \
      default:                                                  \
        xxx_exit("%s: unhandled instruction: %s", __FUNCTION__,inst_name(c)); \
        break ;                                                 \
    }                                                           \
}

/* delete from a hash chain */

static void delete_from_hashchain(ClRef Clause, int Ind, int num_indx)
{  
  CPtr PI = ClRefIndPtr(Clause,Ind) ;
  byte c = get_opcode(PI) ;

  delete_from_chain(c,PI,(WORDS_PER_INDEX*(num_indx-Ind)+1)*sizeof(Cell)/2) ;

  if (get_opcode(PI) == noop)
    *IndRefPrev(PI) = (Cell) &fail_inst ;
  else
    if (get_opcode(PI) == trymeelse)
      *IndRefPrev(PI) = (Cell) IndRefNext(PI) ;
}

/* delete from the chain pointed by a prref - a all chain or a sob chain */
/* Works for SOBs and all sorts of clauses */

static void delete_from_allchain(ClRef Clause)
{  
  PrRef prref ;
  byte c = ClRefTryOpCode(Clause) ;

  delete_from_chain( c, (CPtr)Clause, sizeof(Cell)/2 ) ;

  switch( c )
    {
    case noop:
      prref = ClRefPrRef(ClRefPrev(Clause)) ;
      set_prref_inst(prref,fail) ;
      prref_first_clref(prref) = prref_last_clref(prref) = (ClRef)prref ;
      break ;
    case trymeelse:
      prref = ClRefPrRef(ClRefPrev(Clause)) ;
      prref_first_clref(prref) = (ClRef)ClRefNext(Clause) ;
      break ;
    case dyntrustmeelsefail:
      prref = ClRefPrRef(((ClRef)ClRefNext(Clause))) ;
      prref_last_clref(prref) = ClRefPrev(Clause) ;
      break ;
    }
}

static void delete_from_sobchain(ClRef Clause)
{  
  PrRef Pred ;
  byte c = ClRefTryOpCode(Clause) ;

  delete_from_chain( c, (CPtr)Clause, sizeof(Cell)/2 ) ;

  switch( c )
    {
    case noop:
      Pred = (PrRef)ClRefPrev(Clause) ;
      set_prref_inst(Pred,fail) ;
      prref_first_clref(Pred) = prref_last_clref(Pred) = (ClRef) Pred ;
      break ;
    case trymeelse:
      Pred = (PrRef)ClRefPrev(Clause) ;
      prref_first_clref(Pred) = (ClRef)ClRefNext(Clause) ;
      break ;
    case dyntrustmeelsefail:
      Pred = (PrRef)ClRefNext(Clause) ;
      prref_last_clref(Pred) = ClRefPrev(Clause) ;
      break ;
    }
}

/* circular buffer for retracted clauses */
#define MAX_RETRACTED_CLAUSES   25

static ClRef retracted_buffer[MAX_RETRACTED_CLAUSES+1];
static ClRef *OldestCl = retracted_buffer;
static ClRef *NewestCl = retracted_buffer;

#define next_in_buffer(pCl) ((pCl)>=(retracted_buffer+MAX_RETRACTED_CLAUSES)\
                                ? retracted_buffer : (pCl) + 1)

#define retract_buffer_empty()  (NewestCl == OldestCl)
#define retract_buffer_full()   (OldestCl == next_in_buffer(NewestCl))

#define insert_in_buffer(Cl) (*NewestCl = (Cl),\
                               NewestCl = next_in_buffer(NewestCl))
#define remove_from_buffer() (OldestCl = next_in_buffer(OldestCl))


static int really_delete_clause(ClRef Clause)
{
#ifdef RETRACT_DEBUG
  xxx_dbgmsg("Really deleting clause(%p) op(%x) type(%d)",
	     Clause, ClRefTryOpCode(Clause), ClRefType(Clause) ) ;
#endif
  switch( ClRefType(Clause) )
    {
    case UNINDEXED_CL:
      delete_from_sobchain(Clause) ;
      break ;
    case INDEXED_CL:
      {
	int i, num_indx ;
	ClRef sob ;

	num_indx = ClRefNumInds(Clause) ;
#ifdef RETRACT_DEBUG
	xxx_dbgmsg("Really deleting clause (%p) size %d indexes %d",
		   Clause, ClRefSize(Clause), num_indx) ;
#endif
	delete_from_allchain(Clause) ;
	
	/* remove it from index chains */
	for (i = num_indx; i >= 1; i--)
	  {
	    CPtr IP = ClRefIndPtr(Clause, i) ;
	    while ( get_opcode(IP) != dyntrustmeelsefail &&
		    get_opcode(IP) != noop )
	      IP = IndRefNext(IP) ;
	    /* last pointer in index chain points to indexing SOB */
	    sob = (ClRef)IndRefNext(IP) ;
#ifdef RETRACT_DEBUG
	    xxx_dbgmsg("SOB(%d) - hash size %d - %d clauses",
		       i, ClRefHashSize(sob), ClRefNumClauses(sob) ) ;
	    xxx_dbgmsg("Addr %p : prev %p : next %p",
		       sob, ClRefNext(sob), ClRefPrev(sob) ) ;
#endif
	    delete_from_hashchain(Clause,i,num_indx) ;
	    if (--ClRefNumClauses(sob) == 0)
	      {
#ifdef RETRACT_DEBUG
		xxx_dbgmsg("deleting sob - %p", sob);
#endif
		delete_from_sobchain(sob) ;
		/* mem_dealloc((byte *)ClRefAddr(sob), ClRefSize(sob)); */
	      }
	  }
	break ;
      }
    case SOB_RECORD:
    default :
      xxx_exit( "retract internal error!" ) ;
    }
  mem_dealloc((byte *)ClRefAddr(Clause), ClRefSize(Clause));
  return TRUE ;
}

/*
** Insert in retract buffer and remove old clauses
*/
inline static void delete_clause(ClRef Clause)
{
  if ( retract_buffer_full() )
    {
      ClRef ClauseToDelete = *OldestCl;
      remove_from_buffer();
      really_delete_clause(ClauseToDelete);
    }
  insert_in_buffer(Clause);
}

inline static void force_retract_buffers(void)
{
  while (!retract_buffer_empty()) {
    really_delete_clause(*OldestCl);
    remove_from_buffer();
  }
}

bool db_retract_clause(ClRef Clause, int reclaim_space)
{
#ifdef RETRACT_DEBUG
  xxx_dbgmsg("Retract clause(%p) op(%x) type(%d)",
	     Clause, ClRefTryOpCode(Clause), ClRefType(Clause) ) ;
#endif
  switch( ClRefType(Clause) )
    {
    case UNINDEXED_CL:
      /* set fail for retract_nr AND protection */
      if (get_opcode(ClRefEntryPoint(Clause)) == fail &&
	  inst_operand1(ClRefEntryPoint(Clause)) == 66) 
	reclaim_space = FALSE;  /* previously scheduled for deletion */
      else {
	set_opcode(ClRefEntryPoint(Clause),fail) ;
	inst_operand1(ClRefEntryPoint(Clause)) = 66;
      }
      break ;
    case INDEXED_CL:
      if (get_opcode(ClRefIEntryPoint(Clause,ClRefNumInds(Clause))) == fail &&
	  inst_operand1(ClRefIEntryPoint(Clause,ClRefNumInds(Clause))) == 66)
	reclaim_space = FALSE;  /* previously scheduled for deletion */
      else {
	set_opcode(ClRefIEntryPoint(Clause,ClRefNumInds(Clause)),fail) ;
	inst_operand1(ClRefIEntryPoint(Clause,ClRefNumInds(Clause))) = 66;
      }
      break ;
    case SOB_RECORD:
      xxx_exit( "retracting indexing record!" ) ;
      break ;
    default :
      xxx_exit( "retract internal error!" ) ;
      break ;
    }
  if (reclaim_space) {
#ifdef RETRACT_DEBUG
    xxx_dbgmsg("Inserting clause in delete buffer(%p) op(%x) type(%d)",
	       Clause, ClRefTryOpCode(Clause), ClRefType(Clause) ) ;
#endif
    delete_clause(Clause) ;
  }
  return TRUE ;
}

/* --------------------------------------------------------------------	*/
/*
 * db_get_clause(+CC, ?CI, ?CIL, +Head, +Failed, -Clause, -EntryPoint)
 *
 * gets next clause from predicate
 * Arg 1 is the previous ClRef, or 0 if this is the first call.
 * Args 2 and 3 must initially be variables; they get set by the first
 *   call and those values should continue to be passed for each 
 *   subsequent call.
 * Arg 4 is a prolog term that matches the head of the clause
 * Arg 5 is 0 for "normal" clauses, 1 for clauses that consist of a fail 
 * instruction (generated by retract_nr and to be passed to reclaim_space)
 * Arg 6 returns the clause address
 * Arg 7 returns the jump point into the code
 */

bool db_get_clause(ClRef clause)
{
  Cell  head = ptoc_tag(4);
  CPtr  EntryPoint;
  Psc   psc = term_psc(head);
  PrRef prref = get_real_prref(psc);

#ifdef JITI_DEBUG
  DBG_TRACE(__FUNCTION__);
  fprintf(stderr, "@@@@ CLAUSE 0x%x\n", (unsigned int)clause);
#endif

#ifdef RETRACT_GC_DEBUG
  xxx_dbgmsg("GET CLAUSE P-%p(%x) C-%p(%x) F-%p L-%p",
	     prref, *(byte *)prref, clause,
	     clause ? *(byte *)(clause) : 0,
	     prref_first_clref(prref), prref_last_clref(prref) ) ;
#endif

  if (prref_last_clref(prref) == (ClRef)prref)
    {/* set ClRef and EntryPoint to 0 and get out */
      ctop_int( 6, 0 ) ;
      ctop_int( 7, 0 ) ;
      return TRUE ;
    }
  
  if (clause == 0)
    { int IndexLevel = 0, IndexArg = 1;
      clause = first_clref(prref, head, &IndexLevel, &IndexArg) ;
      ctop_int(2,IndexLevel);
      ctop_int(3,IndexArg);
    }
  else
    { int IndexLevel = ptoc_int(2);
      int IndexArg   = ptoc_int(3);
      int failed     = ptoc_int(5);
      do
      { 
        /* loop until a clause is found:
	      Retracted if looking for failed; 
	      Not Retracted if looking for not failed */
	     clause = next_clref(clause, head, IndexLevel, IndexArg);
      }
      while (clause && ClRefNotRetracted(clause) == failed) ;
    }

  if (clause !=0) {
    if ( ClRefType(clause) == SOB_RECORD )
      { xxx_exit("Error in get clause"); return FALSE; }
    else if ( ClRefType(clause) != INDEXED_CL )
           { EntryPoint = ClRefEntryPoint(clause) ;}
         else /* ClRefType(Clause) == INDEXED_CL */
	   { int num_indx = ClRefNumInds(clause) ;
#ifdef THREADED_TEAM_DEBUG
	     fprintf(stderr, "In db_get_clause: ni = %d\n", num_indx);
#endif
	     EntryPoint = ClRefIEntryPoint(clause,num_indx) ;
	   }
  }
  else EntryPoint = 0 ;

#ifdef RETRACT_GC_DEBUG
  xxx_dbgmsg("GOT CLAUSE C-%p(%x) EntryPoint 0x%x", clause, clause ? *(byte *)clause : 0 ,(unsigned int) EntryPoint) ;
#endif

  ctop_int(6, (Integer)clause) ;
  ctop_int(7, (Integer)EntryPoint) ;

#ifdef JITI_DEBUG
  DBG_TRACE("END:dg_get_clause");
#endif


  return TRUE ;
}

/* -------------------------------------------------------------------- */

bool db_convert_to_dynamic(Cell head, bool is_tabled)
{
  Psc psc = term_psc(head);
DBG_TRACE("db_convert_to_dynamic");
  if (psc) {
    switch (get_type(psc)) {
    case T_ORDI:
    case T_UDEF:
      build_and_return_prref(psc, is_tabled);
      return TRUE;
    case T_DYNA:
      return TRUE;
    case T_PRED:
      xxx_abort("Not permitted to assert to static predicate %s/%d",
		get_name(psc), get_arity(psc));
      return FALSE;
    default:
      xxx_abort("Attempt to assert to something illegal (Type = %d)",
		get_type(psc));
      return FALSE;
    }
  }
  else {
    xxx_err_handle(TYPE_ERROR, 1, "dynamic", 1, "predicate (specification)", head);
    return FALSE;
  }
}

/*----------------------------------------------------------------------*/
/* some stuff for trie_assert                                           */
/*----------------------------------------------------------------------*/

#define try_type_instr_fld(x)  (ClRefTryOpCode(x))
#define code_to_run(x)         (ClRefEntryPoint(x))
#define TRASSERT_ROOT_OFFSET   ENTRY_POINT_OFFSET+(jump_arg_1/sizeof(word))
#define trie_assert_root(x)    (ClRefWord(x,TRASSERT_ROOT_OFFSET))

/*----------------------------------------------------------------------*/

inline static int clref_trie_asserted(CPtr clref)
{
  return((get_opcode(code_to_run(clref)) == jump) && 
	 (get_opcode(trie_assert_root(clref)) == trie_assert_inst));
}

/*----------------------------------------------------------------------*/

static int another_buff(CPtr instr)
{
  int op = get_opcode(instr) ;
  return op != noop && op != dyntrustmeelsefail && op != fail ;
}

/*======================================================================*/
/* The following routine deletes all clauses from a prref.  It is the	*/
/* equivalent of retractall(p(_,_,_,..,_). It is given the address of	*/
/* a buffer and frees it and all buffers it points to.			*/
/*======================================================================*/

static void gen_retract_all(ClRef buff)
{
  ClRef buffers_to_free[200];
  int btop = 0;
  ClRef buffer ;

  force_retract_buffers();
  buffers_to_free[btop++] = buff;
  while (btop > 0) {
    buffer = buffers_to_free[--btop];
    switch (ClRefType(buffer)) {
    case SOB_RECORD: 
      if (another_buff(&ClRefJumpInstr(buffer)))
	buffers_to_free[btop++] = (ClRef) ClRefFirstIndex(buffer);
      if (another_buff(&ClRefTryInstr(buffer)))
	buffers_to_free[btop++] = (ClRef)ClRefNext(buffer);
      mem_dealloc((byte *)ClRefAddr(buffer), ClRefSize(buffer));
      break ;
    case UNINDEXED_CL: 
    case INDEXED_CL:
      if (another_buff(&ClRefTryInstr(buffer)))
	buffers_to_free[btop++] = (ClRef)ClRefNext(buffer);
      if (ClRefNotRetracted(buffer))
	/* db_retract_clause(buffer,1) */
	/* really_delete_clause(buffer); */
	mem_dealloc((byte *)ClRefAddr(buffer), ClRefSize(buffer));
      break;
    }
  }
}

/*======================================================================*/

/* Check this for THREADED */
bool db_retract_abolish(Cell term, bool abolish) /* ? */
{
  Psc psc = term_psc(term);

#ifdef THREADED_TEAM_DEBUG
  fprintf(stderr, "THREADED-TEAM DEBUG: db_retract_abolish: ");
  printterm(term,1,10); /**/
  fprintf(stderr, "\n");
#endif

  if (get_type(psc) == T_DYNA)
    {
      byte *Prref = get_ep(psc);
      if (get_opcode(Prref) == tabletrysingle) {
	if (abolish) {
	  /* free real prref which is obtained from calld instruction */
	  mem_dealloc((byte *)(get_real_prref(psc)), sizeof(PrRefData)); 
	  /* free table header */
	  mem_dealloc(Prref, FIXED_BLOCK_SIZE_FOR_TABLED_PRED);
	  set_type(psc, T_ORDI);
	  set_ep(psc, (byte *)(&(psc->load_inst)));
	}
	return TRUE;
      }
      if (get_opcode(Prref) != fail) {
	ClRef buff = prref_first_clref(Prref);
	gen_retract_all(buff);
	set_opcode(Prref,fail);
      }
      if (abolish) {
	mem_dealloc(Prref, sizeof(PrRefData));
	set_type(psc, T_ORDI);
	set_ep(psc, (byte *)(&(psc->load_inst)));
      }
      return TRUE;
    }
  else if (get_type(psc) == T_ORDI) return TRUE;
  else return TRUE; /* error handling goes here */
}

/*----------------------------------------------------------------------*/

inline static ClRef trie_asserted_clref(PrRef prref)
{
  ClRef clref;

  clref = prref_last_clref(prref);
  if (try_type_instr_fld(prref) != fail) {
    if ((get_opcode(code_to_run(clref)) == jump) &&
	(get_opcode(trie_assert_root(clref)) == trie_assert_inst))
      return clref;
  }
  return NULL;
}

/*----------------------------------------------------------------------*/

bool is_trie_asserted(ClRef clref)
{
  if (ClRefType(clref) == TRASSERTED_CL) {
    if (get_opcode(trie_assert_root(clref)) == trie_assert_inst) return TRUE;
  }
  return FALSE;
}

/*----------------------------------------------------------------------*/

bool trie_assert(Cell clause, bool is_tabled)
{
  Psc    psc;
  PrRef  prref;
  ClRef  clref;
  BTNptr inst_node_ptr;
  int    found = 1;

  psc = term_psc(clause);
  switch (get_type(psc)) {
  case T_DYNA:
    prref = get_real_prref(psc);
    break;
  case T_ORDI:
  case T_UDEF:
    prref = build_and_return_prref(psc, is_tabled);
    break;
  default:
    xxx_abort("TRIE_ASSERT: Something weird in asserting %s/%d",
	      get_name(psc), get_arity(psc));
    return FALSE;
  }

  clref = trie_asserted_clref(prref);

  switch_to_trie_assert;
  
  if (clref == NULL) {
    /*
     * We allocate a block that sort of resembles a ClRef0: In its
     * beginning it contains a clref header block (which is bypassed)
     * and afterwards contains:
     *
     *   - A try-type instruction (whose argument label contains the
     *     chain address) --- these fields are set in add_db_buff.
     *   - The beginning of the trie-code containing a jump to the
     *     trie node which is allocated.
     */
    int alloc_sizeof = sizeof(ClRefHdr) + (try_sizeof+jump_sizeof);
    
    clref = (ClRef)mem_alloc(alloc_sizeof);
    *((CPtr)clref) = alloc_sizeof+TRASSERTED_CL; /* store size, encode type */
    /* from now on, bypass the clref header block */
    clref = (ClRef)((byte *)clref + sizeof(ClRefHdr));

    inst_node_ptr = newBasicTrie(psc,ASSERT_TRIE_TT);
    set_opcode(inst_node_ptr,trie_assert_inst);

    set_opcode((CPtr)clref+ENTRY_POINT_OFFSET,jump);
    trie_assert_root(clref) = (Cell)inst_node_ptr; /* the label of the jump instr */

    db_addbuff((get_arity(psc)+1),clref,prref,1,666/*value not important*/);
  }
  else
    inst_node_ptr = (BTNptr)trie_assert_root(clref);

  one_term_chk_ins(clause,inst_node_ptr,&found);

  switch_from_trie_assert;

  ctop_int(3,found);
  return TRUE;
}

/*----------------------------------------------------------------------*/

bool trie_retract(CPtr clref)
{
  if (clref == NULL) {
    Last_Nod_Sav = NULL;
    return TRUE;
  }
  else if (Last_Nod_Sav == NULL) {
    xxx_abort("TRIE_RETRACT: Last_Nod_Sav is NULL");
    return FALSE;
  }
  else {
    BTNptr inst_node_ptr = (BTNptr)trie_assert_root(clref);
    switch_to_trie_assert;
    delete_branch(Last_Nod_Sav, &(Child(inst_node_ptr)));
    switch_from_trie_assert;
    return TRUE;
  }
}

/*----------------------------------------------------------------------*/

/* Only mark the nodes in the branch as deleted. */

bool trie_retract_safe(void)
{ 
  if (Last_Nod_Sav == NULL)
    return FALSE;
  else {
    safe_delete_branch(Last_Nod_Sav);
    return TRUE;
  }
}

/*----------------------------------------------------------------------*/
