/* File:      cell_xsb.h
** Author(s): David S. Warren, Jiyang Xu, Terrance Swift, Kostis Sagonas
** Contact:   xsb-contact@cs.sunysb.edu
** 
** Copyright (C) The Research Foundation of SUNY, 1993-1999
** 
** XSB is free software; you can redistribute it and/or modify it under the
** terms of the GNU Library General Public License as published by the Free
** Software Foundation; either version 2 of the License, or (at your option)
** any later version.
** 
** XSB is distributed in the hope that it will be useful, but WITHOUT ANY
** WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
** FOR A PARTICULAR PURPOSE.  See the GNU Library General Public License for
** more details.
** 
** You should have received a copy of the GNU Library General Public License
** along with XSB; if not, write to the Free Software Foundation,
** Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
**
** $Id: cell_xsb.h,v 1.10 2004/06/09 15:05:10 kostis Exp $
** 
*/

#ifndef __CELL_XXX_H__
#define __CELL_XXX_H__

#include "xxx_config.h"

/*======================================================================*/
/* CELL: an element in the local stack or global stack (heap).		*/
/*									*/
/* Corresponds to the basic word type of the physical machine.		*/
/* This module also defines the basic INTEGER and REAL types to use	*/
/* by the Prolog interpreter						*/
/*									*/
/* Interface routines							*/
/*	They are put in the different files indicated below, according  */
/*	to the tagging scheme you used.					*/
/* The functions (macros, indeed) include:				*/
/*	cell_tag(cell):		give any cell, return its tag		*/
/*	isnonvar(dcell):	check if the cell is a ref or a var.	*/
/*				when used with derefed cell, check if	*/
/*				the cell is instanciated 		*/
/*	int_val(dcell):		assume derefed integer cell, return its */
/*				value in machine format			*/
/*	clref_val(dcell):	assume derefed struct or list cell,     */
/*				return a CPtr to the address it points  */
/*                              to.	                                */
/*	struct_val(dcell):	assume derefed struct cell, return Pair	*/
/*	bld_int(addr, val):	build a new integer cell		*/
/*	bld_float(addr, val):	build a new float cell			*/
/*	bld_ref(addr, val):	build a new reference cell		*/
/*	bld_struct(addr, str):	build a new srtuct cell			*/
/*	bld_atom(addr, atm):	build a new atom cell			*/
/*	bld_list(addr, list):	build a new list cell			*/
/*	bld_functor(addr, psc):	build a functor cell for a new structure*/
/*	bld_free(addr):		build a new free variable 		*/
/*	bld_copy(dest, source): build a copy of the given cell.		*/
/*			if the source is a free var, the copy is indeed */
/*			a ref cell. Need special checking when free var */
/*			is not a pointer to itself but a special tag.	*/
/*======================================================================*/

/* ==== types of cells =================================================*/

#include "celltags_xsb.h"

/*======================================================================*/
/* CELL: an element in the local stack or global stack (heap).		*/
/*									*/
/* CELL and PROLOG_TERM are defined identically.			*/
/* However, CELL is used to refer to elements of WAM stacks, while	*/
/* PROLOG_TERM is used in the C interface to point to a cell containing	*/
/* the outer functor of a prolog term.					*/
/*======================================================================*/

typedef unsigned long Cell;

typedef Cell *CPtr;

#ifdef THREADED
/* Helps with fail_inst and similar stuff */
typedef struct { 
  CPtr addr;
  Cell inst;
} WAM_inst;
#else
typedef Cell WAM_inst;
#endif

#ifdef BITS64
typedef long Integer;
#else
typedef int  Integer;
#endif
 
#ifdef BITS64
typedef double Float;
#else
typedef float  Float;
#endif

#define cell(cptr) *(cptr)
#define follow(cell) (*(CPtr)(cell))

extern Float asfloat(Cell);
extern Cell  makefloat(Float);
extern Float getfloatval(Cell);

#define isref(cell)  (!((word)(cell)&0x3))
#define isnonvar(cell) ((word)(cell)&0x3)		/* dcell -> bool */

#define cell_tag(cell) ((word)(cell)&0x7)

/*======================================================================*/
/*======================================================================*/

#ifdef BITS64
/* 64 bit machines, take bits 0-1, 61-63 */
#define MALLOC_BIT_MASK	0x8000000000000007
/* Encoding of integers */
#define enc_int(val) ( ((Integer)(val) & 0xe000000000000003) ?\
	(((Integer)(val) << 5) | 0x10) :\
	((Integer)(val) << 3) )
#define dec_int(dcell) ( ((Integer)(dcell) & 0x10) ?\
	((Integer)(dcell) >> 5) :\
	(((Integer)(dcell) >> 3) & 0x1ffffffffffffffc) )
/* encoding of addresses -- no need to steal bits */
#define enc_addr(addr) ((Cell)(addr) << 2)
#define dec_addr(dcell) (((Cell)(dcell) >> 2) & 0x3ffffffffffffffc)

/*-----------------------------------------------------------------*/
/* for HP machines (HP700 & 9000 series) take bits 0-1, 28, 29, 31 */
/*-----------------------------------------------------------------*/
#elif defined(HP700)
#define enc_int(val) ( ((Integer)(val) & 0xb0000003) ?\
			(((Integer)(val) << 5) | 0x10) :\
			(((Integer)(val) << 1) & 0x80000000) |\
			(((Integer)(val) << 3) & 0x7fffffe0) )
#define dec_int(dcell) ( ((Integer)(dcell) & 0x10) ?\
                        ((Integer)(dcell) >> 5) :\
			(((Integer)(dcell) >> 1) & 0x40000000) |\
			(((Integer)(dcell) >> 3) & 0x0ffffffc) )

/* Fewer bit representation of pointers (take bits 0-1, 28-29 only) */
#define enc_addr(addr)  (((Cell)(addr) & 0xc0000000) |\
			 (((Cell)(addr) << 2) & 0x3ffffff0))
#define dec_addr(dcell) (((Cell)(dcell) & 0xc0000000) |\
			 (((Cell)(dcell) >> 2) & 0x0ffffffc))

/*----------------------------------------------------------*/
/* for IBM machines (like RS-6000) take bits 0-1, 27, 30-31 */
/*----------------------------------------------------------*/
#elif defined(IBM)
#define enc_int(val) ( ((Integer)(val) & 0xc8000003) ?\
			(((Integer)(val) << 5) | 0x10) :\
			(((Integer)(val) << 2) & 0xc0000000) |\
			(((Integer)(val) << 3) & 0x3fffffe0) )
#define dec_int(dcell) ( ((Integer)(dcell) & 0x10) ?\
			((Integer)(dcell) >> 5) :\
			(((Integer)(dcell) >> 2) & 0x30000000) |\
			(((Integer)(dcell) >> 3) & 0x07fffffc) )

#define enc_addr(addr)  ((Cell)(addr) << 2)
#define dec_addr(dcell) (((Cell)(dcell) >> 2) & 0x3ffffffc)

/*-------------------------------*/
/* MIPS and Solaris x86 machines */
/*-------------------------------*/
#elif (defined(MIPS_BASED) || defined(SOLARIS_x86)) 
#ifdef BITS64
/* 64 bit machines, take bits 0-1, 61-63 */
/* Encoded integers/addresses */
#define enc_int(val) ( ((Integer)(val) & 0xe000000000000003) ?\
	(((Integer)(val) << 5) | 0x10) :\
	((Integer)(val) << 3) )
#define dec_int(dcell) ( ((Integer)(dcell) & 0x10) ?\
	((Integer)(dcell) >> 5) :\
	(((Integer)(dcell) >> 3) & 0x1ffffffffffffffc) )

#define enc_addr(addr) ((Cell)(addr) << 2)
#define dec_addr(dcell) (((Cell)(dcell) >> 2) & 0x3ffffffffffffffc)
#else
/* take bits 0-1, 29-31 */
/* Encoded integers/addresses */
#define enc_int(val) ( ((Integer)(val) & 0xe0000003) ?\
	(((Integer)(val) << 5) | 0x10) :\
	((Integer)(val) << 3) )
#define dec_int(dcell) ( ((Integer)(dcell) & 0x10) ?\
	((Integer)(dcell) >> 5) :\
	(((Integer)(dcell) >> 3) & 0x1ffffffc) )

#define enc_addr(addr)  ((Cell)(addr) << 2)
#define dec_addr(dcell) (((Cell)(dcell) >> 2) & 0x3ffffffc)
#endif

/*----------------------------------------------------------------------*/
/* This is the case most often used: take bits 0-1, and either 30 or 31 */
/* BIG_MEM allows machines to use 1G of memory.                         */
/*----------------------------------------------------------------------*/
#elif (defined(BIG_MEM) || defined(WIN_NT) || defined(LINUX_ELF))

#if (HIGH_BIT_TO_STEAL == 31)

#define MALLOC_BIT_MASK	0x80000003
/* standard representation of integers */
#define enc_int(val)    ((Integer)(val) << 3)
#define dec_int(val)    ((Integer)(val) >> 3)
/* standard encoding of pointers */
#define enc_addr(addr)  ((Cell)(addr) << 1)
#define dec_addr(dcell) (((Cell)(dcell) >> 1) & 0x7ffffffc)
/*#define dec_addr(dcell) ((Cell)(dcell) >> 1)*/

#elif (HIGH_BIT_TO_STEAL == 30)

#define MALLOC_BIT_MASK	0x40000003
/* standard representation of integers */
#define enc_int(val)    ((Integer)(val) << 3)
#define dec_int(val)    ((Integer)(val) >> 3)
/* Fewer bit representation of pointers (take bits 0-1, 30 only) */
#define enc_addr(addr)  (((Cell)(addr) & 0x80000000) |\
			 (((Cell)(addr) << 1) & 0x7ffffff8))
#define dec_addr(dcell) (((Cell)(dcell) & 0x80000000) |\
                         (((Cell)(dcell) >> 1) & 0x3ffffffc))

#else /* HIGH_BITS_TO_STEAL unknown */

#define enc_int(val)    ( ((Integer)(val) & 0xc0000003) ?\
		          (((Integer)(val) << 4) | 0x08) :\
		          ((Integer)(val) << 2) )
#define dec_int(val)    ( ((Integer)(val) & 0x08) ?\
		          ((Integer)(val) >> 4) :\
		          (((Integer)(val) >> 2) & 0x3ffffffc) )
#define enc_addr(addr)  ((Cell)(addr) << 1)
#define dec_addr(dcell) (((Cell)(dcell) >> 1) & 0x7ffffffc)

#endif

#else
/* standard representation of integers */
#define enc_int(val)    ((Integer)(val) << 3)
#define dec_int(val)    ((Integer)(val) >> 3)
/* standard encoding of pointers */
#define enc_addr(addr)  ((Cell)(addr) << 3)
#define dec_addr(dcell) ((Cell)(dcell) >> 3)
#endif

/*======================================================================*/

/* integer manipulation */
#define int_val(dcell) (Integer)dec_int(dcell)
#define makeint(val)   (Cell)((enc_int(val)) | INT)

/* float manipulation */
#define float_val(dcell) getfloatval(dcell)

/* atom manipulation */
#define atom_val(dcell)    (char *)dec_addr(dcell)
#define makeatom(str)      (Cell)(enc_addr(str) | ATM)
/* special-case atom: [] */
#define	makenil		   makeatom(nil_sym)

/* pointer manipulation */
#define struct_val(dcell)  (Pair)dec_addr(dcell)
#define makestruct(str)    (Cell)(enc_addr(str) | STRUCT)
#define clref_val(dcell)   (CPtr)dec_addr(dcell)
#define makelist(list)     (Cell)(enc_addr(list) | LIST)
#define makeattv(attv)     (Cell)(enc_addr(attv) | ATTV)
#define trievar_val(dcell) (Integer)dec_int(dcell)
#define maketrievar(val)   (Cell)(enc_int(val) | TrieVar)

#define addr_val(dcell)    int_val(dcell)
#define makeaddr(val)      makeint(val)


#define bld_nil(addr)          cell(addr) = makenil
#define bld_atom(addr, atm)    cell(addr) = makeatom(atm)
#define bld_int(addr, val)     cell(addr) = makeint(val)
#define bld_float(addr, val)   cell(addr) = makefloat((float)val)
#define bld_ref(addr, val)     cell(addr) = (Cell)(val)
#define bld_struct(addr, str)  cell(addr) = makestruct(str)
#define bld_list(addr, list)   cell(addr) = makelist(list)
#if defined(ATTR_VARS)
#define bld_attv(addr, attv)   cell(addr) = makeattv(attv)
#endif
#define bld_functor(addr, psc) cell(addr) = (word)psc
#define bld_copy(addr, val)    cell(addr) = val
#define bld_free(addr)         cell(addr) = (Cell)(addr)/* CPtr => FREE cell */

#define isinteger(dcell) (cell_tag(dcell)==INT)	   /* dcell -> bool */
#define isfloat(dcell)   (cell_tag(dcell)==FLOAT)  /* dcell -> bool */
#define isatom(dcell)	 (cell_tag(dcell)==ATM)
#define islist(dcell)    (cell_tag(dcell)==LIST)   /* dcell -> bool */
#define isstruct(dcell)  (cell_tag(dcell)==STRUCT) /* dcell -> bool */
#define isattv(dcell)    (cell_tag(dcell)==ATTV)   /* dcell -> bool */

#define numequal(num1, num2) num1 == num2
#define isnumber(dcell)	((isinteger(dcell)) || (isfloat(dcell)))
#define isatomic(dcell)	((isatom(dcell)) || (isnumber(dcell)))
#define isnil(dcell)    (dcell == makenil)  /* simplified Dec 24th */

#define get_str_psc(dcell) ((struct_val(dcell))->psc_ptr)

/*======================================================================*/
/* Miscellaneous							*/
/*======================================================================*/

#define MAX_ARITY	255

#define arity_integer(dcell) \
			(isinteger(dcell) && int_val(dcell) >= 0 \
					  && int_val(dcell) <= MAX_ARITY)

#endif
