/* File:      cst.c
** Author(s): Daniel Aldman, Kostis Sagonas
** Contact:   kostis@csd.uu.se
**
** Copyright (C) Daniel Aldman, Kostis Sagonas, Uppsala University, 2003
**
** XXX is free software; you can redistribute it and/or modify it under the
** terms of the GNU Library General Public License as published by the Free
** Software Foundation; either version 2 of the License, or (at your option)
** any later version.
**
** XXX is distributed in the hope that it will be useful, but WITHOUT ANY
** WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
** FOR A PARTICULAR PURPOSE.  See the GNU Library General Public License for
** more details.
**
** You should have received a copy of the GNU Library General Public License
** along with XXX; if not, write to the Free Software Foundation,
** Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
**
** $Id: cst.c,v 1.39 2005/02/15 15:42:19 kostis Exp $
**
*/

#include "xxx_config.h"

#include <stdio.h>       /* just for debugging... */
#include <stdlib.h>
#include <string.h>

#include "auxlry.h"
#include "celltags_xsb.h"
#include "cell_xsb.h"
#include "struct_manager.h"
#include "heap_xsb.h"
#include "psc_xsb.h"
#include "deref.h"
#include "macro_xsb.h"
#include "register.h"
#include "memory_xsb.h"
#include "binding.h"
#include "flags_xsb.h"
#include "inst_xsb.h"
#include "trie_internals.h"
#include "tr_utils.h"
#include "error_xsb.h"
#include "cst.h"

#if defined(CST_SUPPORT)

/* Comment out the following line to stop seeing any debug output */
/* #define CST_DEBUG */

#ifdef CST_DEBUG
/* Comment out the following line to stop seeing debug output from
   the code which build CST's from attributes */
/* #define CST_DEBUG_GATHER */
/* Comment out the following line to stop seeing debug output from
   the code which builds attributes from CST's  */
/* #define CST_DEBUG_BUILD */
int dbglvl = 0;

#define dbgpush dbglvl++
#define dbgpop dbglvl--
#define dbg \
{ \
    int dbgcntr; \
    fprintf(stderr, "CST DEBUG: "); \
    for(dbgcntr = 0; dbgcntr < dbglvl; dbgcntr++) \
        fprintf(stderr, " "); \
}

static void CST_print_constraint(CST_constraint	con, FILE *fp)
{
    if(isinteger(con.variable_1))
	fprintf(fp, "%d", int_val(con.variable_1));
    else if(cell_tag(con.variable_1) == TrieVar)
	fprintf(fp, "V_%d", trievar_val(con.variable_1));
    else
	fprintf(fp, "(Unknown)");
    
    if(con.type == CST_less_than)
	fprintf(fp, " <  ");
    else if(con.type == CST_less_than_or_equal)
	fprintf(fp, " <= ");
    else if(con.type == CST_equal)
	fprintf(fp, " =  ");
    else if(con.type == CST_more_than_or_equal)
	fprintf(fp, " >= ");
    else if(con.type == CST_more_than)
	fprintf(fp, " >  ");
    else 
	fprintf(fp, " (Unknown constraint type %d) ", con.type);
    
    if(isinteger(con.variable_2))
	fprintf(fp, "%d", int_val(con.variable_2));
    else if(cell_tag(con.variable_2) == TrieVar)
	fprintf(fp, "V_%d", trievar_val(con.variable_2));
    else
	fprintf(fp, "(Unknown)");
    
    fprintf(fp, " + ");
    
    if(isinteger(con.constant))
	fprintf(fp, "%d", int_val(con.constant));
    else if(isfloat(con.constant))
	fprintf(fp, "%f", float_val(con.constant));
    else
	fprintf(fp, "(Unknown)");
    
}

static void CL_print(CL list, FILE *fp)
{
    if(list == NULL)
	fprintf(fp, "\n");
    while(!(list == NULL))
    {
	CST_print_constraint(list->con, fp);
	if(list->next != NULL)
	    fprintf(fp, " /\\ ");
	else
	    fprintf(fp, "\n");
	list = list->next;
    }
}
#endif

static CST last_inserted_tree, last_inserted_node;

/* 
   These variables keep track of the total number of used/free CST and
   CL nodes, for statistical purposes.
*/
static unsigned long 
total_number_of_used_cst_nodes = 0, 
    total_number_of_unused_cst_nodes = 0,
    total_number_of_used_cl_nodes = 0, 
    total_number_of_unused_cl_nodes = 0;

/* 
   This variable holds the free CST nodes.
*/
static CST free_cst_nodes_storage = (CST) NULL;

/* 
   This variable holds the free CL nodes.
*/
static CL free_cl_nodes_storage = (CL) NULL;

/* 
   This define holds the number of CST nodes that should be allocated
   each time a node is requested and none are free.
*/
#define NUMBER_OF_NODES_PER_BLOCK (1024)

static Structure_Manager sm_new_cst_nodes = 
SM_InitDecl(
    struct CST_node, 
    NUMBER_OF_NODES_PER_BLOCK, 
    "constraint search tree nodes");

static Structure_Manager sm_new_cl_nodes = 
SM_InitDecl(
    struct CL_node, 
    NUMBER_OF_NODES_PER_BLOCK, 
    "constraint list nodes");

#define	CST_constraints_equal(con1,con2)	\
    (((con1).variable_1 == (con2).variable_1) &&    \
     ((con1).type == (con2).type) &&		\
     ((con1).variable_2 == (con2).variable_2) &&	\
     ((con1).constant == (con2).constant))

void CST_release_resources(void)
{
    SM_ReleaseResources(sm_new_cst_nodes);
    SM_ReleaseResources(sm_new_cl_nodes);
    free_cl_nodes_storage = NULL;
    total_number_of_unused_cl_nodes = 0;
    total_number_of_used_cl_nodes = 0;
    free_cst_nodes_storage = NULL;
    total_number_of_unused_cst_nodes = 0;
    total_number_of_used_cst_nodes = 0;
}

static void CL_allocate_more_nodes(void)
{
    CL	 node;
    long i;
    
    for(i = 0; i < NUMBER_OF_NODES_PER_BLOCK; i++)
    {
	/* Allocate one more node */
	SM_AllocateStruct(sm_new_cl_nodes, node);
	/* insert it into the free node storage */
	node->next = free_cl_nodes_storage;
	free_cl_nodes_storage = node;
	/* update the statistics counter */
	total_number_of_unused_cl_nodes++;
    }
}

CL CL_insert_node(CST_constraint con, CL next)
{
    CL	node;
    
    /* do we need to allocate more CL nodes? */
    if(free_cl_nodes_storage == NULL)
	/* yes */
	CL_allocate_more_nodes();
    
    /* take the first node out of storage */
    node = free_cl_nodes_storage;
    free_cl_nodes_storage = free_cl_nodes_storage->next;
    
    /* update statistics */
    total_number_of_used_cl_nodes++;
    total_number_of_unused_cl_nodes--;
    
    /* fill in the data */
    node->con = con;
    node->next = next;
    
    return node;
}

CL CL_remove_node(CL node)
{
    CL	reply = NULL;
    
    /* Return the next element in the list - later. */
    if(node->next != NULL)
	reply = node->next;
    
    /* Hook this node into the free nodes list */
    node->next = free_cl_nodes_storage;
    free_cl_nodes_storage = node;
    
    /* update statistics */
    total_number_of_used_cl_nodes--;
    total_number_of_unused_cl_nodes++;
    
    return reply;
}

/* Given a node in a list, this function returns the first node in that list */
static CL CL_get_last_node(CL list)
{
    if(list == NULL)
	return NULL;
    
    while(list->next != NULL)
	list = list->next;
    
    return list;
}

/* Given a node in a list, this function creates an identical copy of
   that list, and returns the first element in the copy */
static CL CL_clone(CL list)
{
    CL  newlist, oldnode;

    list = CL_get_last_node(list);
    newlist = NULL;
    oldnode = NULL;
    while(list != NULL)
    {
	CL newnode;
	newnode = CL_insert_node(list->con, NULL);
	if(oldnode != NULL)
	    oldnode->next = newnode;
	oldnode = newnode;
	if(newlist == NULL)
	    newlist = oldnode;
    }

    return newlist;
}

void CL_release(CL list)
{
    while(list != NULL)
	list = CL_remove_node(list);
}

static CL CL_remove_constraint(CL list, CST_constraint con)
{
    CL	tmp_list = list, last_element;
    
    last_element = list;
    while(tmp_list != NULL)
    {
	if(CST_constraints_equal(tmp_list->con, con))
	{
	    if(tmp_list == list)
	    {
		return CL_remove_node(tmp_list);
	    }else{
		last_element->next = CL_remove_node(tmp_list);
		return list;
	    }
	}
	last_element = tmp_list;
	tmp_list = tmp_list->next;
    }
    
    return list;
}

/* 
   This function, given a list and a constraint, returns true iff the
   constraint was present in the list.  It also deletes the node
   containing the constraint iff it was found.  This function will not
   search backwards in the list.
*/
static bool CL_has_constraint(CL list, CST_constraint con)
{
    while(list != NULL)
    {
	if(CST_constraints_equal(list->con, con))
	    return 1;
	
	list = list->next;
    }
    return 0;
}

typedef struct {
    CL		list1,list2,list3;
}cl_3_tuple;

/* 
   This function, given two lists, determines which is the first
   element of list1 which is not in both lists, unlinks it from the
   list and return it as list3 of the return value. 
*/
static cl_3_tuple CL_split(CL list1, CL list2)
{
    cl_3_tuple reply;
    CL         lastnode;
    
    reply.list1 = list1;
    reply.list2 = list2;
    reply.list3 = NULL;
    
    lastnode = NULL;
    while(!((list1 == NULL) || (list2 == NULL)))
    {
	if(!CL_has_constraint(reply.list2, list1->con))
	{
	    reply.list3 = CL_insert_node(list1->con, NULL);
	    
	    if(lastnode != NULL)
		lastnode->next = CL_remove_node(list1);
	    else
		reply.list1 = CL_remove_node(list1);
	    
	    return reply;
	}else{
	    lastnode = list1;
	    list1 = list1->next;
	}
    }
    return reply;
}

static void CST_allocate_more_nodes(void)
{
    CST	 node;
    long i;
    
    for(i = 0; i < NUMBER_OF_NODES_PER_BLOCK; i++)
    {
	/* Allocate one more node */
	SM_AllocateStruct(sm_new_cst_nodes, node);
	/* insert it into the free node storage */
	node->left = free_cst_nodes_storage;
	free_cst_nodes_storage = node;
	/* update the statistics counter */
	total_number_of_unused_cst_nodes++;
    }
}

void CST_fix_opcode(CST node)
{
#if defined(CST_DEBUG_GATHER)
    dbg;fprintf(stderr, 
		"Setting opcode (%p,%p)...",
		node,
		node->auxdata);
#endif
    if(node->auxdata != NULL)
    {
	if(CST_is_root_node(node) || node->parent == NULL)
	{
	    set_opcode(node, cst_return);
#if defined(CST_DEBUG_GATHER)
	    fprintf(stderr, "cst_return\n");
#endif
	}else if((node->parent->right == node &&
		  node->parent->left != NULL) ||
		 (node->parent->auxdata != NULL))
	{
	    set_opcode(node, cst_return_cp);
#if defined(CST_DEBUG_GATHER)
	    fprintf(stderr, "cst_return_cp\n");
#endif
	}else{
	    set_opcode(node, cst_return);
#if defined(CST_DEBUG_GATHER)
	    fprintf(stderr, "cst_return\n");
#endif
	}
    }else{
	if(CST_is_root_node(node) || node->parent == NULL)
	{
	    set_opcode(node, cst_continue);
#if defined(CST_DEBUG_GATHER)
	    fprintf(stderr, "cst_continue\n");
#endif
	}else if((node->parent->right == node &&
		  node->parent->left != NULL) ||
		 (node->parent->auxdata != NULL))
	{
	    set_opcode(node, cst_continue_cp);
#if defined(CST_DEBUG_GATHER)
	    fprintf(stderr, "cst_continue_cp\n");
#endif
	}else{
	    set_opcode(node, cst_continue);
#if defined(CST_DEBUG_GATHER)
	    fprintf(stderr, "cst_continue\n");
#endif
	}
    }
}

static CST CST_create_node(CL cons, 
			   void *auxdata, 
			   CST parent, 
			   CST left, 
			   CST right)
{
    CST	node;
    
    /* do we need to allocate more CST nodes? */
    if(free_cst_nodes_storage == NULL)
	/* yes */
	CST_allocate_more_nodes();
    
    /* take the first node out of storage*/
    node = free_cst_nodes_storage;
    free_cst_nodes_storage = free_cst_nodes_storage->left;
    
    /* update statistics */
    total_number_of_used_cst_nodes++;
    total_number_of_unused_cst_nodes--;
    
    /* fill in the data */
    node->constraints = cons;
    node->auxdata = auxdata;
    node->parent = parent;
    node->left = left;
    node->right = right;
    CST_fix_opcode(node);
    return node;
}

/*
  This function takes care of the memory used by a CST node when it is
  no longer used.  Do not call this function twice with the same node. 
  Any children of this node will not be deallocated.
*/
static void CST_release_node(CST node)
{
    /* just checking to make sure we don't dereference NULL pointers */
    if(node == NULL)
	return;
    
    CL_release(node->constraints);
    
    /* put the node into the node storage */
    node->left = free_cst_nodes_storage;
    free_cst_nodes_storage = node;
    
    /* update statistics */
    total_number_of_used_cst_nodes--;
    total_number_of_unused_cst_nodes++;
}

void CST_release_tree(CST node)
{
    CST previous_node;

    /* just checking to make sure we don't dereference NULL pointers */
    if(node == NULL)
	return;
    
    previous_node = NULL;
    do{
	/* proceed to the leftmost child */
	while(node->left != NULL &&
	      node->left != previous_node)
	    node = node->left;
	/* if there is a right child, go there first */
	if(node->right != NULL &&
	   node->right != previous_node)
	{
	    previous_node = node;
	    node = node->right;
	}else{
	    /* Found a node whose children have been deleted. Delete
               and proceed with the parent (if one exists) */
	    previous_node = node;
	    if(!CST_is_root_node(node))
		node = node->parent;
	    else
		node = NULL;
	    CST_release_node(node);
	}
    }while(node != NULL);
}

void CST_statistics(unsigned long *cst_space, 
		    unsigned long *cst_unused_space,
		    unsigned long *cl_space,
		    unsigned long *cl_unused_space)
{
    *cst_space = sizeof(struct CST_node) * total_number_of_used_cst_nodes;
    *cst_unused_space = 
	sizeof(struct CST_node) * total_number_of_unused_cst_nodes;
    *cl_space = sizeof(struct CL_node) * total_number_of_used_cl_nodes;
    *cl_unused_space = 
	sizeof(struct CL_node) * total_number_of_unused_cl_nodes;
}

#define LESS_THAN		     	(0)
#define LESS_THAN_OR_EQUAL	        (1)
#define EQUAL			       	(2)
#define NOT_EQUAL			(3)
#define MORE_THAN_OR_EQUAL	        (4)
#define MORE_THAN		       	(5)
#define CONSTRAINT_HOLDER               (6)
#define NUM_STRUCTURES			(7)

char struct_name[NUM_STRUCTURES][64] = {
    "<",
    "=<",
    "=",
    "\\=",
    ">=",
    ">",
    "constraint"
};

#define LEFT_CONSTRAINT_NAME		(struct_name[LEFT_CONSTRAINT])
#define RIGHT_CONSTRAINT_NAME	       	(struct_name[RIGHT_CONSTRAINT])
#define LESS_THAN_NAME		       	(struct_name[LESS_THAN])
#define LESS_THAN_OR_EQUAL_NAME		(struct_name[LESS_THAN_OR_EQUAL])
#define EQUAL_NAME		       	(struct_name[EQUAL])
#define NOT_EQUAL_NAME                  (struct_name[NOT_EQUAL])
#define MORE_THAN_OR_EQUAL_NAME		(struct_name[MORE_THAN_OR_EQUAL])
#define MORE_THAN_NAME		       	(struct_name[MORE_THAN])
#define CONSTRAINT_HOLDER_NAME		(struct_name[CONSTRAINT_HOLDER])

char typestr[8][32] = {
    "free variable or reference(0) ",
    "structure(1) ",
    "integer(2) ",
    "list(3) ",
    "reference(4) ",
    "atom(5) ",
    "floating point number(6) ",
    "attributed variable(7) "
};

/* 
 * Code to store constrained variables' constraints in a constraint list
 */

#define get_attv_attribute(attv) (clref_val(attv)+1)
#define head_p(list) (clref_val(cell(list)))
#define tail_p(list) (clref_val(cell(list))+1)

#define head(list) (cell(clref_val(list)))
#define tail(list) (cell((CPtr) clref_val(list)+1))

#define StandardizeAndTrailVariable(addr,n)     \
    StandardizeVariable(addr,n);                \
    if(trail_top != NULL)                       \
    {                                           \
       if(trail_goes_up)                        \
          *(++(*trail_top)) = addr;             \
       else                                     \
          *(--(*trail_top)) = addr;             \
    }

static CL CL_parse_constraint_struct(CPtr strct, 
				     CL cons, 
				     int *trievar_ctr,
				     int *trieattv_ctr,
				     CPtr **trail_top,
				     int trail_goes_up);
static CL CL_parse_constraint_list(CPtr list, 
				   CL cons, 
				   int *trievar_ctr,
				   int *trieattv_ctr,
				   CPtr **trail_top,
				   int trail_goes_up);

CL CL_insert_constraints(Cell attv, 
			 CL cons, 
			 int *trievar_ctr, 
			 int *trieattv_ctr,
			 CPtr **trail_top,
			 int trail_goes_up)
{
    CPtr attr;
    
#ifdef CST_DEBUG_GATHER
    fprintf(stderr, "\n");
    dbg;fprintf(stderr, "CL_insert_constraints called.\n");
    dbg;fprintf(stderr, 
		"variables: %d, attributed variables: %d\n",
		*trievar_ctr, 
		*trieattv_ctr);
#endif
    /* check that the cell is in fact an attributed variable */
    if(!isattv(attv))
    {
	xsb_abort(
	    "CL_insert_constraints: attempted to add constraints from "
	    "something that was not an attributed variable (it was a %s).",
	    typestr[cell_tag(attv)]);
	return cons;
    }
    
    /* get this attributed variable's attributes */
    attr = get_attv_attribute(attv);
    
    if(isstruct(cell(attr)))
    {
	/* the arity of the structure should be 2 */
	if(get_arity((Psc)follow(struct_val(cell(attr)))) == 2)
	{
	    /* the name of the attribute should be v */
	    if(strcmp(get_str_psc(cell(attr))->nameptr, "v") == 0)
	    {
		Cell arg = cell((CPtr) (clref_val(cell(attr))+2));
		deref(arg);
		cons = 
		    CL_parse_constraint_list(
			(CPtr) &arg, 
			cons, 
			trievar_ctr, 
			trieattv_ctr,
			trail_top, 
			trail_goes_up);
	    }else
		xsb_abort(
		    "cl_insert_constraints: attribute name "
		    "is incorrect, it should be v, it is %s\n",
		    get_str_psc(cell(attr))->nameptr);
	}else
	    xsb_abort(
		"cl_insert_constraints: attribute arity "
		"is incorrect, it should be 2, it is %d\n",
		get_arity((Psc)follow(struct_val(cell(attr)))));
    }else{
#ifdef CST_DEBUG_GATHER
	dbg;fprintf(stderr, "attribute is unknown\n");
#endif
	xsb_abort(
	    "cl_insert_constraints: attribute is not a structure. "
	    "Attribute not processed.\n");
    }
#ifdef CST_DEBUG_GATHER
    dbg; fprintf(stderr, "CL_insert_constraints done.\n");
#endif
    return cons;
}

static CL CL_parse_constraint_list(CPtr list, 
				   CL cons, 
				   int *trievar_ctr,
				   int *trieattv_ctr,
				   CPtr **trail_top,
				   int trail_goes_up)
{
    /* go through the list */
#ifdef CST_DEBUG_GATHER
    dbgpush;
    dbg;fprintf(stderr, "CL_parse_constraint_list called.\n");
#endif
    
    if(!islist(cell(list)))
    {
	xsb_abort(
	    "CL_parse_constraint_list: attempted to parse a non-list. "
	    "This is probably due to a nonstandard second argument in v/2\n");
    }

    while(islist(cell(list)))
    {
	Cell head;
	head = cell(head_p(list));
	deref(head);
#ifdef CST_DEBUG_GATHER
	dbg;fprintf(stderr,
		    "list head: %s\n", 
		    typestr[cell_tag(head)]);
#endif
	if(isstruct(head))
	    cons = 
		CL_parse_constraint_struct((CPtr) &head, 
					   cons, 
					   trievar_ctr,
					   trieattv_ctr,
					   trail_top,
					   trail_goes_up);
	else if(islist(head))
	    cons = 
		CL_parse_constraint_list((CPtr) &head,
					 cons,
					 trievar_ctr,
					 trieattv_ctr,
					 trail_top,
					 trail_goes_up);
	else if(isref(head))
	{
	    /* Empty list - do nothing. */
	}else{
	    xsb_abort(
		"cl_parse_constraint_list: the list element is not "
		"a structure or a list (it is a %s), and has been skipped.\n",
		typestr[cell_tag(head)]);
	}
	list = tail_p(list);
    }
#ifdef CST_DEBUG_GATHER
    dbg;fprintf(stderr, "CL_parse_constraint_list done.\n");
    dbgpop;
#endif
    
    /* done - return the new CL */
    return cons;
}

static CL CL_make_trie_var(Cell var, 
			   CL cons,
			   int *trievar_ctr,
			   int *trieattv_ctr,
			   CPtr **trail_top,
			   int trail_goes_up)
{
    CPtr temp_var;
    
    deref(var);
    if(isattv(var))
    {
	/* 
	   This is an attributed variable, so it's standardized (and
	   trailed), and then it's constraints are also added to the
	   CST.
	*/
	if(trail_goes_up)
	    *(hreg++) = (Cell) var;
	temp_var = clref_val(var);
	StandardizeAndTrailVariable(
	    temp_var, 
	    (*trievar_ctr));
	(*trievar_ctr)++;
	(*trieattv_ctr)++;
#ifdef CST_DEBUG_GATHER
	dbg;fprintf(stderr, 
		    "New ATTV discovered - calling CL_insert_constraints "
		    "recursively.\n");
#endif
	cons =
	    CL_insert_constraints(var,
				  cons,
				  trievar_ctr,
				  trieattv_ctr,
				  trail_top,
				  trail_goes_up);
    }else if(isref(var) && !IsStandardizedVariable(var))
    {
	/* 
	   This is a simple variable, so just standardize (and trail)
	   it.
	*/
	StandardizeAndTrailVariable(
	    (CPtr)var, 
	    *(trievar_ctr));
	(*trievar_ctr)++;
#ifdef CST_DEBUG_GATHER
	dbg;fprintf(stderr, "New variable discovered - standardized.\n");
#endif
    }
    return cons;
}

static CL CL_parse_constraint_struct(CPtr strct, 
				     CL cons, 
				     int *trievar_ctr,
				     int *trieattv_ctr,
				     CPtr **trail_top,
				     int trail_goes_up)
{
    CST_constraint constr;
    int negate_constant = 0;
    
#ifdef CST_DEBUG_GATHER
    dbg;fprintf(stderr, "CL_parse_constraint_struct called.\n");
#endif    
    /* ...just in case */
    if(!isstruct(cell(strct)))
	return cons;
    
#ifdef CST_DEBUG_GATHER
    dbgpush;
    
    dbg;fprintf(stderr, 
		"structure name: %s\n", 
		get_str_psc(cell(strct))->nameptr);
    dbg;fprintf(stderr, 
		"structure arity: %d\n", 
		get_arity((Psc)follow(struct_val(cell(strct)))));
#endif
    
    /* 
     * for now, assume (constraint name)/2 structures to hold actual
     * constraints 
     */
    if(get_arity((Psc)follow(struct_val(cell(strct)))) != 2)
    {
	xsb_abort(
	    "cl_parse_constraint_struct: Error in constraint. The "
	    "constraint struct's arity is incorrect.\n");
#ifdef CST_DEBUG_GATHER
	dbgpop;
#endif
	return cons;
    }
    
    if(strcmp(get_str_psc(cell(strct))->nameptr, LESS_THAN_NAME) == 0)
	constr.type = CST_less_than;
    else if(strcmp(get_str_psc(cell(strct))->nameptr, 
		   LESS_THAN_OR_EQUAL_NAME) == 0)
	constr.type = CST_less_than_or_equal;
    else if(strcmp(get_str_psc(cell(strct))->nameptr, EQUAL_NAME) == 0)
	constr.type = CST_equal;
    else if(strcmp(get_str_psc(cell(strct))->nameptr, NOT_EQUAL_NAME) == 0)
	constr.type = CST_not_equal;
    else if(strcmp(get_str_psc(cell(strct))->nameptr, 
		   MORE_THAN_OR_EQUAL_NAME) == 0)
	constr.type = CST_more_than_or_equal;
    else if(strcmp(get_str_psc(cell(strct))->nameptr, 
		   MORE_THAN_NAME) == 0)
	constr.type = CST_more_than;
    else{
	xsb_abort(
	    "cl_parse_constraint_struct: Error in constraint. The "
	    "constraint struct's name is not one that is recognized. "
	    "(name = %s)\n",
	    get_str_psc(cell(strct))->nameptr);
#ifdef CST_DEBUG_GATHER
	dbgpop;
#endif
	return cons;
    }
    
#ifdef CST_DEBUG_GATHER
    dbg;fprintf(stderr, "structure determined to be a correct constraint\n");
#endif
    
    /* find first argument */
    constr.variable_1 = (Cell) (clref_val(cell(strct))+1);
    /* make sure that it is a trie var or attv */
    cons = 
	CL_make_trie_var(
	    constr.variable_1,
	    cons,
	    trievar_ctr,
	    trieattv_ctr,
	    trail_top,
	    trail_goes_up);
    deref(constr.variable_1);
#ifdef CST_DEBUG_GATHER
    dbg;
    fprintf(stderr, 
	    "first argument (should be a standardized variable), type: %s, "
	    "(%X) status: ",
	    typestr[cell_tag(constr.variable_1)],
	    (unsigned int) constr.variable_1);
#endif
    /* check that the first argument's type is correct */ 
    if(IsStandardizedVariable(constr.variable_1))
    {
#ifdef CST_DEBUG_GATHER
        fprintf(stderr, "variable - ok!\n");
#endif
	constr.variable_1 = EncodeTrieVar(IndexOfStdVar(constr.variable_1));
    }else{
#ifdef CST_DEBUG_GATHER
	fprintf(stderr, "failed!\n");
#endif
	xsb_abort(
	    "cl_parse_constraint_struct: Error in constraint. The "
	    "constraint's first argument is not a variable. Constraint "
	    "not processed.\n");
#ifdef CST_DEBUG_GATHER
	dbgpop;
#endif
	return cons;
    }
    
    /* find second argument */
    constr.variable_2 = (Cell) (clref_val(cell(strct))+2);
    /* make sure that it is a trie var or attv (if possible) */
    cons =
	CL_make_trie_var(
	    constr.variable_2,
	    cons,
	    trievar_ctr,
	    trieattv_ctr,
	    trail_top,
	    trail_goes_up);
    deref(constr.variable_2);
#ifdef CST_DEBUG_GATHER
    dbg;
    fprintf(stderr, 
	    "second argument (should be either constant, variable, +/2 "
	    "or -/2), type: %sstatus: ",
	    typestr[cell_tag(constr.variable_2)]);
#endif
    /* check that the second argument's type is correct */
    if(IsStandardizedVariable(constr.variable_2))
    {
#ifdef CST_DEBUG_GATHER
        fprintf(stderr, "variable - ok!\n");
#endif
	constr.variable_2 = EncodeTrieVar(IndexOfStdVar(constr.variable_2));
	constr.constant = makeint(0);
    }else if(isinteger(constr.variable_2) || 
	     isfloat(constr.variable_2))
    {
#ifdef CST_DEBUG_GATHER
	fprintf(stderr, "constant - ok!\n");
#endif
	constr.constant = constr.variable_2;
	constr.variable_2 = makeint(0);
    }else if(isstruct(constr.variable_2))
    {
#ifdef CST_DEBUG_GATHER
	fprintf(stderr, "structure - ok!\n");
#endif
	/* test the arity of the structure */
	if(get_arity((Psc)follow(struct_val(constr.variable_2))) != 2)
	{
	    xsb_abort(
		"cl_parse_constraint_struct: Error in constraint. The arity "
		"of the constraint struct's second argument is not 2.\n");
#ifdef CST_DEBUG_GATHER
	    dbgpop;
#endif
	    return cons;
	}
	
	if(strcmp("+", get_str_psc(constr.variable_2)->nameptr) == 0)
	{
#ifdef CST_DEBUG_GATHER
	    dbg;fprintf(stderr, "structure checked out ok.\n");
#endif
	}else if(strcmp("-", get_str_psc(constr.variable_2)->nameptr) == 0)
	{
#ifdef CST_DEBUG_GATHER
	    fprintf(stderr, "structure checked out ok.\n");
#endif
	    negate_constant = 1;
	}else{
	    xsb_abort(
		"cl_parse_constraint_struct: Error in constraint. The "
		"constraint struct's name is not one that is recognized. "
		"(name = %s)\n",
		get_str_psc(constr.variable_2)->nameptr);
#ifdef CST_DEBUG_GATHER
	    dbgpop;
#endif
	    return cons;
	}

	/* find the actual things to put into the constraint */
	constr.constant = (Cell) (clref_val(constr.variable_2)+2);
	if(negate_constant == 1)
	{
	    if(isinteger(constr.constant))
		constr.constant = makeint(-int_val(constr.constant));
	    else if(isfloat(constr.constant))
		constr.constant = makefloat(-float_val(constr.constant));
	}
	constr.variable_2 = (Cell) (clref_val(constr.variable_2)+1);
	
	deref(constr.constant);
	if(!isinteger(constr.constant) && !isfloat(constr.constant))
	{
	    xsb_abort(
		"cl_parse_constraint_struct: Error in constraint. The "
		"constraint's constant is not a number.\n");
#ifdef CST_DEBUG_GATHER
	    dbgpop;
#endif
	    return cons;
	}
	
	cons = CL_make_trie_var(constr.variable_2,
				cons,
				trievar_ctr,
				trieattv_ctr,
				trail_top,
				trail_goes_up);
	deref(constr.variable_2);
        if(IsStandardizedVariable(constr.variable_2))
	{
	    constr.variable_2 = 
		EncodeTrieVar(IndexOfStdVar(constr.variable_2));
	}else{
	    xsb_abort(
		"cl_parse_constraint_struct: Error in constraint. The "
		"constraint's second variable is not a variable or attv.\n");
#ifdef CST_DEBUG_GATHER
	    dbgpop;
#endif
	    cons = CL_insert_node(constr, cons);
            return cons;
	}
    }else{
	xsb_abort(
	    "cl_parse_constraint_struct: Error in constraint. The "
	    "constraint's second variable is not a variable, constant, "
	    "+/2 or -/2..\n");
#ifdef CST_DEBUG_GATHER
	dbgpop;
#endif
	return cons;
    }
    cons = CL_insert_node(constr, cons);
#ifdef CST_DEBUG_GATHER
    dbgpop;
    dbg;fprintf(stderr, "CL_parse_constraint_struct done.\n");
#endif
    return cons;
}

typedef struct {
    CL l;
    char i;
} CL_int_tuple;

static char constraint_compare(CST_constraint *c1, CST_constraint *c2)
{
    if(c1->type < c2->type)
	return -1;
    if(c1->type > c2->type)
	return 1;
    if(trievar_val(c1->variable_1) < trievar_val(c2->variable_1))
	return -1;
    if(trievar_val(c1->variable_1) > trievar_val(c2->variable_1))
	return 1;
    if(trievar_val(c1->variable_2) < trievar_val(c2->variable_2))
	return -1;
    if(trievar_val(c1->variable_2) > trievar_val(c2->variable_2))
	return 1;
    if(isfloat(c1->constant) && isfloat(c2->constant))
    {
	if(float_val(c1->constant) < float_val(c2->constant))
	    return -1;
	if(float_val(c1->constant) > float_val(c2->constant))
	    return 1;
	return 0;
    }else if(isfloat(c1->constant) && isinteger(c2->constant))
    {
	if(float_val(c1->constant) < int_val(c2->constant))
	    return -1;
	if(float_val(c1->constant) > int_val(c2->constant))
	    return 1;
	return 0;
    }else if(isinteger(c1->constant) && isfloat(c2->constant))
    {
	if(int_val(c1->constant) < float_val(c2->constant))
	    return -1;
	if(int_val(c1->constant) > float_val(c2->constant))
	    return 1;
	return 0;
    }else if(isinteger(c1->constant) && isinteger(c2->constant))
    {
	if(int_val(c1->constant) < int_val(c2->constant))
	    return -1;
	if(int_val(c1->constant) > int_val(c2->constant))
	    return 1;
	return 0;
    }else if(c1->constant == makeint(0) && isfloat(c2->constant))
    {
	if(0 <= float_val(c2->constant))
	    return -1;
	else
	    return 1;
	return 0;
    }else if(c1->constant == makeint(0) && isinteger(c2->constant))
    {
	if(0 <= int_val(c2->constant))
	    return -1;
	else
	    return 1;
	return 0;
    }else if(c2->constant == makeint(0) && isfloat(c1->constant))
    {
	if(float_val(c1->constant) <= 0)
	    return -1;
	else
	    return 1;
	return 0;
    }else if(c2->constant == makeint(0) && isinteger(c1->constant))
    {
	if(int_val(c1->constant) <= 0)
	    return -1;
	else
	    return 1;
	return 0;
    }else if(c1->constant == makeint(0) && c2->constant == makeint(0))
	return 0;
    
    xsb_abort(
	"constraint_compare: when comparing constraints, one of the two's "
	"constant was not a float, an int or NULL. Tags: %ld & %ld\n",
	cell_tag(c1->constant),
	cell_tag(c2->constant));
    return 0;
}

static CL_int_tuple CAVL_perform_LL_rotation(CL tree)
{
    CL_int_tuple retval;
    CL           leftnode = tree->next;
    
    retval.i = 0;

    tree->next = leftnode->AVL_right;
    tree->AVL_balance = 0;
    leftnode->AVL_right = tree;
    leftnode->AVL_balance = 0;
    retval.l = leftnode;
    return retval;
}

static CL_int_tuple CAVL_perform_LR_rotation(CL tree)
{
    CL_int_tuple retval;
    CL           leftnode = tree->next;
    CL           leftrightnode = leftnode->AVL_right;
    
    retval.i = 0;

    tree->next = leftrightnode->AVL_right;
    tree->AVL_balance = 0;
    leftnode->AVL_right = leftrightnode->next;
    leftnode->AVL_balance = 0;
    leftrightnode->next = leftnode;
    leftrightnode->AVL_right = tree;
    leftrightnode->AVL_balance = 0;
    retval.l = leftrightnode;
    return retval;
}

static CL_int_tuple CAVL_perform_RR_rotation(CL tree)
{
    CL_int_tuple retval;
    CL           rightnode = tree->AVL_right;
    
    retval.i = 0;
    tree->AVL_right = rightnode->next;
    tree->AVL_balance = 0;
    rightnode->next = tree;
    rightnode->AVL_balance = 0;
    retval.l = rightnode;
    return retval;
}

static CL_int_tuple CAVL_perform_RL_rotation(CL tree)
{
    CL_int_tuple retval;
    CL           rightnode = tree->AVL_right;
    CL           rightleftnode = rightnode->next;
    
    retval.i = 0;
    tree->AVL_right = rightleftnode->next;
    tree->AVL_balance = 0;
    rightnode->next = rightleftnode->AVL_right;
    rightnode->AVL_balance = 0;
    rightleftnode->next = tree;
    rightleftnode->AVL_right = rightnode;
    rightleftnode->AVL_balance = 0;
    retval.l = rightleftnode;
    return retval;
}

static CL_int_tuple CAVL_insert(CL tree, CL new_node)
{
    char         compval;
    CL_int_tuple retval;

    if(tree == NULL)
    {
	retval.l = new_node;
	retval.i = 1;
	return retval;
    }
    
    compval = constraint_compare(&(tree->con), &(new_node->con));
    if(compval == 0)
    {
	CL_remove_node(new_node);
	retval.l = tree;
	retval.i = 0;
	return retval;
    }else if(compval > 0)
    {
	CL_int_tuple new_left;
	
	if(tree->next == NULL)
	{
	    new_left.l = new_node;
	    new_left.i = 1;
	}else{
	    new_left = CAVL_insert(tree->next, new_node);
	}
	tree->next = new_left.l;
	if(new_left.i == 0)
	{
	    retval.l = tree;
	    retval.i = 0;
	    return retval;
	}else if(new_left.i > 0)
	{
	    tree->AVL_balance++;
	    if(tree->AVL_balance > 1)
	    {
		if(tree->next->AVL_balance > 0)
		    return CAVL_perform_LL_rotation(tree);
		else
		    return CAVL_perform_LR_rotation(tree);
	    }else{
		retval.l = tree;
		if(tree->AVL_balance > 0)
		    retval.i = 1;
		else
		    retval.i = 0;
		return retval;
	    }
	}else{
	    tree->AVL_balance--;
	    if(tree->AVL_balance < -1)
	    {
		if(tree->AVL_right->AVL_balance < 0)
		    return CAVL_perform_RR_rotation(tree);
		else
		    return CAVL_perform_RL_rotation(tree);
	    }else{
		retval.l = tree;
		if(tree->AVL_balance < 0)
		    retval.i = 1;
		else
		    retval.i = 0;
		return retval;
	    }
	}
    }else{
	CL_int_tuple new_right;
	
	if(tree->AVL_right == NULL)
	{
	    new_right.l = new_node;
	    new_right.i = 1;
	}else{
	    new_right = CAVL_insert(tree->AVL_right, new_node);
	}
	tree->AVL_right = new_right.l;
	if(new_right.i == 0)
	{
	    retval.l = tree;
	    retval.i = 0;
	    return retval;
	}else if(new_right.i > 0)
	{
	    tree->AVL_balance--;
	    if(tree->AVL_balance < -1)
	    {
		if(tree->AVL_right->AVL_balance < 0)
		    return CAVL_perform_RR_rotation(tree);
		else
		    return CAVL_perform_RL_rotation(tree);
	    }else{
		retval.l = tree;
		if(tree->AVL_balance < 0)
		    retval.i = 1;
		else
		    retval.i = 0;
		return retval;
	    }
	}else{
	    tree->AVL_balance++;
	    if(tree->AVL_balance > 1)
	    {
		if(tree->next->AVL_balance > 0)
		    return CAVL_perform_LL_rotation(tree);
		else
		    return CAVL_perform_LR_rotation(tree);
	    }else{
		retval.l = tree;
		if(tree->AVL_balance > 0)
		    retval.i = 1;
		else
		    retval.i = 0;
		return retval;
	    }
	}
    }
}

static CL CAVL_linearize(CL tree)
{
    CL left, right;

    if(tree == NULL)
	return NULL;
    
    left = CAVL_linearize(tree->next);
    right = CAVL_linearize(tree->AVL_right);
    
    tree->next = right;
    if(right == NULL)
	tree->AVL_right = tree;
    else
	tree->AVL_right = tree->AVL_right;
    
    if(left == NULL)
	return tree;
    
    left->AVL_right->next = tree;
    left->AVL_right = tree->AVL_right;
    return left;
}

CL CL_remove_duplicates(CL constraints)
{
    CL tree, node;
    
    tree = NULL;
    while(constraints != NULL)
    {
	node = constraints;
	constraints = node->next;
	node->next = NULL;
	node->AVL_right = NULL;
	node->AVL_balance = 0;
	tree = CAVL_insert(tree, node).l;
    }
    
    return CAVL_linearize(tree);
}

static CST_2_bool_tuple CST_insert_constraints_hlp(CL cons, 
						   CST tree, 
						   CST parent)
{
    CST	        newroot;
    CST_2_bool_tuple retval;
    
    if(tree == NULL)
    {
#if defined(CST_DEBUG_GATHER)
	dbg;fprintf(stderr, 
		    "CST_insert_constraints_hlp: tree == NULL, "
		    "creating new node\n");
#endif
	/* there are some constraints to insert - better insert them here */
	newroot = CST_create_node(cons, (void *)1, parent, NULL, NULL);
	/* Make sure that this node is in the list of nodes that
           answers can be returned from. */
	retval.tree1 = newroot;
	retval.tree2 = newroot;
	retval.b = 1;
	return retval;
    }

    if(tree->constraints == NULL)
    {
#if defined(CST_DEBUG_GATHER)
	dbg;fprintf(stderr, 
		    "CST_insert_constraints_hlp: true node, "
		    "creating new node as parent to this node\n");
#endif
	/* this is a "true" node - create a new node and put it as the
           parent of this node, which is put as the right subtree of
           the new node.*/
	retval.tree1 = retval.tree2 =
	    CST_create_node(cons,
			    (void *)1,
			    parent,
			    NULL,
			    tree);
	tree->parent = retval.tree1;
	CST_fix_opcode(tree);
	retval.b = 1;
	return retval;
    }else if(tree->constraints->next == NULL)
    {
#if defined(CST_DEBUG_GATHER)
	dbg;fprintf(stderr, 
		    "CST_insert_constraints_hlp: slim node, "
		    "checking which way to go: ");
	CL_print(tree->constraints, stderr);
#endif
	/* only one constraint in the node - check wether or not it is
	   in the constraint list, to see which way to go */
	if(CL_has_constraint(cons, tree->constraints->con))
	{
#if defined(CST_DEBUG_GATHER)
	    dbg;fprintf(stderr, 
			"CST_insert_constraints_hlp: slim node, "
			"removing constraint and going left\n");
#endif
	    /* it's in the list - go left */
	    cons = CL_remove_constraint(cons, tree->constraints->con);
	    if(cons != NULL)
	    {
		retval = CST_insert_constraints_hlp(cons, tree->left, tree);
		if(retval.b == 1)
		{
		    tree->left = retval.tree1;
		    CST_fix_opcode(tree->left);
		    if(tree->right != NULL)
			CST_fix_opcode(tree->right);
		}
		retval.tree1 = tree;
		return retval;
	    }else{
#if defined(CST_DEBUG_GATHER)
		dbg;fprintf(stderr, 
			    "CST_insert_constraints_hlp: out of constraints, "
			    "flagging node and returning.\n");
#endif
		/* there are no more constraints left to insert into
		   the tree. Make sure that this node is in the list
		   of nodes that answers can be returned from. */
		retval.tree1 = tree;
		retval.tree2 = tree;
		if(is_start_node(tree))
		{
#if defined(CST_DEBUG_GATHER)
		    dbg;fprintf(stderr, 
				"CST_insert_constraints_hlp: "
				"node was already flagged.\n");
#endif
		    retval.b = 0;
		}else{
		    tree->auxdata = (void *)1;
		    CST_fix_opcode(tree);
		    retval.b = 1;
		}
		return retval;
	    }
	}else{
#if defined(CST_DEBUG_GATHER)
	    dbg;fprintf(stderr, 
			"CST_insert_constraints_hlp: slim node, "
			"going right\n");
#endif
	    /* It's not in the list - go right */
	    retval = CST_insert_constraints_hlp(cons, tree->right, tree);
	    if(retval.b == 1)
	    {
		tree->right = retval.tree1;
		CST_fix_opcode(tree->right);
	    }
	    retval.tree1 = tree;
	    return retval;
	}
    }else{
	/* This is a fat node. time to split. */
	cl_3_tuple		tup;
	
#if defined(CST_DEBUG_GATHER)
	dbg;fprintf(stderr, 
		    "CST_insert_constraints_hlp: fat node - ");
	CL_print(tree->constraints, stderr);
#endif
	tup = CL_split(tree->constraints, cons);
	if(tup.list3 == NULL)
	{
	    /* There were no discriminating constraints in the tree's
               current constriants */
#if defined(CST_DEBUG_GATHER)
	    dbg;fprintf(stderr, 
			"CST_insert_constraints_hlp: fat node, "
			"splitting(2)\n");
#endif
	    tup = CL_split(cons, tree->constraints);
	    if(tup.list3 == NULL)
	    {
#if defined(CST_DEBUG_GATHER)
		dbg;fprintf(stderr, 
			    "CST_insert_constraints_hlp: fat node, "
			    "couldn't split.\n");
#endif
		/* There were no discriminating constraints at
		   all. Make sure that this node is in the list of
		   nodes that answers can be returned from. */
		retval.tree1 = tree;
		retval.tree2 = tree;
		if(is_start_node(tree))
		{
#if defined(CST_DEBUG_GATHER)
		    dbg;fprintf(stderr, 
				"CST_insert_constraints_hlp: "
				"node was already flagged.\n");
#endif
		    retval.b = 0;
		}else{
		    tree->auxdata = (void *)1;
		    CST_fix_opcode(tree);
		    retval.b = 1;
		}
		return retval;
	    }else if(tup.list1 == NULL)
	    {
#if defined(CST_DEBUG_GATHER)
		dbg;fprintf(stderr, 
			    "CST_insert_constraints_hlp: fat node, "
			    "splitting revealed: no new constraints.\n");
#endif
		newroot = 
		    CST_create_node(tup.list3,
				    (void *)1,
				    parent,
				    NULL,
				    tree);
		tree->parent = newroot;
		CST_fix_opcode(tree);
		CST_fix_opcode(newroot);
		/* there are no more constraints left to insert into
		   the tree. Make sure that this node is in the list
		   of nodes that answers can be returned from. */
		retval.tree1 = newroot;
		retval.tree2 = newroot;
		retval.b = 1;
		return retval;
	    }else{
#if defined(CST_DEBUG_GATHER)
		dbg;fprintf(stderr, 
			    "CST_insert_constraints_hlp: fat node, "
			    "split(2)\n");
#endif
		newroot = 
		    CST_create_node(tup.list3,
				    NULL,
				    parent,
				    CST_create_node(
					tup.list1,
					(void *)1,
					NULL,
					NULL,
					NULL),
				    tree);
		tree->parent = newroot;
		tree->constraints = tup.list2;
		newroot->left->parent = newroot;
		CST_fix_opcode(tree);
		CST_fix_opcode(newroot);
		CST_fix_opcode(newroot->left);
		/* Make sure that newroot->left is in the list of
		   nodes that answers can be returned from. */
		retval.tree1 = newroot;
		retval.tree2 = newroot->left;
		retval.b = 1;
		return retval;
	    }
	}else{
#if defined(CST_DEBUG_GATHER)
	    dbg;fprintf(stderr, 
			"CST_insert_constraints_hlp: fat node, "
			"split(1)\n");
#endif
	    newroot = NULL;
	    if(tree->right != NULL)
		if(tree->right->constraints == NULL)
		{
		    newroot =
			CST_create_node(
			    tup.list3,
			    NULL,
			    parent,
			    tree,
			    CST_create_node(
				tup.list2,
				(void *)1,
				NULL,
				NULL,
				tree->right));
		    tree->right = NULL;
		}

	    if(newroot == NULL)
		newroot = 
		    CST_create_node(
			tup.list3,
			NULL,
			parent,
			tree,
			CST_create_node(
			    tup.list2,
			    (void *)1,
			    NULL,
			    NULL,
			    NULL));
	    tree->parent = newroot;
	    tree->constraints = tup.list1;
	    newroot->right->parent = newroot;
	    CST_fix_opcode(tree);
	    CST_fix_opcode(newroot);
	    CST_fix_opcode(newroot->right);
	    /* Make sure that newroot->right is in the list of nodes
               that answers can be returned from. */
	    retval.tree1 = newroot;
	    retval.tree2 = newroot->right;
	    retval.b = 1;
	    return retval;
	}
    }
}

static CST_2_bool_tuple CST_insert_constraints_hlp_2(CST tree, CST parent)
{
    CST_2_bool_tuple tup;
    
    if(tree == NULL)
    {
	tup.tree1 = tup.tree2 = 
	    CST_create_node(NULL, 
			    (void *)1,
			    parent, 
			    NULL, 
			    NULL);
	tup.b = 1;
    }else if(tree->constraints != NULL)
    {
	tup = CST_insert_constraints_hlp_2(tree->right, tree);
	if(tup.b == 1)
	{
	    tree->right = tup.tree1;
	    CST_fix_opcode(tree);
	}
	tup.tree1 = tree;
    }else{
	tup.tree1 = tree;
	tup.tree2 = tree;
	tup.b = 0;
    }
    return tup;
}

CST_2_bool_tuple CST_insert_constraints(CL constraints, CST tree)
{
    CST_2_bool_tuple tup;
#if defined(CST_DEBUG_GATHER)
    dbg;fprintf(stderr, 
		"inserting constriants (%p) into tree %p: ", 
		constraints, 
		tree);
    CL_print(constraints, stderr);
#endif
    if(constraints == NULL)
    {
	/* no constraints - insert "true" node */
	tup = CST_insert_constraints_hlp_2(tree, NULL);
    }else{
	/* insert constraints */
	tup = CST_insert_constraints_hlp(constraints, 
					 tree, 
					 NULL);
    }
    last_inserted_tree = tup.tree1;
    last_inserted_node = tup.tree2;
    if(tup.b == 1)
    {
	CST_fix_opcode(tup.tree1);
#if defined(CST_DEBUG_GATHER)
	dbg;fprintf(stderr, 
		    "inserted node and tree: %p and %p resp.\n",
		    last_inserted_node,
		    last_inserted_tree);
    }else{
	dbg;fprintf(stderr, 
		    "did not insert node and tree: %p and %p resp.\n",
		    last_inserted_node,
		    last_inserted_tree);
#endif
    }
    return tup;
}

CST CST_find_node(CST tree, CL constraints)
{
    if(tree == NULL)
    {
	/* Encountered an empty tree, or a true node */
	CL_release(constraints);
	return NULL;
    }

    if(tree->constraints == NULL)
    {
	/* True node - this is ok, iff we have no constraints left */
	if(constraints == NULL)
	{
	    return tree;
	}else{
	    CL_release(constraints);
	    return NULL;
	}
    }else if(tree->constraints->next == NULL)
    {
	/* Only one constraint in the node, see if it is one of the
           constraints we are looking for. */
	if(CL_has_constraint(constraints, tree->constraints->con))
	{
	    /* It was. Remove the constraint from the constraint list. */
	    constraints = CL_remove_constraint(constraints, 
					       tree->constraints->con);
	    
	    /* Iff there are any more constraints, continue down to
               the left child, otherwise this is the node we're
               looking for */
	    if(constraints == NULL)
		return tree;
	    else
		return CST_find_node(tree->left, constraints);
	}else{
	    /* It wasn't. Continue down to the right child. */
	    return CST_find_node(tree->right, constraints);
	}
    }else{
	/* There are more than one constraint in the node. Iff there
           are no constraints at all in the constraint list, try the
           right child as a true node. */
	if(constraints == NULL)
	    return CST_find_node(tree->right, constraints);
	else{
	    /* There were constraints in the constraint list. Iff
               these lists are equal, return this node, otherwise
               return NULL. */
	    cl_3_tuple tup;
	    CL         treeconstraints;
	    
	    treeconstraints = CL_clone(tree->constraints);
	    tup = CL_split(treeconstraints, constraints);
	    if(tup.list3 == NULL)
	    {
		tup = CL_split(constraints, treeconstraints);
		if(tup.list3 == NULL)
		{
		    CL_release(treeconstraints);
		    CL_release(constraints);
		    return tree;
		}else{
		    CL_release(tup.list1);
		    CL_release(tup.list2);
		    CL_release(tup.list3);
		    return NULL;
		}
	    }else{
		CL_release(tup.list1);
		CL_release(tup.list2);
		CL_release(tup.list3);
		return NULL;
	    }
	}
    }
}
/*
*/
#ifdef CST_DEBUG
/*#define CST_DEBUG_BUILD_CONSTRAINT */
#endif

static void CL_insert_constraint_struct(CPtr cstruct, CPtr Variable)
{
    Cell tailoflist, listloc;
    
#ifdef CST_DEBUG_BUILD_CONSTRAINT
    dbg;fprintf(stderr, "inserting constraint\n");
    dbg;fprintf(stderr, "hreg == %lX\n", (long) hreg);
#endif

    if(isattv(*Variable))
    {
#ifdef CST_DEBUG_BUILD_CONSTRAINT
	dbg;fprintf(stderr, "constraint already exists\n");
#endif
	tailoflist = *get_attv_attribute(*Variable);
#ifdef CST_DEBUG_BUILD_CONSTRAINT
	dbg;fprintf(stderr, "tail of list = %lX\n", tailoflist);
	dbg;printterm(tailoflist, 7, 7);fprintf(stderr, "\n");
#endif
	if(!islist(tailoflist))
	    tailoflist = makenil;
    }else if(isref(*Variable))
    {
#ifdef CST_DEBUG_BUILD_CONSTRAINT
	dbg;fprintf(stderr, "no previous constraint\n");
#endif
	tailoflist = makenil;
#ifdef CST_DEBUG_BUILD_CONSTRAINT
	dbg;fprintf(stderr, "tail of list = nil\n");
#endif
	cell(Variable) = makeattv(hreg);
	new_heap_free(hreg);
	new_heap_node(hreg, makenil);
    }else{
	xsb_abort(
	    "CL_insert_constraint_struct: Variable (%p / %p) not "
	    "a variable! (tag = %ld)\n", 
	    (void *) Variable, (void *) *Variable, (*Variable) & 0x7);
	return;
    }
#ifdef CST_DEBUG_BUILD_CONSTRAINT
    dbg;fprintf(stderr, "creating list...\n");
#endif
    listloc = (Cell) hreg;
    new_heap_node(hreg, (Cell)cstruct);
    new_heap_node(hreg, tailoflist);
#ifdef CST_DEBUG_BUILD_CONSTRAINT
    dbg;fprintf(stderr, "inserting list element in attribute\n");
#endif
    *(get_attv_attribute(*Variable)) = makelist(listloc);
#ifdef CST_DEBUG_BUILD_CONSTRAINT
    dbg;fprintf(stderr, "constraint inserted\n");
#endif
}

static void CL_return_answer_hlp(CL list, CPtr array)
{
    Cell constraint_location, tmp;
    Pair sym;
    int  new_indicator;
    
#ifdef CST_DEBUG_BUILD_CONSTRAINT
    dbg;fprintf(stderr,"CL_return_answer_hlp called\n");
#endif
    
    while(list != NULL)
    {
#ifdef CST_DEBUG_BUILD_CONSTRAINT
	dbg;CST_print_constraint(list->con, stderr);fprintf(stderr, "\n");
#endif
	if(isinteger(list->con.variable_1))
	{
	    /* nothing to do */
	}else if(cell_tag(list->con.variable_1) == TrieVar)
	{
	    sym = 
		(Pair)insert(
		    ((list->con.type == CST_less_than)?LESS_THAN_NAME:
		     ((list->con.type == CST_less_than_or_equal)?
		      LESS_THAN_OR_EQUAL_NAME:
		      ((list->con.type == CST_equal)?EQUAL_NAME:
		       ((list->con.type == CST_not_equal)?NOT_EQUAL_NAME:
			((list->con.type == CST_more_than_or_equal)?
			 MORE_THAN_OR_EQUAL_NAME:
			 MORE_THAN_NAME))))), 
		    2, 
		    (Psc)flags[CURRENT_MODULE],
		    &new_indicator);
	    
	    constraint_location = (Cell) hreg;
	    bld_struct(hreg, hreg + 1);
	    hreg++;
	    new_heap_functor(hreg, sym->psc_ptr); 
	    /* Variable 1 - 1 cell */
	    tmp = list->con.variable_1;
	    if(cell_tag(tmp) == TrieVar)
		tmp = (Cell) (&array[trievar_val(tmp)]);
	    new_heap_node(hreg, tmp);
	    
	    if(list->con.constant == makeint(0))
	    {
		/* Variable 2 - 1 cell */
		tmp = list->con.variable_2;
		if(cell_tag(tmp) == TrieVar)
		    tmp = (Cell) (&array[trievar_val(tmp)]);
		new_heap_node(hreg, tmp);
	    }else if(list->con.variable_2 == makeint(0))
	    {
		/* Constant - 1 cell */
		new_heap_node(hreg, list->con.constant);
	    }else{
                /* +/2 structure - 3 cells */
		bld_struct(hreg, hreg + 1);
		hreg++;
		sym =
		    (Pair)insert(
			"+",
			2,
			(Psc)flags[CURRENT_MODULE],
			&new_indicator);
                new_heap_functor(hreg, sym->psc_ptr);
                /* Variable 2 - 1 cell */
                tmp = list->con.variable_2;
                if(cell_tag(tmp) == TrieVar)
                    tmp = (Cell) (&array[trievar_val(tmp)]);
                new_heap_node(hreg, tmp);
                /* Constant - 1 cell */
                new_heap_node(hreg, list->con.constant);
	    }
	    
	    CL_insert_constraint_struct(
		(CPtr) constraint_location, 
		&array[trievar_val(list->con.variable_1)]);
	}else{
	    xsb_abort("CL_return_answer_hlp: Constraint broken!\n");
	}
	
	if(isinteger(list->con.variable_2))
	{
	    /* nothing to do */
	}else if(cell_tag(list->con.variable_2) == TrieVar)
	{
            sym =
                (Pair)insert(
                    ((list->con.type == CST_less_than)?LESS_THAN_NAME:
                     ((list->con.type == CST_less_than_or_equal)?
                      LESS_THAN_OR_EQUAL_NAME:
                      ((list->con.type == CST_equal)?EQUAL_NAME:
		       ((list->con.type == CST_not_equal)?NOT_EQUAL_NAME:
			((list->con.type == CST_more_than_or_equal)?
			 MORE_THAN_OR_EQUAL_NAME:
			 MORE_THAN_NAME))))),
                    2,
                    (Psc)flags[CURRENT_MODULE],
                    &new_indicator);

            constraint_location = (Cell) hreg;
            bld_struct(hreg, hreg + 1);
            hreg++;
            new_heap_functor(hreg, sym->psc_ptr);
            /* Variable 1 - 1 cell */
            tmp = list->con.variable_1;
            if(cell_tag(tmp) == TrieVar)
                tmp = (Cell) (&array[trievar_val(tmp)]);
            new_heap_node(hreg, tmp);

            if(list->con.constant == makeint(0))
            {
                /* Variable 2 - 1 cell */
                tmp = list->con.variable_2;
                if(cell_tag(tmp) == TrieVar)
                    tmp = (Cell) (&array[trievar_val(tmp)]);
                new_heap_node(hreg, tmp);
            }else if(list->con.variable_2 == makeint(0))
            {
                /* Constant - 1 cell */
                new_heap_node(hreg, list->con.constant);
            }else{
                /* +/2 structure - 3 cells */
                bld_struct(hreg, hreg + 1);
                hreg++;
		sym =
		    (Pair)insert(
			"+",
			2,
			(Psc)flags[CURRENT_MODULE],
			&new_indicator);
                new_heap_functor(hreg, sym->psc_ptr);
                /* Variable 2 - 1 cell */
                tmp = list->con.variable_2;
                if(cell_tag(tmp) == TrieVar)
                    tmp = (Cell) (&array[trievar_val(tmp)]);
                new_heap_node(hreg, tmp);
                /* Constant - 1 cell */
                new_heap_node(hreg, list->con.constant);
            }
	    
            CL_insert_constraint_struct(
                (CPtr) constraint_location,
                &array[trievar_val(list->con.variable_2)]);
	}else{
	    xsb_abort("CL_return_answer_hlp: Constraint broken!\n");
	}
	
	list = list->next;
    }
#ifdef CST_DEBUG_BUILD_CONSTRAINT
    dbg;fprintf(stderr,"CL_return_answer_hlp done\n");
#endif
}

static long CL_max_trievar_index(CL list)
{
    long max, tmp;
    
    max = -1;
    while(list != NULL)
    {
	tmp = -1;
	if(cell_tag(list->con.variable_1) == TrieVar)
	    tmp = trievar_val(list->con.variable_1);
	if(tmp > max)
	    max = tmp;
	if(cell_tag(list->con.variable_2) == TrieVar)
	    tmp = trievar_val(list->con.variable_2);
	if(tmp > max)
	    max = tmp;
	
	list = list->next;
    }
    return max;
}

static long CST_max_trievar_index(CST tree)
{
    long max, tmp, skipnode;
    
    max = -1;
    skipnode = 0;
    while(tree != NULL)
    {
	if(skipnode == 0)
	{
	    tmp = CL_max_trievar_index(tree->constraints);
	    if(tmp > max)
		max = tmp;
	}
	if(!CST_is_root_node(tree))
	{
	    if(tree->parent->right == tree)
		skipnode = 1;
	    else
		skipnode = 0;
	    tree = tree->parent;
	}else
	    tree = NULL;
    }
    return max;
}

void CST_return_answer(CST node, CPtr array, long arraysize, int goes_up)
{
    long maxindex, i, numvars;
    CST  tmpnode, tmpchild;
    CPtr newarray;
    Cell oldvar;
    
    maxindex = CST_max_trievar_index(node);
    
#ifdef CST_DEBUG_BUILD
    dbg;fprintf(stderr,"CST_return_answer called\n");
    dbg;fprintf(stderr,"node = %p\n", node);
    dbg;fprintf(stderr,"last inserted node = %p\n", last_inserted_node);
/*    dbg;fprintf(stderr,"tree = %p\n", last_inserted_tree);*/
    dbg;fprintf(stderr,"node list = %p\n", node->constraints);
    dbg;fprintf(stderr,"node list = "); CL_print(node->constraints, stderr);
    dbg;fprintf(stderr,"hreg = %p\n", hreg);
    dbg;fprintf(stderr,"answer template = %p\n", array);
    dbg;fprintf(stderr,"answer template size = %ld\n", arraysize);
    dbg;fprintf(stderr,"max trievar index = %ld\n", maxindex);
    dbg;
    if(goes_up == 1)
	fprintf(stderr,"Answer template goes up\n");
    else
	fprintf(stderr,"Answer template goes down\n");
    /*print_all_stacks(NULL);*/
#endif
    
    if(node == NULL)
	return;

    newarray = hreg;
    numvars = 0;
    /* create the arguments */
    for(i = 0; i < arraysize; i++)
    {
	oldvar = (Cell) (&array[(goes_up==1?-i:i)]);
#ifdef CST_DEBUG_BUILD
	dbg;fprintf(stderr, "dereferencing: ");
	{
	    Cell newvar;
	    newvar = oldvar;
	    fprintf(stderr, "%p ", (char *) newvar);
	    while(newvar != NULL && ((newvar & 0x3) == 0) && (*((CPtr) newvar)) != newvar)
	    {
		newvar = *((CPtr) newvar);
		fprintf(stderr, "%p ", (char *) newvar);
	    }
	    fprintf(stderr, "\n");
	}
#endif
	deref(oldvar);
#ifdef CST_DEBUG_BUILD
	dbg;fprintf(stderr, "done dereferencing, result = %p\n", (CPtr) (oldvar));
#endif
	if(i < arraysize && (isref(oldvar) || isattv(oldvar)))
	{
	    if(isattv(oldvar))
	    {
#ifdef CST_DEBUG_BUILD
		dbg;fprintf(stderr,"ATTV found (%ld), bound to %p\n", i, hreg);
#endif
		bind_ref(clref_val(oldvar), hreg);
		/* create a new variable on the heap */
		new_heap_free(hreg);
		numvars++;
	    }else if(isref(oldvar))
	    {
#ifdef CST_DEBUG_BUILD
		dbg;fprintf(stderr,
			    "Variable found (%ld), bound to %p\n", 
			    i, 
			    hreg);
#endif
		bind_ref((CPtr) oldvar, hreg);
		/* create a new variable on the heap */
		new_heap_free(hreg);
		numvars++;
	    }
	}
    }
    if(maxindex == -1)
	return;
#ifdef CST_DEBUG_BUILD
    dbg;fprintf(stderr,
		"array contained %ld variables. Creating %ld new ones\n", 
		numvars, 
		maxindex - (numvars - 1));
#endif
    for(;numvars <= maxindex; numvars++)
    {
#ifdef CST_DEBUG_BUILD
	dbg;fprintf(stderr, "adding variable to array (hreg = %p)\n", hreg);
#endif
	new_heap_free(hreg);
    }
    array = newarray;
#ifdef CST_DEBUG_BUILD
    dbg;fprintf(stderr,"new array at %p\n", newarray);
    dbg;fprintf(stderr,"returning answer...\n");
#endif
    tmpnode = node;
    tmpchild = node->left;
    while(tmpnode != NULL)
    {
#ifdef CST_DEBUG_BUILD
	dbg;fprintf(stderr,"tmpchild = %p\n", tmpchild);
	dbg;fprintf(stderr,"tmpnode = %p\n", tmpnode);
	dbg;fprintf(stderr,"tmpnode->parent = %p\n", tmpnode->parent);
#endif
	if(tmpchild == tmpnode->left)
	    CL_return_answer_hlp(tmpnode->constraints, array);
	
	tmpchild = tmpnode;
	if(!CST_is_root_node(tmpnode))
	    tmpnode = tmpnode->parent;
	else
	    tmpnode = NULL;
    }
#ifdef CST_DEBUG_BUILD
    dbg;fprintf(stderr,"answer returning complete, finishing up\n");
#endif

    for(i = 0; i <= maxindex; i++)
    {
	Cell attv_location;
	Pair sym;
	int  new_indicator;
	
	if(isattv(array[i]))
	{
	    sym = 
		(Pair)insert(
		    "v", 
		    2, 
		    (Psc)flags[CURRENT_MODULE],
		    &new_indicator);
	    
	    attv_location = (Cell) hreg;
	    new_heap_free(hreg);
	    bld_struct(hreg, hreg + 1);
	    hreg++;
	    new_heap_functor(hreg, sym->psc_ptr); 
	    new_heap_num(hreg, 1);
	    new_heap_node(hreg, (Cell) get_attv_attribute(array[i]));
#ifdef CST_DEBUG_BUILD
	    dbg;fprintf(stderr, 
			"variable %ld (%p / %p / %p), attribute: ", 
			i, 
			attv_location, 
			&(array[i]), 
			dec_addr(array[i]));
	    printterm((Cell) (((CPtr)attv_location) + 1), 7, 7);
	    fprintf(stderr, "\n");
#endif
	    array[i] = makeattv(attv_location);
	}
    }
#ifdef CST_DEBUG_BUILD
    dbg;fprintf(stderr,"CST_return_answer done\n");
#endif
}

CPtr cst_last_answer_return_array;
long cst_last_answer_return_arraysize;
int  cst_last_answer_return_array_goes_up;

void CST_setup_answer_return(CPtr array, long arraysize, int goes_up)
{
    cst_last_answer_return_array = array;
    cst_last_answer_return_arraysize = arraysize;
    cst_last_answer_return_array_goes_up = goes_up;
}

Cell CST_get_vars_and_attributes(CST node)
{
    Pair sym;
    int  new_indicator;
    long count_of_variables;
    CPtr struct_pos;
    long i;
    Cell retval;
    
    count_of_variables = CST_max_trievar_index(node) + 1;
    if(count_of_variables == -1)
	return makeint(0);
    sym = 
	(Pair)insert(
	    "con", 
	    count_of_variables*2, 
	    (Psc)flags[CURRENT_MODULE],
	    &new_indicator);
    /* bind variable to struct */
    retval = (Cell) hreg;
    /* construct reference to struct on heap */
    bld_struct(hreg, hreg + 1);
    hreg++;
    new_heap_functor(hreg, sym->psc_ptr);
    /* remember the position of the variables */
    struct_pos = hreg;
    /* Place the correct amount of unbound variables on the heap */
    for(i = 0; i < count_of_variables*2; i++)
    {
	new_heap_free(hreg);
    }
    /* let CST_return_answer fill in the variables... */
    CST_return_answer(node, struct_pos, count_of_variables, 0);
    /* find each variable's attribute, and put it in the last half of
       the struct */
    for(i = 0; i < count_of_variables; i++)
    {
        Cell var;
	var = (Cell) (struct_pos + i);
	deref(var);
	if(isattv(var)) /* dummy check */
	    *(struct_pos + count_of_variables + i) = 
		(Cell) get_attv_attribute(var);
    }
    return retval;
}

/*static void CST_return_answer_last(CST node)
{
    last_inserted_node = node;
    last_inserted_tree = node;
    CST_return_answer(last_answer_return_array, 
		      last_answer_return_arraysize, 
		      last_answer_return_array_goes_up);
}*/

BTNptr CST_owning_trie_node(CST tree)
{
    while(!CST_is_root_node(tree))
	tree = tree->parent;
    return CST_get_parent_trie(tree);
}

void CST_delete_branch(CST tree, BTNptr *root_of_trie)
{
    tree->auxdata = NULL;
    while(tree->left == NULL &&
	  tree->right == NULL &&
	  tree->auxdata == NULL)
    {
	if(CST_is_root_node(tree))
	{
	    BTNptr parent_trie;
	    
	    parent_trie = CST_get_parent_trie(tree);
	    CL_release(tree->constraints);
	    CST_release_node(tree);
	    delete_branch(parent_trie, root_of_trie);
	    return;
	}else{
	    CST    parent_cst;
	    
	    parent_cst = tree->parent;
	    if(parent_cst->left == tree)
		parent_cst->left = NULL;
	    else if(parent_cst->right == tree)
		parent_cst->right = NULL;
	    CL_release(tree->constraints);
	    CST_release_node(tree);
	    tree = parent_cst;
	}
    }
}

void CST_safe_delete_branch(CST tree)
{
    tree->auxdata = NULL;
    CST_fix_opcode(tree);
}
#endif
