/* File:      cst.h
** Author(s): Daniel Aldman, Kostis Sagonas
** Contact:   kostis@csd.uu.se
**
** Copyright (C) Daniel Aldman, Kostis Sagonas, Uppsala University, 2003
**
** XXX is free software; you can redistribute it and/or modify it under the
** terms of the GNU Library General Public License as published by the Free
** Software Foundation; either version 2 of the License, or (at your option)
** any later version.
**
** XXX is distributed in the hope that it will be useful, but WITHOUT ANY
** WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
** FOR A PARTICULAR PURPOSE.  See the GNU Library General Public License for
** more details.
**
** You should have received a copy of the GNU Library General Public License
** along with XXX; if not, write to the Free Software Foundation,
** Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
**
** $Id: cst.h,v 1.15 2003/09/05 15:19:48 daal8963 Exp $
**
*/

#ifndef __CST_H__
#define __CST_H__

#if defined(CST_SUPPORT)

/*

CST constraint representation

This data structure is used to represent an equation of the types:

variable 1 < variable 2 + constant
variable 1 <= variable 2 + constant
variable 1 = variable 2 + constant
variable 1 >= variable 2 + constant
variable 1 > variable 2 + constant

*/

typedef struct {
    Cell variable_1, variable_2,  constant;
    enum {
	CST_less_than,
	CST_less_than_or_equal,
	CST_equal,
	CST_not_equal,
	CST_more_than_or_equal,
	CST_more_than
    } type;
} CST_constraint;

/* Constraint list representation */

typedef struct CL_node *CL;

/* Constraint list node representation */

struct CL_node {
    CST_constraint con;            /* The constraint */
    char           AVL_balance;    /* Used internally to remove duplicates */
    CL		   next,           /* Next pointer */
                   AVL_right;      /* Used internally to remove duplicates */
};

/* 
   This function inserts a new node between the previous and next
   nodes in a list.  To specify that there are no nodes before or
   after this node, set that to NULL.
*/
CL CL_insert_node(CST_constraint con, CL next);

/* 
   This function removes a node from it's list. It returns the next
   element in the list 
*/
CL CL_remove_node(CL node);

/* This function removes an entire list. */
void CL_release(CL list);

/*
  This function finds the given attributed variable's constraints and
  inserts them into the CL.
*/
extern CL CL_insert_constraints(Cell attv,
				CL cons,
				int *trievar_ctr,
				int *trieattv_ctr,
				CPtr **trail_top,
				int trail_goes_up);

/*
  This function removes all duplicate constraints from a CL.
*/
extern CL CL_remove_duplicates(CL constraints);

/* Constraint search tree representation */

typedef struct CST_node *CST;

/* Constraint search tree node representation */

struct CST_node {
#ifdef THREADED
    Cell instr_addr;
#endif
    byte instr;
#ifdef BITS64
    byte padding[7];
#else
    byte padding[3];
#endif
    CL   constraints;
    void *auxdata;
    int  is_in_answer_trie;
    CST	 parent, left, right;
};

/* This function makes sure that the node has the right opcode in it. */
void CST_fix_opcode(CST node);


/* Macros to access subgoal frames pointed out by call tries */
#define CST_get_SGF(cst) ((SGFrame)((cst)->auxdata))
#define CST_set_SGF(cst, frame) \
((cst)->auxdata = (void *) (frame), CST_fix_opcode(cst))

#define CST_get_ALN(cst) ((ALNptr)((cst)->auxdata))
#define CST_set_ALN(cst, aln) \
((cst)->auxdata = (void *) (aln), CST_fix_opcode(cst))

#include "tries.h"

#define CST_set_parent_trie(cst,trie) \
    (cst)->parent = (CST) (((long)(trie)) | 1)
#define CST_get_parent_trie(cst) \
    (BTNptr) (((long)((cst)->parent)) ^ 1)
#define CST_is_root_node(cst) ((((long)(cst)->parent) & 1) == 1)

/* 

A note on the auxdata field:

The auxdata field of the CST_node structure is in part a flag field,
in part a pointer to a AS_info structure/SGFrame structure/Delay list
element. When used with call tries, the auxdata field is set to point
to the SGFrame structure representing the call, when used with
incomplete answer tries, the auxdata field is set to point to the
AS_info structure representing the answer, when used with completed
answer tries, the auxdata field may be set to a delay list element.

*/

#define is_start_node(t) ((t)->auxdata != NULL)

/* 
   This function returns a new CST node, initialized with the given values.
*/
/*extern CST CST_create_node(CL   cons, 
			   void *auxdata, 
			   CST  parent, 
			   CST  left, 
			   CST  right);
*/
/*
  This function takes care of the memory used by an entire CST tree
  when it is no longer used. Do not call this function twice with the
  same tree.
*/
extern void CST_release_tree(CST tree);

/*
  A tuple. Used by CST_insert_constraint - see that function's
  explanation on how it's used.
*/
typedef struct {
    CST tree1, tree2;
    int b;
} CST_2_bool_tuple;

/*
  Given a CL, this function inserts the constraints into the CST,
  rebuilding the CST if neccessary.

  On return, the tuple contains three values: tree1 holds the root
  node of the tree, and, iff an insertion has been made, tree2 holds
  the node that represents the constraint. The b field is true iff an
  insertion has been made. 

  The CL is made a part of the CST, and should not be used after a
  call to this function.
*/
extern CST_2_bool_tuple CST_insert_constraints(CL constraints, CST tree);

/*
  Used to save some variables used by the CST system for returning answers.
*/
void CST_setup_answer_return(CPtr array, long arraysize, int goes_up);

/*
  Returns answers from a CST node.
*/
void CST_return_answer(CST node, CPtr array, long arraysize, int goes_up);

/*
  Given a CST and a CL, this function finds the node in the CST that
  represents the constraints in the CL.

  Iff there is a node that directly corresponds to the constraints in
  the CL, this function returns a pointer to that node, otherwise the
  function returns NULL.

  The CL is destroyed by this function.
*/
extern CST CST_find_node(CST tree, CL constraints);

/* 
   This function reports the space usage of the CST module.
*/
extern void CST_statistics(unsigned long *cst_space, 
			   unsigned long *cst_unused_space,
			   unsigned long *cl_space,
			   unsigned long *cl_unused_space);

/* 
   This function, given a node in a CST will return the trie node that
   owns the CST.
*/
BTNptr CST_owning_trie_node(CST tree);

/*
  This function does the same thing for CSTs as delete_branch does for
  tries.
*/
void CST_delete_branch(CST tree, BTNptr *root_of_trie);

/*
  This function does the same thing for CSTs as safe_delete_branch
  does for tries.
*/
void CST_safe_delete_branch(CST tree);

/*
  This function releases all memory taken up by the CST system.
*/
void CST_release_resources(void);

/*
  This function, given a valid CST node, will construct a structure on
  the heap containing all variables in that answer which have
  attributes and their attributes. Used in get_lastnode_cs_retskel
  builtin.
*/
Cell CST_get_vars_and_attributes(CST node);

#endif
#endif
