/* File:      error_xsb.c
** Author(s): Sagonas, Demoen
** Contact:   xsb-contact@cs.sunysb.edu
** 
** Copyright (C) The Research Foundation of SUNY, 1986, 1993-1998
** 
** XSB is free software; you can redistribute it and/or modify it under the
** terms of the GNU Library General Public License as published by the Free
** Software Foundation; either version 2 of the License, or (at your option)
** any later version.
** 
** XSB is distributed in the hope that it will be useful, but WITHOUT ANY
** WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
** FOR A PARTICULAR PURPOSE.  See the GNU Library General Public License for
** more details.
** 
** You should have received a copy of the GNU Library General Public License
** along with XSB; if not, write to the Free Software Foundation,
** Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
**
** $Id: error_xsb.c,v 1.11 2004/09/29 22:30:27 kostis Exp $
** 
*/

#include "xxx_config.h"

#include <stdio.h>
#include <stdarg.h>
#include <string.h>
#include <signal.h>

#include "auxlry.h"
#include "cell_xsb.h"
#include "psc_xsb.h"
#include "subp.h"
#include "error_xsb.h"
#include "io_builtins_xsb.h"
#include "cinterf.h"
#include "macro_xsb.h"
#include "choice.h"
#include "inst_xsb.h"
#include "register.h"
#include "chat.h"
#include "xxx_inst_defs.h"	/* needed for some instruction sizes */

FILE *stdmsg;	     	     	/* stream for benign messages */
FILE *stddbg;	     	     	/* stream for debug msgs */
FILE *stdwarn;	     	     	/* stream for warnings */
FILE *stdfdbk;	     	     	/* stream for feedback messages */

/*----------------------------------------------------------------------*/

static char *err_msg[] = {
	"Calculation", "Database", "Evaluation", "Implementation",
	"Instantiation", "I/O Control", "I/O End-of-file", "I/O Formatting",
	"Operator", "Overflow", "Range", "Syntax", "Type",
	"Undefined predicate/function", "Undefined value",
	"Underflow", "Zero division", "Domain" };

/*----------------------------------------------------------------------*/

jmp_buf xxx_abort_fallback_environment;

/*----------------------------------------------------------------------*/

/* you can pass either 1 argument---a full description (a string),
   or a variable number of arguments -- a format followed by arguments.
*/
void xxx_abort(char *description, ...)
{
  char message[MAXBUFSIZE];
  va_list args;

  va_start(args, description);

  strcpy(message, "++Error: ");
  vsprintf(message+strlen(message), description, args);
  if (message[strlen(message)-1] != '\n')
    strcat(message, "\n");

  va_end(args);
  pcreg = exception_handler(message);

  /* this allows xxx_abort to jump out even from nested loops */
  longjmp(xxx_abort_fallback_environment, (Integer) pcreg);
}

void xxx_bug(char *description, ...)
{
  char message[MAXBUFSIZE];
  va_list args;

  va_start(args, description);

  strcpy(message, "++XXX bug: ");
  vsprintf(message+strlen(message), description, args);
  if (message[strlen(message)-1] != '\n')
    strcat(message, "\n");

  va_end(args);
  pcreg = exception_handler(message);

  /* this allows xxx_abort to jump out even from nested loops */
  longjmp(xxx_abort_fallback_environment, (Integer) pcreg);
}

/*----------------------------------------------------------------------*/

void xxx_arithmetic_abort(Cell op1, char *OP, Cell op2, char *what_expected)
{
  static vstrDEFINE(str_op1);
  static vstrDEFINE(str_op2);

  print_pterm(op1, TRUE, &str_op1);
  print_pterm(op2, TRUE, &str_op2);
  if (isref(op1) || isref(op2)) {
    xxx_abort("Uninstantiated argument of evaluable arithmetic function %s/2\n%s %s %s %s%s",
	      OP, "   Goal:",
	      (isref(op1) ? "_Var": str_op1.string),
	      OP,
	      (isref(op2) ? "_Var": str_op2.string),
	      ", probably as 2nd arg of is/2");
  }
  else {
    xxx_abort("Wrong domain in evaluable arithmetic function %s/2\n         %s %s %s %s %s found",
	      OP, what_expected, "expected, but",
	      str_op1.string, OP, str_op2.string);
  }
}

void xxx_arithmetic_abort1(char *OP, Cell op)
{
  static vstrDEFINE(str_op);
  
  vstrSET(&str_op,"_Var");
  if (! isref(op)) print_pterm(op, TRUE, &str_op);
  xxx_abort("%s evaluable arithmetic function %s/2\n%s %s(%s) %s",
	    (isref(op) ? "Uninstantiated argument of" : "Wrong domain in"),
	    OP, "   Goal:", OP, str_op.string,
	    ", probably as 2nd arg of is/2");  
}

void xxx_arithmetic_comp_abort(Cell op1, char *OP, int op2)
{
  static vstrDEFINE(str_op1);

  vstrSET(&str_op1,"_Var");
  if (! isref(op1)) print_pterm(op1, TRUE, &str_op1);
  xxx_abort("%s arithmetic comparison %s/2\n%s %s %s %d",
	    (isref(op1) ? "Uninstantiated argument of" : "Wrong type in"),
	    OP, "   Goal:", str_op1.string, OP, op2);
}

/*----------------------------------------------------------------------*/

/* this is a soft type of error msg compared to xxx_abort. It doesn't
   abort the computation, but sends stuff to stderr */
void xxx_error(char *description, ...)
{
  va_list args;

  va_start(args, description);
  fprintf(stderr, "\n++Error: ");
  vfprintf(stderr, description, args);
  va_end(args);
  fprintf(stderr, "\n");
}

void xxx_warn(char *description, ...)
{
  va_list args;

  va_start(args, description);
  fprintf(stdwarn, "\n++Warning: ");
  vfprintf(stdwarn, description, args);
  va_end(args);
  fprintf(stdwarn, "\n");
}

void xxx_mesg(char *description, ...)
{
  va_list args;

  va_start(args, description);
  vfprintf(stdmsg, description, args);
  va_end(args);
  fprintf(stdmsg, "\n");
}

void xxx_dbgmsg(char *description, ...)
{
  va_list args;

  va_start(args, description);
  vfprintf(stddbg, description, args);
  va_end(args);
  fprintf(stddbg, "\n");
}

/*----------------------------------------------------------------------*/

void xxx_exit(char *description, ...)
{
  va_list args;

  va_start(args, description);
  vfprintf(stderr, description, args);
  va_end(args);

  fprintf(stdfdbk, "\nExiting XXX abnormally...\n");
  exit(1);
}

/*----------------------------------------------------------------------*/

void xxx_err_handle(int description, int arg, const char *f,
		    int ar, const char *expected, Cell found)
{
  char message[160];	/* Allow 2 lines of error reporting.	*/
  
  switch (description) {
  case INSTANTIATION_ERROR:
    sprintf(message, 
	    "! %s error in argument %d of %s/%d\n",
	    err_msg[description], arg, f, ar);
    break;
  case RANGE_ERROR:	/* I assume expected != NULL */
    sprintf
      (message,
       "! %s error: in argument %d of %s/%d\n! %s expected, but %d found\n",
       err_msg[description], arg, f, 
       ar, expected, (int) int_val(found));
    break;
  case TYPE_ERROR:
  case DOMAIN_ERROR:
    if (expected == NULL) {
      sprintf(message, 
	      "! %s error in argument %d of %s/%d\n",
	      err_msg[description], arg, f, ar);
    } else  
      sprintf
	(message,
	 "! %s error: in argument %d of %s/%d\n! %s expected, but %s found\n",
	 err_msg[description], arg, f, ar, expected,
	 "something else");
    break;
  case ZERO_DIVIDE_ERROR:
    sprintf(message,
	    "! %s error in %s\n! %s expected, but %lx found\n",
	    err_msg[description], f, expected, found);
    break;
  default:
    sprintf(message, 
	    "! %s error (not completely handled yet)\n",
	    err_msg[description]);
    break;
  }
  pcreg = exception_handler(message);
}

/*----------------------------------------------------------------------*/

void check_and_clean_table_cut(CPtr cp)
{
  register byte instr_cut_over;

  instr_cut_over = get_opcode(cp_pcreg(cp));
  switch (instr_cut_over) {
  case check_complete:
  case resume_compl_suspension:
  case answer_return:
  case tabletrust:
  case tableretry:
    if (!is_completed(tcp_subgoal_ptr(breg))) {
      Psc psc = TIF_PSC(subg_tif_ptr(tcp_subgoal_ptr(breg)));
      Psc call_psc = *(*((Psc **)ereg-1)-1);
      xxx_abort("Illegal cut over a tabled predicate: %s/%d\n%s %s/%d\n",
		get_name(psc), get_arity(psc),
		"\t  from within a call to",
		get_name(call_psc), get_arity(call_psc));
    }
    break;
  default:
    break;
  }
  if (instr_cut_over == resume_compl_suspension)
    chat_free_cp_compl_susp_chat_areas((ComplSuspFrame)breg);
}

/*************************************************************************/
/*
   Builtins for exception handling using a Prolog-based catch-throw

              $$set_scope_marker/0
              $$unwind_stack/0
              $$clean_up_block/0

   Written by Bart Demoen, after the CW report 98:
              A 20' implementation of catch and throw

   7 Febr 1999
*/

static byte *scope_marker;

int set_scope_marker(void)
{
  /*     printf("%x %x\n",cp_ereg(breg),ereg);*/
   scope_marker = pcreg;
   /* skipping a putpval and a call instruction */
   /* is there a portable way to do this ?      */
   /* instruction builtin has already made pcreg point to the putpval */
   scope_marker += THROWPAD;
   return(TRUE);
} /* set_scope_marker */

/* TLS: slightly modified to abort if throwing over an open table.
   Im not positive if this is right */
int unwind_stack(void)
{
   byte *cp, *cpmark;
   CPtr e,b;

   cpmark = scope_marker;
   /*   printf("sm 2 %d  x%x\n",scope_marker,scope_marker);*/
   /* first find the right environment */
   e = ereg;
   cp = cpreg; /* apparently not pcreg ... maybe not good in general */
   while ( (cp != cpmark) && e )
     {
       /*       printf("cp %d x%x\n",cp,cp);*/
       cp = (byte *)e[-1];
       e = (CPtr)e[0];
     }

   if (!e) xxx_exit("Throw failed because no catcher for throw");

   /* now find the corresponding breg */
   b = breg;
   while (cp_ereg(b) <= e) {
     check_and_clean_table_cut(b);      
     b = cp_prevbreg(b);
   }
   check_and_clean_table_cut(b);
   breg = b;
   return(FALSE);
} /* unwind_stack */

int clean_up_block(void)
{
   if (cp_ereg(breg) > ereg) {
     /*     printf("%x %x\n",cp_ereg(breg),ereg); */
     check_and_clean_table_cut(breg);      
     breg = (CPtr)cp_prevbreg(breg);
   }
   return(TRUE);
} /* clean_up_block */

/*---------------------------- end of error_xsb.c --------------------------*/
