/* File:      loader_xsb.c
** Author(s): David S. Warren, Jiyang Xu, Terrance Swift, Kostis Sagonas
** Contact:   xsb-contact@cs.sunysb.edu
** 
** Copyright (C) Uppsala University, 2002
** Copyright (C) The Research Foundation of SUNY, 1986, 1993-1998
** Copyright (C) ECRC, Germany, 1990
** 
** XXX is free software; you can redistribute it and/or modify it under the
** terms of the GNU Library General Public License as published by the Free
** Software Foundation; either version 2 of the License, or (at your option)
** any later version.
** 
** XXX is distributed in the hope that it will be useful, but WITHOUT ANY
** WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
** FOR A PARTICULAR PURPOSE.  See the GNU Library General Public License for
** more details.
** 
** You should have received a copy of the GNU Library General Public License
** along with XXX; if not, write to the Free Software Foundation,
** Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
**
** $Id: loader_xsb.c,v 1.40 2006/08/21 11:26:13 kostis Exp $
** 
*/


/************************************************************************/
/*
	This file contains routines for loading a byte code
	file into the emulator's permanent work space (pspace).
*/
/************************************************************************/

#include "xxx_config.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "auxlry.h"
#include "psc_xsb.h"
#include "psc_defs.h"
#include "cell_xsb.h"
#include "heap_xsb.h"
#include "flags_xsb.h"
#include "macro_xsb.h"
#include "loader_xsb.h"
#include "loader_defs.h"
#include "error_xsb.h"
#include "inst_xsb.h"
#include "xxx_inst_defs.h"
#include "xxx_inst_sizes.h"
#include "memory_xsb.h"
#include "register.h"
#include "varstring.h"

#ifdef FOREIGN
#include "dynload.h"
#endif

/* === macros =========================================================	*/

#define store_ptrpsc(I_ADDR)  (cell(I_ADDR) = *reloc_table[cell(I_ADDR)])

#define store_pscname(I_ADDR) \
    cell(I_ADDR) = (Cell) get_name((Psc)(*reloc_table[cell(I_ADDR)]))
#define store_tagged_pscname(I_ADDR) \
    cell(I_ADDR) = makeatom(get_name((Psc)(*reloc_table[cell(I_ADDR)])))

/* NOTE: the following macro assumes that try/retry/trust instructions
         all have the same size; size is expressed in number of words */
#define TRY_RETRY_TRUST_SIZE  (try_sizeof/sizeof(Cell))

#define gentry(OPCODE, ARG1, ARG2, EP) {  	   \
    set_opcode((EP), (OPCODE));			   \
    *((byte *)(EP)+try_arg_1) = (ARG1);  	   \
    cell((CPtr)((byte *)(EP)+try_arg_2)) = (ARG2); \
    (EP) += TRY_RETRY_TRUST_SIZE;		   \
  }

#define genareg(OPCODE, AREG, SYMBOL, LABEL, EP) { \
    set_opcode((EP), (OPCODE));    		   \
    *((byte *)(EP)+areg_equal_arg_1) = (AREG);	   \
    cell((CPtr)((byte *)(EP)+areg_equal_arg_2)) = (SYMBOL);	 \
    cell((CPtr)((byte *)(EP)+areg_equal_arg_3)) = (Cell)(LABEL); \
    (EP) += (areg_equal_sizeof / sizeof(Cell));	   \
}

/* In the following, Y is the number of bytes we want to read from fd   */
#define get_obj_data(X,Y)	(fread((char *)(X), 1, (Y), fd))

#define get_obj_byte(X)		(get_obj_data((X),1))
#define get_obj_word(X)		(get_obj_data((X),OBJ_WORD_SIZE))
#define get_obj_string(X,LEN)	(get_obj_data((X),(LEN)))

#define get_obj_word_bb(X)	{ get_obj_word(X) ; fix_bb((void *)X) ; }
#define get_obj_word_bbsig(X)	{ get_obj_word(X) ; fix_bb4((void *)X) ; \
			          *(Integer *)((void *)X) = *(int *)((void *)X); }
#define get_tagged_int_bbsig(X)	{ int x; \
			          get_obj_word(X) ; fix_bb4((void *)X) ; \
			          x = *(int *)(X) ; \
			          *(Integer *)(X) = makeint(x) ; }
#define get_obj_word_bbflt(X)	{ get_obj_word(X) ; fix_bb4((void *)X) ; \
			          *(Float *)(X) = *(float *)(X); }

/*----------------------------------------------------------------------*/

/* Number of entries in one "segment" of the index relocation table.
   The table isn't actually segmented, but it is allocated in
   chunks of this size. */
#define NUM_INDEX_BLKS 256

/* === variables also used in other parts of the system =============== */

Psc global_mod;	/* points to "global", whose ep is globallist */

TIFptr first_tip = NULL;

/* === working variables ==============================================	*/

static TIFptr last_tip = NULL;
static word **reloc_table = NULL;
static pseg last_text = NULL;	/* permanent var, chain of text seg */
static pseg current_seg;	/* starting address -- used for relocation */
static CPtr *index_reloc;         	/* index relocation table */
static unsigned int num_index_reloc;    /* number of chunks in index_reloc */
static pindex *index_block_chain;	/* index block chain */

#ifdef THREADED
static unsigned int *orig_inst_offset = NULL;
#endif

/* === return an appropriate hash table size ==========================	*/

unsigned int hsize(const unsigned int numentry)
{
  int i, j, temp;

  if (numentry > 251/*was 16*/) temp = numentry; 
  else temp = 2 * numentry + 1;
  j = temp / 2 + 1;
  for (i = 2; i <= j; i++) {
    if ((i != temp) && ((temp % i) == 0)) { temp++; j = temp/2+1; }
  }
  /* fprintf(stderr, "Numentry = %d, returning = %d\n", numentry, temp); */
  return temp;
}

/* == unload a segment ================================================	*/

void unload_seg(pseg s)
{
  pindex i1, i2 ;
  pseg prev, next ;

  /* free the index blocks */
  i1 = seg_index(s) ;
  while (i1) {
    i2 = i_next(i1) ;
    mem_dealloc((byte *)i1, i_size(i1));
    i1 = i2;
  }
  /* delete segment from segment dllist and deallocate it */
  next = seg_next(s) ;
  prev = seg_prev(s) ;
  if (next) seg_prev(next) = prev ;
  if (prev) seg_next(prev) = next ;
  if (last_text == s) last_text = prev ;
  mem_dealloc((byte *)seg_hdr(s), seg_size(s));
}

/*----------------------------------------------------------------------*/

inline static CPtr reloc_addr(const CPtr base, const Integer offset)
{
  return (CPtr)(((byte *)(base))+(offset*ZOOM_FACTOR));
}

/*-----------------------------------------------------------------------*/

#ifdef THREADED
inline static Integer reloc_offset(const Integer offset,
				   const unsigned int inst_num)
{
  Integer value;
  unsigned int i, low, high;
  
  low = 0;
  high = inst_num - 1;
  while (low <= high) {
    i = (low + high) / 2;
    value = orig_inst_offset[i];
    if (offset < value) high = i - 1;
    else if (offset > value) low = i + 1;
    else break; /* found match */
  }
  return offset + (i * WORD_SIZE);
}
#endif

/*-----------------------------------------------------------------------*/

#ifdef THREADED
static void reloc_threaded(const unsigned int text_bytes,
			   const unsigned int inst_num)
{
  register CPtr inst_addr, first_addr;
  
  inst_addr = first_addr = seg_text(current_seg);

  /*
  inst_addr = seg_text(current_seg);
  end_addr = (CPtr)((byte *)inst_addr + text_size);
  while (inst_addr < end_addr && get_obj_word(inst_addr)) {
  */
  while ((inst_addr-first_addr) * sizeof(Cell) < text_bytes) {
    unsigned int oprand, current_opcode;

    current_opcode = get_opcode(inst_addr);
    inst_addr = cell_after_opcode(inst_addr);

    for (oprand = 1; oprand <= INST_LAST_OP_NUM; oprand++) {
      switch (inst_table[current_opcode][oprand]) {
      case A:
      case V:
      case P:
      case R:
      case X:
      case PP:
      case PPP:
      case RRR:
      case RRV:
      case PPR:
	break;
      case S:
      case C:
      case N:  /* tagged integer number */
      case UN: /* untagged integer number */
      case F:
      case I:
      case T: 
	inst_addr++;
	break;
      case L:
      case E:
	{
	  Integer offset = *inst_addr;
	  if (offset < 0) 
	    *(CPtr *)inst_addr = (CPtr)&fail_inst;
	  else {
	    Integer new_offset;
	    new_offset = reloc_offset(offset, inst_num);
	    *(CPtr *)inst_addr = reloc_addr(first_addr, new_offset);
	  }
	  inst_addr++;
	  break;
	}
      default:
	break;
      } /* switch */
    } /* for */
  }
} /* end of reloc_threaded() */
#endif

/*----------------------------------------------------------------------*/

inline
void inserth(struct hash_space *hspace,
	     const CPtr label, const Cell val) 
{ 
  CPtr temp;
  CPtr cptr;
  long hashval;
  struct hrec *bucket;

  hashval = ihash(val, hspace->size);
  bucket = &hspace->indextab[hashval];
  bucket->l++;
  temp = (CPtr)((void *)&(bucket->link));
  if (bucket->l > 1) {
    temp = (CPtr)*temp;
    while ((CPtr)*temp != temp) 
      temp = (CPtr)*(temp+1);
  }
  cptr = hspace->chaintab_ptr;
  *temp = (Cell)cptr;
  *cptr = (Cell)label; cptr++;
  *cptr = (Cell)cptr;  cptr++;
  hspace->chaintab_ptr = cptr;
}

/*----------------------------------------------------------------------*/

static inline
int hashing_is_discr_default(struct hash_space * pHashSpace)
{
  long lResult = 0;
  unsigned int nIndex = 0;
  unsigned int nUsedBuckets = 0;

  while((nIndex < pHashSpace->size) && (nUsedBuckets < 2))
  {
    if(pHashSpace->indextab[nIndex].l > 0)
    {
      nUsedBuckets++;
    }
    nIndex++;
  }
  return lResult=(nUsedBuckets>1)?1:0;
}

int hashing_is_discriminatory(struct hash_space * pHashSpace)
{
  int iResult = 0;

#if   defined( _HASHING_METHOD_1_ )

#elif defined( _HASHING_METHOD_2_ )

#else
  iResult = hashing_is_discr_default(pHashSpace);
#endif
  return iResult;
}

/*----------------------------------------------------------------------*/

struct hash_space *alloc_hspace(const unsigned int clause_no)
{
  unsigned int j;
  struct hash_space *hspace;
  unsigned int size = hsize(clause_no);
  static struct hrec *indextab;

  hspace = (struct hash_space *)malloc(sizeof(struct hash_space));
  hspace->size = size;
  indextab = (struct hrec *)malloc(size*sizeof(struct hrec));
  /* initialize the index table */
  for (j = 0; j < size; j++) {
    indextab[j].l = 0;
    indextab[j].link = (CPtr)((void *)&(indextab[j].link));
  }
  hspace->indextab = indextab;
  hspace->chaintab = malloc(2*clause_no*sizeof(CPtr));
  hspace->chaintab_ptr = hspace->chaintab;
  return hspace;
}

void free_hspace(struct hash_space *hspace)
{
  free(hspace->chaintab);
  free(hspace->indextab);
  free(hspace);
}

/*----------------------------------------------------------------------*/
/* Severely changed to hash atomic data on TAGGED values; this avoids   */
/* unnecessary untagging on hash_on_???? instructions.                  */
/*----------------------------------------------------------------------*/

static unsigned int get_index_tab(FILE *fd,
				  const unsigned int clause_no,
				  const unsigned int inst_num,
				  struct hash_space *hspace)
{
  unsigned int j, size, count = 0;
  byte type;
  CPtr label;
  Integer ival;
  Cell val;

  size = hspace->size;

  for (j = 0; j < clause_no; j++)
  {
    get_obj_byte(&type);
    switch (type)
      {
      case 'i':
	{
	  get_obj_word_bbsig(&ival);
	  val = makeint(ival);
	  count += ARGLABEL_SIZEOF_C;
	  break;
	}
      case 'l':
	{
	  val = (Cell)(list_str); 
	  count += ARGLABEL_SIZEOF_LorN;
	  break;
	}
      case 'n': 
	{
	  val = makenil; /* CHANGED 24th Dec 1999 - Kostis */
	  count += ARGLABEL_SIZEOF_LorN;
	  break;
	}
      case 'c':
	{
	  get_obj_word_bb(&ival);
	  val = (Cell)ival;
	  count += ARGLABEL_SIZEOF_C;
	  store_tagged_pscname(&val);
	  break;
	}
      case 's':
	{
	  get_obj_word_bb(&ival);
	  val = (Cell)ival;
	  count += ARGLABEL_SIZEOF_S;
	  store_ptrpsc(&val);
	  break;
	}
      default:
	{
	  xxx_exit("%s: Unknown index type %c\n", __FUNCTION__, type);
	  return 0;
	}
      }
    get_obj_word_bbsig(&label);
    if ((Integer)label < 0) label = (CPtr)&fail_inst;
    else {
#ifdef THREADED
      Integer new_label = reloc_offset((Integer)label, inst_num);
      label = reloc_addr(seg_text(current_seg), new_label);
#else
      label = reloc_addr(seg_text(current_seg), (Integer)label);
#endif
    }
    inserth(hspace, label, val);
  }
  return count;
} /* end of get_index_tab() */

/*----------------------------------------------------------------------*/

inline static pindex new_index_seg(const unsigned int no_cells)
{
  pindex new_i;
  unsigned int size;

  size = SIZE_IDX_HDR + (no_cells * sizeof(Cell));
  new_i = (pindex)mem_alloc(size);
  /* initialize fields of new index segment header */
  i_next(new_i) = 0;
  i_size(new_i) = size;
  /* append at tail of block chain */
  *index_block_chain = new_i;
  index_block_chain = &i_next(new_i);
  return new_i;
}

/*----------------------------------------------------------------------*/

inline static pindex new_2nd_level_index_seg(const unsigned int no_cells)
{
  pindex new_i;
  unsigned int size;

  size = SIZE_IDX_HDR + (areg_equal_sizeof * no_cells) + fail_sizeof;
  new_i = (pindex)mem_alloc(size);
  /* initialize fields of new index segment header */
  i_next(new_i) = 0;
  i_size(new_i) = size;
  return new_i;
}

/*----------------------------------------------------------------------*/

static pindex gen_opt_2nd_lev_index(const unsigned int item_no,
				    const CPtr bucket_link)
{
  unsigned int i;
  pindex new_i;
  CPtr   ep1, ep2, temp;

  new_i = new_2nd_level_index_seg(item_no);
  ep1 = ep2 = i_block(new_i);
  temp = bucket_link;
  for (i = 0; i < item_no; i++)
    {
      bool bypass;
      byte *label = (byte *)(*temp);
      byte  points_to_inst = get_opcode(label);

      if (points_to_inst == allocate_gc) {
	bypass = 0;
	label += allocate_gc_sizeof;
	points_to_inst = get_opcode(label);
      }
      else bypass = 1;

      /* check whether the optimization makes sense -- perhaps conservative */
      if (points_to_inst == getcon || points_to_inst == getnumcon ||
	  points_to_inst == getstr || points_to_inst == getnil)
	{
	  byte areg;
	  register byte *ep3;
	  word symbol;

	  /*if (points_to_inst == allocate_gc)
	    symbol = *((word *)(label+allocate_gc_sizeof+getcon_arg_2)); */
	  if (points_to_inst == getnil) symbol = makenil;
	  else symbol = *((word *)(label+getcon_arg_2));

	  for (ep3 = (byte *)ep1+areg_equal_arg_2;
	       ep3 < (byte *)ep2; ep3 += areg_equal_sizeof)
	    if (*((word *)ep3) == symbol) goto abandon_optimization_attempt;
	  
	  if (points_to_inst == getcon || points_to_inst == getnumcon)
	    { /* here the instruction could be bypassed ! */
	      areg = *(label+getcon_arg_1);
	      if (bypass) label += getcon_sizeof;
	    }
	  else
	    if (points_to_inst == getnil)
	      { /* here the instruction could be bypassed ! */
		areg = *(label+getnil_arg_1);
		if (bypass) label += getnil_sizeof;
	      }
	    else
	      areg = 1;
	  genareg(areg_equal, areg, symbol, label, ep2);
#ifdef DEBUG_LOADER
	  fprintf(stderr, "    areg_eq r%d\t0x%lx\t%p\n", areg, symbol, label);
#endif
	}
      else
	{
#ifdef DEBUG_LOADER
	  fprintf(stderr, "--> %s\n", inst_name(points_to_inst));
#endif
	  goto abandon_optimization_attempt;
	}
      ++temp, temp = (CPtr)*(temp);
    }
  set_opcode(ep2, fail);
  *index_block_chain = new_i;
  index_block_chain = &i_next(new_i);
  return new_i;

 abandon_optimization_attempt:
  mem_dealloc((byte *)new_i, i_size(new_i));
  return NULL;
}

/*----------------------------------------------------------------------*/

void gen_index(const byte index_arg,
		      CPtr sob_arg_p,
		      const byte arity,
		      struct hash_space *hspace)
{
  CPtr   ep1;
  pindex new_i;
  struct hrec *indextab;
  unsigned int j, bucket_l, size; 
  bool second_level_opt_possible;
 
  size = hspace->size;
  indextab = hspace->indextab;

  new_i = new_index_seg(size);

  second_level_opt_possible = ((index_arg != 1) ? 0 : 1);

  ep1 = i_block(new_i);
  *(sob_arg_p) = (Cell)ep1;
  for (j = 0; j < size; j++) {
    bucket_l = indextab[j].l;
    if (bucket_l == 0) *(ep1) = (Cell) &fail_inst;
    else if (bucket_l == 1) *(ep1) = *(indextab[j].link);
    else
      if (second_level_opt_possible &&
	  (new_i = gen_opt_2nd_lev_index(bucket_l, indextab[j].link)) != NULL)
	*(ep1) = (Cell)i_block(new_i);
      else {
	register CPtr ep2, temp;

	/* otherwise create try/retry/trust instruction */
	new_i = new_index_seg(TRY_RETRY_TRUST_SIZE * bucket_l);
	ep2 = i_block(new_i);
	*(ep1) = (Cell)ep2;
	temp = indextab[j].link;
	/* fprintf(stderr, "has label = %p\n", *temp); */
	gentry(try, arity, *temp, ep2);
	
	for (temp++; *temp != (Cell)temp; temp++) {
	  temp = (CPtr) *temp;		/* generate "retry" */
	  /* fprintf(stderr, "has label = %p\n", *temp); */
	  gentry(retry, arity, *temp, ep2);
	}
	/* change last "retry" to "trust" */
	set_opcode((ep2-TRY_RETRY_TRUST_SIZE), trust);
#ifdef DEBUG_LOADER
	fprintf(stderr, "\nSecond-level hash of size %d\n", bucket_l);
	for (ep2 = ep2-(TRY_RETRY_TRUST_SIZE*bucket_l); bucket_l > 0 ; bucket_l--) {
	  fprint_inst(stderr, ep2);
	  fprint_inst(stderr, *(cell_after_opcode(ep2)));
	  ep2 += TRY_RETRY_TRUST_SIZE;
	}
#endif/**/
      }
    ep1++;
  }
}

/************************************************************************
*                                                                       *
*  load_text() loads the byte code intruction from a byte code file to	*
*  the byte code program space.  References to indices to the PCS table	*
*  are resolved with the use of the macro st_index.  New index relies   *
*  on the symbol table array which is assigned values by load_sms.	*
*                                                                       *
************************************************************************/

static int load_text(FILE *fd, const unsigned int text_size, byte *index_arg)
{
#ifdef THREADED
  unsigned int i = 0, offset = 0;
#endif
  register CPtr inst_addr, end_addr;

  inst_addr = seg_text(current_seg);
  end_addr = (CPtr)((byte *)inst_addr + text_size);
  while (inst_addr < end_addr && get_obj_word(inst_addr)) {
    int oprand, current_opcode;

#ifdef THREADED
    orig_inst_offset[i++] = offset;
    set_cellinst(inst_addr, *inst_addr);
#endif
    current_opcode = get_opcode(inst_addr);
    inst_addr = cell_after_opcode(inst_addr);
#ifdef THREADED
    offset += inst_sizeof_in_O_file[current_opcode];
#endif
    for (oprand = 1; oprand <= INST_LAST_OP_NUM; oprand++) {
      switch (inst_table[current_opcode][oprand]) {
      case A:
      case V:
      case P:
      case R:
      case X:
      case PP:
      case PPP:
      case RRR:
      case RRV:
	break;
      case PPR:
	if (current_opcode == hash_on_bound ||
	    current_opcode == hash_on_atomic ||
	    current_opcode == hash_on_struct) {
	  *index_arg = *((byte *)(inst_addr)-1);
	}
	break;
      case S:
	get_obj_word_bb(inst_addr);
	store_ptrpsc(inst_addr);
	inst_addr++;
	break;
      case C:
	get_obj_word_bb(inst_addr);
	store_tagged_pscname(inst_addr);
	inst_addr++;
	break;
      case L:  /* label */
      case E:  /* entry point */
	get_obj_word_bbsig(inst_addr);
#ifndef THREADED
	{ /* for not directly threaded code,
	     relocation of offsets can take place immediately */
	  Integer offset = *inst_addr;
	  if (offset < 0)
	    *(CPtr *)inst_addr = (CPtr)&fail_inst;
	  else {
	    *(CPtr *)inst_addr = reloc_addr(seg_text(current_seg), offset);
	  }
	}
#endif
	inst_addr++;
	break;
      case N:  /* tagged integer number */
	get_tagged_int_bbsig(inst_addr);
	inst_addr++;
	break;
      case UN: /* untagged integer number */
	get_obj_word_bbsig(inst_addr);
	inst_addr++;
	break;
      case F:
	get_obj_word_bbflt(inst_addr);
	inst_addr++;
	break;
      case I: /* word-sized argument of an indexing instruction */
	get_obj_word_bb(inst_addr);
	if (oprand == 2) {	/* second operand of hash_on_* and jiti_on_* */
	  if (cell(inst_addr) >= (Cell)(NUM_INDEX_BLKS*num_index_reloc)) {
	    num_index_reloc = (cell(inst_addr)/NUM_INDEX_BLKS)+1;
	    index_reloc = (CPtr *)realloc(index_reloc,NUM_INDEX_BLKS*
					  num_index_reloc*sizeof(CPtr));
	    if (!index_reloc) {
	      xxx_error("Couldn't allocate index relocation space");
	      return FALSE;
	    }
	  }
	  index_reloc[cell(inst_addr)] = inst_addr;
	}
	else {	   /* third operand of hash_on_* and jiti_on_* instructions */
	  if (current_opcode != jiti_on_atomic)
	    *inst_addr = hsize(cell(inst_addr));
	}
	inst_addr++;
	break;
      case T: {
	Cell tab_config_hold;
	static TIFptr tif_ptr; /* need to retain its value */

	if (current_opcode == tabletry || current_opcode == tabletrysingle) {
	  New_TIF(tif_ptr,NULL);
	  if (first_tip == 0) first_tip = tif_ptr;
	  else TIF_NextTIF(last_tip) = tif_ptr;
	  last_tip = tif_ptr;
	}
	get_obj_word(&tab_config_hold);          /* space holder */
	*inst_addr = (Cell)tif_ptr;
	inst_addr++;
	break;
      }
      default:
	break;
      }  /* switch */
    } /* for */
  }
  if (inst_addr != end_addr) {
    xxx_error("inst_addr %p, end_addr %p", (void*)inst_addr, (void*)end_addr);
    return FALSE;
  }
  else return TRUE;
} /* end of load_text() */

/*----------------------------------------------------------------------*/

static void load_index(FILE *fd,
		       const byte index_arg,
		       const int index_bytes,
		       const unsigned int inst_num)
{
  struct  hash_space *hspace;
  Integer index_bno, clause_no, t_len;
  byte    index_inst, arity;
  int     count = 0;
  CPtr    sob_arg_p;

  while (count < index_bytes) {
    get_obj_byte(&index_inst);
    /*fprintf(stderr, "%d, %s", index_inst, "foo"inst_name(index_inst));*/
    get_obj_byte(&arity);
    get_obj_word_bb(&index_bno);
    sob_arg_p = index_reloc[index_bno];
    get_obj_word_bb(&clause_no);
    
    hspace = alloc_hspace(clause_no);
    t_len = get_index_tab(fd, clause_no, inst_num, hspace);
    
    gen_index(index_arg, sob_arg_p, arity, hspace);
    free_hspace(hspace);
    count += CASES_SIZEOF + t_len;
  }
}

/*== the load_seg function =============================================*/

static pseg load_seg(FILE *fd,
		     const unsigned int text_bytes,
		     const unsigned int index_bytes,
		     const unsigned int inst_num)
{
  byte index_arg = 0;
  unsigned int seg_size, text_size;

  text_size = text_bytes * ZOOM_FACTOR;
#ifdef THREADED
  text_size += (inst_num * WORD_SIZE); /* increase due to direct threading */
#endif
  /* Allocate space, including space for segment header */
  seg_size = SIZE_SEG_HDR + text_size;
  current_seg = (pseg)mem_alloc(seg_size);

  /* Allocate first chunk of index_reloc -- do not count it in perm_space */
  index_reloc = (CPtr *)malloc(NUM_INDEX_BLKS*sizeof(CPtr));
  if (!index_reloc) {
    xxx_error("Couldn't allocate index relocation space");
    return NULL;
  }
  num_index_reloc = 1;
#ifdef THREADED
  /* Allocate auxiliary info table for relocation of instruction offsets */
  orig_inst_offset = (unsigned int *)malloc(inst_num*sizeof(unsigned int));
  if (!orig_inst_offset) {
     xxx_error("Couldn't allocate space for offset relocation table");
    return NULL;
  }
#endif

  current_seg++;
  seg_next(current_seg)  = 0;
  seg_prev(current_seg)  = last_text;
  seg_index(current_seg) = 0;
  seg_size(current_seg)  = seg_size;

  if (!load_text(fd, text_size, &index_arg)) {
    mem_dealloc((byte *)seg_hdr(current_seg), seg_size);
    return NULL;
  }
#ifdef THREADED
  reloc_threaded(text_size, inst_num); /* relocation for L and E operands */
#endif
  index_block_chain = &seg_index(current_seg);
  load_index(fd, index_arg, index_bytes, inst_num);
  free(index_reloc);
#ifdef THREADED
  free(orig_inst_offset);
#endif

  /* set text-index segment chain */
  if (last_text) seg_next(last_text) = current_seg;
  last_text = current_seg;
  return current_seg;
}

/************************************************************************/
/*  Routines to check environment consistency.				*/
/************************************************************************/

#define T_NEW 3
#define E_HIDDEN -1
#define E_NOUSE -2

static int env_check[4][5] = {
/*		   T_EXPORT   T_LOCAL   T_IMPORTED   T_IMEX   T_GLOBAL	*/
/*======================================================================*/
/* T_VISIBLE  */ { T_VISIBLE, T_HIDDEN, T_VISIBLE,  E_NOUSE, T_VISIBLE  },
/* T_HIDDEN   */ { T_HIDDEN,  T_HIDDEN,	E_HIDDEN,   E_NOUSE, T_VISIBLE  },
/* T_UNLOADED */ { T_VISIBLE, E_HIDDEN, T_UNLOADED, E_NOUSE, T_UNLOADED },
/* T_NEW      */ { T_VISIBLE, T_HIDDEN, T_UNLOADED, E_NOUSE, T_VISIBLE  }
};

void env_type_set(Psc psc, byte t_env, byte t_type, bool is_new)
{
  int env;
  byte type;

  if (is_new) {
    set_env(psc, env_check[T_NEW][t_env]);
    set_type(psc, t_type);
  } else {
    env = env_check[get_env(psc)][t_env];
    if (env < 0) {
      xxx_error("Environment conflict in the use of %s/%d !", 
		get_name(psc), get_arity(psc));
      /* In the following I am not sure whether setting the environment */
      /* in the presence of an environment conflict error is the right  */
      /* thing to do!  But an "imported_from" vs "local" (non-exported) */
      /* symbol conflict must definitely be resolved in favour of the   */
      /* "local" declaration.						*/
      if (env == E_HIDDEN) {
	if (t_env == T_IMPORTED) 
	  /* Here the psc record of the symbol has already been created */
	  /* by another module that imported (mistakenly) this symbol.  */
	  set_env(psc, T_LOCAL);	
	else { /* We are trying to load a module
		  that imports sth not exported. */
	  Psc mod_psc = get_data(psc);
	  if (mod_psc != NULL)
	    xxx_mesg("         %s/%d is imported from %s but is not exported\n",
		      get_name(psc), get_arity(psc), get_name(mod_psc));
	  else
	    xxx_mesg("         %s/%d is imported from somewhere but is not exported\n",
		      get_name(psc), get_arity(psc));
	}
      }
    }
    else set_env(psc, env);
    type = get_type(psc);
    if (t_type && type && t_type != type) {
      if (t_type==T_UDEF && (type==T_PRED || type==T_DYNA || type==T_FORN)) ;
      else if (t_type==T_FORN && type==T_UDEF) set_type(psc, T_FORN);
      else xxx_error("incompatible types in the use of %s/%d (%x with %x)",
		     get_name(psc), get_arity(psc), type, t_type);
    } else set_type(psc, type | t_type);  
  }
}

/*----------------------------------------------------------------------*/

unsigned int read_magic(FILE *fd)
{
  unsigned long num;

  if (get_obj_word(&num) < 4) return 0;
  fix_bb4((byte *)&num);
  return num;
}

/*----------------------------------------------------------------------*/

inline static void get_obj_atom(FILE *fd, VarString *atom)
{
  byte x;
  unsigned long len;
  
  get_obj_data((&x),1);
  /* ``x'' gets the length of the string or > SHORT_LDOPTIONLEN.
     The latter means we have a long atom.
     In this case, the length is stored in 4 bytes & we use get_obj_word_bb */
  if (x > SHORT_LDOPTIONLEN) { /* handle unusual case specially */
    get_obj_word_bb(&len);
    /* xxx_dbgmsg("get_obj_len = %d... Case is not handled yet!\n",len); */
  } else
    len = x;

  vstrENSURE_SIZE(atom,len+1);
  get_obj_string(atom->string, len);
  atom->length = len;
  vstrNULL_TERMINATE(atom);
}

/*----------------------------------------------------------------------*/

static bool load_one_sym(FILE *fd, Psc cur_mod, unsigned int count, int exp)
{
  static vstrDEFINE(str);
  bool is_new;
  byte t_arity, t_type, t_env;
  Pair temp_pair;
  Psc  mod;

  get_obj_byte(&t_env);
  /* xxx_dbgmsg("%s: t_env = %d\n", __FUNCTION__, t_env); */

  /* this simple check can avoid worse situations in case of compiler bugs */
  if (t_env > T_GLOBAL) 
    xxx_abort("LOADER: The loaded bytecode file %s.O is corrupted",
	      cur_mod->nameptr);

  get_obj_byte(&t_type);
  get_obj_byte(&t_arity);
  get_obj_atom(fd, &str);
  if (t_type == T_MODU)
    temp_pair = insert_module(0, str.string);
  else {
    if (t_env == T_IMPORTED) {
      byte t_modlen;
      char modname[MAXNAME+1];
      get_obj_byte(&t_modlen);
      get_obj_string(modname, t_modlen);
      modname[t_modlen] = '\0';
      temp_pair = insert_module(0, modname);
      mod = temp_pair->psc_ptr;
    }
    else
      if (t_env == T_GLOBAL) mod = global_mod;
      else mod = cur_mod;
    temp_pair = insert(str.string, t_arity, mod, &is_new);
    if (is_new && t_env==T_IMPORTED)
      set_data(temp_pair->psc_ptr, mod);
    /* set psc_data to the psc record of the module name */
    env_type_set(temp_pair->psc_ptr, t_env, t_type, is_new);
    /* dsw added following, maybe wrongly */
    if (exp && t_env == T_EXPORTED) {
      /* xxx_dbgmsg("exporting: %s from: %s",str.string,cur_mod->nameptr); */
      if (is_new) set_data(temp_pair->psc_ptr, mod);
      link_sym(temp_pair->psc_ptr, (Psc)flags[CURRENT_MODULE]);
    }
  }
  if (!temp_pair) return FALSE;
  
  /*
    if (count >= REL_TAB_SIZE) {
      xxx_dbgmsg("Reloc_table overflow");
      return FALSE;
    }
  */
  
  reloc_table[count] = (word *)temp_pair;
  return TRUE;
} /* load_one_sym() */

/************************************************************************
*                                                                       *
* Load_syms is a function which loads a symbol table given in a byte    *
* code file into an appropriate format in the PSC table.  As part of    *
* its function it resolves entry points for byte code intructions (call *
* to relloc_addr), and maintains a tableau so that instructions with    *
* indices into the PSC table may have those indices resolved before     *
* loading them in the instruction array (byte code program space).      *
* The intructions are loaded by a separate function.                    *
*                                                                       *
************************************************************************/

static bool load_syms(FILE *fd, unsigned int psc_count, Psc cur_mod, int exp)
{
  unsigned int i;
  
  reloc_table = (word **)calloc((psc_count), sizeof(word *));
  /* xxx_dbgmsg("reloc_table %p, psc_count %d", reloc_table, psc_count); */

  for (i = 0; i < psc_count; i++) {
    if (!load_one_sym(fd, cur_mod, i, exp)) return FALSE;
  }
  return TRUE;
}

/*----------------------------------------------------------------------*/

static unsigned int magic_num;

static byte *loader1(FILE *fd, int exp)
{
  char name[FOREIGN_NAMELEN], arity;
  byte name_len;
  bool is_new;
  int  seg_count = 0;
  unsigned long psc_count;
  Integer text_bytes, index_bytes, inst_num;
  pseg seg_first_inst, first_inst = NULL;
  Psc  psc, cur_mod;
  Pair pair;
  TIFptr tip;
 
  get_obj_byte(&name_len);

  if (name_len >= FOREIGN_NAMELEN)
    xxx_abort("LOADER: foreign module name is too long");

  get_obj_string(name, name_len);
  if (name_len == 0) cur_mod = global_mod;
  else {
    Pair pair;
    name[name_len] = '\0';
    pair = insert_module(T_MODU, name);
    cur_mod = pair_psc(pair);
  }
  get_obj_word_bb(&psc_count);
  if (!load_syms(fd, psc_count, cur_mod, exp)) 
    return FALSE;
  /*	xxx_dbgmsg("symbol table of module %s loaded", name);	*/
  while (read_magic(fd) == 0x11121306)
    {
      seg_count++;
      /*		xxx_dbgmsg("Seg count: %d",seg_count); */
      /* get the header of the segment */
      get_obj_byte(&arity);
      get_obj_byte(&name_len);
      
      if (name_len >= FOREIGN_NAMELEN)
	xxx_abort("LOADER: module name is too long");

      get_obj_string(name, name_len);
      name[name_len] = '\0';
      get_obj_word_bb(&text_bytes);
      if (magic_num == 0x11121301) get_obj_word_bb(&inst_num);
      get_obj_word_bb(&index_bytes);

      /* load the text-index segment */
      seg_first_inst = load_seg(fd, text_bytes, index_bytes, inst_num);
      if (!seg_first_inst) return FALSE;
      if (seg_count == 1) first_inst = seg_first_inst;
      /* 1st inst of file */
      /* set the entry point of the predicate */
      pair = insert(name, arity, cur_mod, &is_new);
      psc = pair_psc(pair);
      switch (get_type(psc))
	{
	case T_ORDI:
	case T_UDEF:
	  if (strcmp(name, "_$main")!=0) {
	    set_type(psc, T_PRED);
	    set_ep(psc, (byte *)seg_first_inst);
	  }
	  if ((tip = get_tip(psc)) != NULL) { TIF_PSC(tip) = psc; }
	  break;
	case T_PRED:
	  if (strcmp(name, "_$main")!=0) {
	    unload_seg((pseg)get_ep(psc));
	    set_ep(psc, (byte *)seg_first_inst);
	  }
	  if ((tip = get_tip(psc)) != NULL) { TIF_PSC(tip) = psc; }
	  break;
	default:
	  xxx_error("the predicate %s/%d cannot be loaded", name, arity);
	  unload_seg(seg_first_inst);
	  return NULL;
	}
    }
  /*
    xxx_dbgmsg("The first instruction of module %s is %x",
    get_name(cur_mod), first_inst);
  */
  return (byte *)first_inst;
} /* loader1 */

/************************************************************************/
/*  Routine for the foreign language interface.				*/
/************************************************************************/

#ifdef FOREIGN
static byte *loader_foreign(char *filename, FILE *fd, int exp)
{
  byte name_len;
  char name[FOREIGN_NAMELEN];
  static vstrDEFINE(ldoption);
  unsigned long psc_count;
  Psc  cur_mod;
  Pair pair;

  get_obj_byte(&name_len);
  if (name_len >= FOREIGN_NAMELEN) {
    xxx_error("LOADER: foreign module name is too long");
    return FALSE;
  }
  get_obj_string(name, name_len);
  name[name_len] = '\0';
  get_obj_atom(fd, &ldoption);
  pair = insert_module(T_MODU, name);
  cur_mod = pair_psc(pair);
  get_obj_word_bb(&psc_count);
  if (!load_syms(fd, psc_count, cur_mod, exp)) return FALSE;
  return load_obj(filename, cur_mod, ldoption.string);
} /* end of loader_foreign */
#endif

/************************************************************************/
/*									*/
/* Loads the file into permanent space.					*/
/* Data segment first (mixed psc entries and name strings), then text	*/
/* segment.  Under normal circumstances it returns the address of first	*/
/* instruction; if errors occur, it returns NULL.			*/
/*									*/
/************************************************************************/

byte *loader(char *file, int exp)
{
  FILE *fd;	      /* file descriptor */
  byte *first_inst = NULL;

  fd = fopen(file, "rb"); /* "b" needed for DOS. */
  if (!fd) return NULL;
  if (flags[HITRACE]) xxx_mesg("\n     ...... loading file %s", file);
  magic_num = read_magic(fd);

  if (magic_num == 0x11121301 /* || magic_num == 0x11121300 */)
    first_inst = loader1(fd, exp);
  else if (magic_num == 0x11121308) {
#ifdef FOREIGN
    first_inst = loader_foreign(file, fd, exp);
#else
    xxx_abort("Loading a foreign file: %s", file);
#endif
  }
  else {
    xxx_abort("File: %s does not have proper byte code format...\n%s",
	      file, "\t Please recompile it");
    first_inst = NULL;
  }

  fclose(fd);
  if (reloc_table) { free(reloc_table); reloc_table = 0; }
  return first_inst;
} /* loader */

/*----------------------------------------------------------------------*/
