/* File:      string_xsb.c  -- string manipulation stuff
** Author(s): kifer
** Contact:   xsb-contact@cs.sunysb.edu
** 
** Copyright (C) The Research Foundation of SUNY, 1999
** 
** XSB is free software; you can redistribute it and/or modify it under the
** terms of the GNU Library General Public License as published by the Free
** Software Foundation; either version 2 of the License, or (at your option)
** any later version.
** 
** XSB is distributed in the hope that it will be useful, but WITHOUT ANY
** WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
** FOR A PARTICULAR PURPOSE.  See the GNU Library General Public License for
** more details.
** 
** You should have received a copy of the GNU Library General Public License
** along with XSB; if not, write to the Free Software Foundation,
** Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
**
** $Id: string_xsb.c,v 1.6 2003/11/23 22:55:30 kostis Exp $
** 
*/


#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "xxx_config.h"

#include "auxlry.h"
#include "cell_xsb.h"
#include "error_xsb.h"
#include "psc_xsb.h"
#include "hash_xsb.h"
#include "macro_xsb.h"
#include "choice.h"
#include "deref.h"
#include "memory_xsb.h"
#include "heap_xsb.h"
#include "register.h"
#include "flags_xsb.h"
#include "binding.h"
#include "builtin.h"
#include "cinterf.h"
#include "sp_unify_xsb_i.h"
#include "string_xxx.h"

static Cell term, term2;

static vstrDEFINE(input_buffer);
static vstrDEFINE(subst_buf);
static vstrDEFINE(output_buffer);

#include "ptoc_tag_xsb_i.h"


/* R1: +Substring; R2: +String; R3: ?Pos
   Check if Arg1 is a substring of Arg2; unify pos of match with Arg3
*/
bool str_sub(void)
{
  static char *subptr, *stringptr, *matchptr;
  static int substr_pos;

  term = ptoc_tag(1);
  term2 = ptoc_tag(2);
  if (isatom(term) && isatom(term2)) { 
    subptr = atom_val(term);
    stringptr = atom_val(term2);
    matchptr = strstr(stringptr, subptr);
    substr_pos = matchptr-stringptr+1; /* relative pos of substring */
    if (matchptr == NULL)
      return FALSE;
    else {
      return int_unify(makeint(substr_pos), ptoc_tag(3));
    }
  } else return FALSE;
}


bool str_cat(void)
{
  static char *str1, *str2, *tmpstr;

  term = ptoc_tag(1);
  term2 = ptoc_tag(2);
  if (isatom(term) && isatom(term2)) {
    str1 = atom_val(term);
    str2 = atom_val(term2);
    
    tmpstr = (char *)malloc(strlen(str1) + strlen(str2) + 1);
    strcpy(tmpstr, str1);
    strcat(tmpstr, str2);
    str1 = intern_atom(tmpstr);
    free(tmpstr);
    return atom_unify(makeatom(str1), ptoc_tag(3));
  } else return FALSE;
}


/* XSB string substitution entry point
   In: 
       Arg1: string
       Arg2: beginning offset
       Arg3: ending offset. `-' : end of string, -1 : char before last, etc.
   Out:
       Arg4: new (output) string
   Always succeeds, unless error.
*/
bool substring(void)
{
  /* Prolog args are first assigned to these, so we could examine the types
     of these objects to determine if we got strings or atoms. */
  prolog_term input_term, output_term;
  prolog_term beg_offset_term, end_offset_term;
  char *input_string=NULL;    /* string where matches are to be found */
  int beg_offset=0, end_offset=0, input_len=0, substring_len=0;
  int conversion_required=FALSE;

  vstrSET(&output_buffer,"");

  input_term = reg_term(1);  /* Arg1: string to find matches in */
  if (isatom(input_term)) /* check it */
    input_string = atom_val(input_term);
  else if (islist(input_term)) {
    input_string = p_charlist_to_c_string(input_term, &input_buffer,
					  "SUBSTRING", "input string");
    conversion_required = TRUE;
  } else
    xxx_abort("SUBSTRING: Arg 1 (the input string) must be an atom or a character list");

  input_len = strlen(input_string);

  /* arg 2: beginning offset */
  beg_offset_term = reg_term(2);
  if (! isinteger(beg_offset_term))
    xxx_abort("SUBSTRING: Arg 2 (the beginning offset) must be an integer");
  beg_offset = int_val(beg_offset_term);
  if (beg_offset < 0)
    beg_offset = 0;
  else if (beg_offset > input_len)
    beg_offset = input_len;

  /* arg 3: ending offset */
  end_offset_term = reg_term(3);
  if (isref(end_offset_term))
    end_offset = input_len;
  else if (! isinteger(end_offset_term))
    xxx_abort("SUBSTRING: Arg 3 (the end offset) must be integer or _");
  else end_offset = int_val(end_offset_term);

  if (end_offset < 0)
    end_offset = input_len + end_offset;
  else if (end_offset > input_len)
    end_offset = input_len;
  else if (end_offset < beg_offset)
    end_offset = beg_offset;

  output_term = reg_term(4);
  if (! isref(output_term))
    xxx_abort("SUBSTRING: Arg 4 (the output string) must be an unbound variable");

  /* do the actual replacement */
  substring_len = end_offset-beg_offset;
  vstrAPPENDBLK(&output_buffer, input_string+beg_offset, substring_len);
  vstrNULL_TERMINATE(&output_buffer);
  
  /* get result out */
  if (conversion_required)
    c_string_to_p_charlist(output_buffer.string, output_term,
			   "SUBSTRING", "Arg 4");
  else
    c2p_string(output_buffer.string, output_term);
  
  return(TRUE);
}


/* XSB string substitution entry point: replace substrings specified in Arg2
   with strings in Arg3.
   In: 
       Arg1: string
       Arg2: substring specification, a list [s(B1,E1),s(B2,E2),...]
       Arg3: list of replacement string
   Out:
       Arg4: new (output) string
   Always succeeds, unless error.
*/
bool string_substitute(void)
{
  /* Prolog args are first assigned to these, so we could examine the types
     of these objects to determine if we got strings or atoms. */
  prolog_term input_term, output_term;
  prolog_term subst_reg_term, subst_spec_list_term, subst_spec_list_term1;
  prolog_term subst_str_term=(prolog_term)0,
    subst_str_list_term, subst_str_list_term1;
  char *input_string=NULL;    /* string where matches are to be found */
  char *subst_string=NULL;
  prolog_term beg_term, end_term;
  int beg_offset=0, end_offset=0, input_len;
  int last_pos = 0; /* last scanned pos in input string */
  /* the output buffer is made large enough to include the input string and the
     substitution string. */
  int conversion_required=FALSE; /* from C string to Prolog char list */

  vstrSET(&output_buffer,"");

  input_term = reg_term(1);  /* Arg1: string to find matches in */
  if (isatom(input_term)) /* check it */
    input_string = atom_val(input_term);
  else if (islist(input_term)) {
    input_string = p_charlist_to_c_string(input_term, &input_buffer,
					  "STRING_SUBSTITUTE", "input string");
    conversion_required = TRUE;
  } else
    xxx_abort("STRING_SUBSTITUTE: Arg 1 (the input string) must be an atom or a character list");

  input_len = strlen(input_string);

  /* arg 2: substring specification */
  subst_spec_list_term = reg_term(2);
  if (!islist(subst_spec_list_term) && !isnil(subst_spec_list_term))
    xxx_abort("STRING_SUBSTITUTE: Arg 2 must be a list [s(B1,E1),s(B2,E2),...]");

  /* handle substitution string */
  subst_str_list_term = reg_term(3);
  if (! islist(subst_str_list_term))
    xxx_abort("STRING_SUBSTITUTE: Arg 3 must be a list of strings");

  output_term = reg_term(4);
  if (! isref(output_term))
    xxx_abort("STRING_SUBSTITUTE: Arg 4 (the output) must be an unbound variable");

  subst_spec_list_term1 = subst_spec_list_term;
  subst_str_list_term1 = subst_str_list_term;

  if (isnil(subst_spec_list_term1)) {
    vstrSET(&output_buffer, input_string);
    goto EXIT;
  }
  if (isnil(subst_str_list_term1))
    xxx_abort("STRING_SUBSTITUTE: Arg 3 must not be an empty list");

  do {
    subst_reg_term = p2p_car(subst_spec_list_term1);
    subst_spec_list_term1 = p2p_cdr(subst_spec_list_term1);

    if (!isnil(subst_str_list_term1)) {
      subst_str_term = p2p_car(subst_str_list_term1);
      subst_str_list_term1 = p2p_cdr(subst_str_list_term1);

      if (isatom(subst_str_term)) {
	subst_string = atom_val(subst_str_term);
      } else if (islist(subst_str_term)) {
	subst_string = p_charlist_to_c_string(subst_str_term, &subst_buf,
					      "STRING_SUBSTITUTE",
					      "substitution string");
      } else 
	xxx_abort("STRING_SUBSTITUTE: Arg 3 must be a list of strings");
    }

    beg_term = p2p_arg(subst_reg_term,1);
    end_term = p2p_arg(subst_reg_term,2);

    if (!isinteger(beg_term) || !isinteger(end_term))
      xxx_abort("STRING_SUBSTITUTE: Non-integer in Arg 2");
    else{
      beg_offset = int_val(beg_term);
      end_offset = int_val(end_term);
    }
    /* -1 means end of string */
    if (end_offset < 0)
      end_offset = input_len;
    if ((end_offset < beg_offset) || (beg_offset < last_pos))
      xxx_abort("STRING_SUBSTITUTE: Substitution regions in Arg 2 not sorted");

    /* do the actual replacement */
    vstrAPPENDBLK(&output_buffer, input_string+last_pos, beg_offset-last_pos);
    vstrAPPEND(&output_buffer, subst_string);
    
    last_pos = end_offset;

  } while (!isnil(subst_spec_list_term1));

  vstrAPPEND(&output_buffer, input_string+end_offset);

 EXIT:
  /* get result out */
  if (conversion_required)
    c_string_to_p_charlist(output_buffer.string, output_term,
			   "STRING_SUBSTITUTE", "Arg 4");
  else
    c2p_string(output_buffer.string, output_term);
  
  return(TRUE);
}

