/* File:      table_stats.c
** Author(s): Ernie Johnson
** Contact:   xsb-contact@cs.sunysb.edu
** 
** Copyright (C) The Research Foundation of SUNY, 1986, 1993-1998
** 
** XSB is free software; you can redistribute it and/or modify it under the
** terms of the GNU Library General Public License as published by the Free
** Software Foundation; either version 2 of the License, or (at your option)
** any later version.
** 
** XSB is distributed in the hope that it will be useful, but WITHOUT ANY
** WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
** FOR A PARTICULAR PURPOSE.  See the GNU Library General Public License for
** more details.
** 
** You should have received a copy of the GNU Library General Public License
** along with XSB; if not, write to the Free Software Foundation,
** Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
**
** $Id: table_stats.c,v 1.4 2003/11/23 22:56:33 kostis Exp $
** 
*/


#include "xxx_config.h"

#include <stdio.h>

#include "auxlry.h"
#include "cell_xsb.h"
#include "psc_xsb.h"
#include "table_stats.h"
#include "macro_xsb.h"
#include "error_xsb.h"
#include "flags_xsb.h"


/*==========================================================================*/

/*
 *		  T A B L I N G   S T A T I S T I C S
 *		  ===================================
 */


/*
 *                      Recording Current Usage
 *                      -----------------------
 */


NodeStats node_statistics(Structure_Manager *sm)
{
  NodeStats stats;

  SM_CurrentCapacity(*sm, NodeStats_NumBlocks(stats),
		     NodeStats_NumAllocNodes(stats));
  SM_CountFreeStructs(*sm, NodeStats_NumFreeNodes(stats));
  NodeStats_NodeSize(stats) = SM_StructSize(*sm);

  return stats;
}

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

HashStats hash_statistics(Structure_Manager *sm)
{
  HashStats ht_stats;
  counter num_used_hdrs;
  BTHTptr pBTHT;
  BTNptr *ppBTN;

  ht_stats.hdr = node_statistics(sm);

  num_used_hdrs = 0;
  HashStats_NumBuckets(ht_stats) = 0;
  HashStats_TotalOccupancy(ht_stats) = 0;
  HashStats_NonEmptyBuckets(ht_stats) = 0;
  HashStats_BucketSize(ht_stats) = sizeof(void *);
  pBTHT = SM_AllocList(*sm);
  while ( IsNonNULL(pBTHT) ) {
#ifdef DEBUG
    /* Counter for contents of current hash table
       ------------------------------------------ */
    counter num_contents = 0;
#endif
    num_used_hdrs++;
    HashStats_NumBuckets(ht_stats) += BTHT_NumBuckets(pBTHT);
    HashStats_TotalOccupancy(ht_stats) += BTHT_NumContents(pBTHT);
    for ( ppBTN = BTHT_BucketArray(pBTHT);
	  ppBTN < BTHT_BucketArray(pBTHT) + BTHT_NumBuckets(pBTHT);
	  ppBTN++ )
      if ( IsNonNULL(*ppBTN) ) {
#ifdef DEBUG
	/* Count the objects in each bucket
	   -------------------------------- */
	BTNptr pBTN = *ppBTN;
	do {
	  num_contents++;
	  pBTN = BTN_Sibling(pBTN);
	} while ( IsNonNULL(pBTN) );
#endif
	HashStats_NonEmptyBuckets(ht_stats)++;
      }
#ifdef DEBUG
    /* Compare counter and header values
       --------------------------------- */
    if ( num_contents != BTHT_NumContents(pBTHT) )
      xxx_warn("Inconsistent %s Usage Calculations:\n"
	       "\tHash table occupancy mismatch.", SM_StructName(*sm));
#endif
    pBTHT = BTHT_NextBTHT(pBTHT);
  }
  if ( HashStats_NumAllocHeaders(ht_stats) !=
       (num_used_hdrs + HashStats_NumFreeHeaders(ht_stats)) )
    xxx_warn("Inconsistent %s Usage Calculations:\n"
	     "\tHeader count mismatch:  Alloc: %d  Used: %d  Free: %d",
	     SM_StructName(*sm), HashStats_NumAllocHeaders(ht_stats),
	     num_used_hdrs,  HashStats_NumFreeHeaders(ht_stats));

  return ht_stats;
}

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

SubgStats subgoal_statistics(void)
{
  SubgStats sg_stats;
  SGFrame pProdSF;

  sg_stats.sf = node_statistics(&smSF);

  SubgStats_NumProducers(sg_stats) = SubgStats_NumConsumers(sg_stats) = 0;
  for ( pProdSF = SM_AllocList(smSF);  IsNonNULL(pProdSF);
        pProdSF = subg_next_subgoal(pProdSF) ) {
    SubgStats_NumProducers(sg_stats)++;
  }
  if ( SubgStats_NumAllocFrames(sg_stats) !=
       (SubgStats_NumProducers(sg_stats) + SubgStats_NumFreeFrames(sg_stats)) )
    xxx_warn("Inconsistent Subgoal Frame Usage Calculations:\n"
	     "\tSubgoal Frame count mismatch");

  return sg_stats;
}

/*-------------------------------------------------------------------------*/

/*
 *                      Displaying Current Usage
 *                      ------------------------
 */

void print_detailed_tablespace_stats() {

  NodeStats
    btn,		/* Basic Trie Nodes */
    aln;		/* Answer List Nodes */

  SubgStats  sf;	/* Subgoal Frames */

  HashStats
    btht;		/* Basic Trie Hash Tables */
  

  btn = node_statistics(&smTableBTN);
  btht = hash_statistics(&smTableBTHT);
  sf = subgoal_statistics();
  aln = node_statistics(&smALN);

  printf("\n"
	 "  Table Space\n"
	 "  -----------\n");  
  printf("  Current Total Allocation:   %12u bytes\n"
	 "  Current Total Usage:        %12u bytes\n",
	 CurrentTotalTableSpaceAlloc(btn,btht,sf,aln),
	 CurrentTotalTableSpaceUsed(btn,btht,sf,aln));
  printf("\n"
	 "    Basic Tries:\n");
  printf("      Basic Trie Nodes (%u blocks):\n"
	 "        Allocated:   %10u  (%8u bytes)\n"
	 "        Used:        %10u  (%8u bytes)\n"
	 "        Free:        %10u  (%8u bytes)\n",
	 NodeStats_NumBlocks(btn),
	 NodeStats_NumAllocNodes(btn),  NodeStats_SizeAllocNodes(btn),
	 NodeStats_NumUsedNodes(btn),  NodeStats_SizeUsedNodes(btn),
	 NodeStats_NumFreeNodes(btn),  NodeStats_SizeFreeNodes(btn));
  printf("      Basic Trie Hash Tables (%u blocks):\n"
	 "        Headers:     %10u  (%8u bytes)\n"
	 "          Used:      %10u  (%8u bytes)\n"
	 "          Free:      %10u  (%8u bytes)\n"
	 "        Buckets:     %10u  (%8u bytes)\n"
	 "          Used:      %10u\n"
	 "          Empty:     %10u\n"
	 "        Occupancy:   %10u BTNs\n",
	 HashStats_NumBlocks(btht),
	 HashStats_NumAllocHeaders(btht),  HashStats_SizeAllocHeaders(btht),
	 HashStats_NumUsedHeaders(btht),  HashStats_SizeUsedHeaders(btht),
	 HashStats_NumFreeHeaders(btht),  HashStats_SizeFreeHeaders(btht),
	 HashStats_NumBuckets(btht),  HashStats_SizeAllocBuckets(btht),
	 HashStats_NonEmptyBuckets(btht),  HashStats_EmptyBuckets(btht),
	 HashStats_TotalOccupancy(btht));
  printf("\n"
	 "    Subgoal Frames (%u blocks):\n"
	 "      Allocated:     %10u  (%8u bytes)\n"
	 "      Used:          %10u  (%8u bytes)\n"
	 "        Producers:   %8u    (%6u bytes)\n"
	 "        Consumers:   %8u    (%6u bytes)\n"
	 "      Free:          %10u  (%8u bytes)\n",
	 SubgStats_NumBlocks(sf),
	 SubgStats_NumAllocFrames(sf),  SubgStats_SizeAllocFrames(sf),
	 SubgStats_NumUsedFrames(sf),  SubgStats_SizeUsedFrames(sf),
	 SubgStats_NumProducers(sf),
	 SubgStats_NumProducers(sf) * SubgStats_FrameSize(sf),
	 SubgStats_NumConsumers(sf),
	 SubgStats_NumConsumers(sf) * SubgStats_FrameSize(sf),
	 SubgStats_NumFreeFrames(sf),  SubgStats_SizeFreeFrames(sf));
  printf("\n"
	 "    Answer List Nodes (%u blocks):\n"
	 "      Allocated:     %10u  (%8u bytes)\n"
	 "      Used:          %10u  (%8u bytes)\n"
	 "      Free:          %10u  (%8u bytes)\n",
	 NodeStats_NumBlocks(aln),
	 NodeStats_NumAllocNodes(aln),  NodeStats_SizeAllocNodes(aln),
	 NodeStats_NumUsedNodes(aln),  NodeStats_SizeUsedNodes(aln),
	 NodeStats_NumFreeNodes(aln),  NodeStats_SizeFreeNodes(aln));

  if (flags[TRACE_STA]) {
    /* Report Maximum Usages
       --------------------- */
    update_maximum_tablespace_stats(&btn,&btht,&sf,&aln);
    printf("\n"
	   "  Maximum Total Usage:        %12ld bytes\n",
	   maximum_total_tablespace_usage());
    printf("  Maximum Structure Usage:\n"
	   "    ALNs:            %10u  (%8u bytes)\n",
	   maximum_answer_list_nodes(),
	   maximum_answer_list_nodes() * NodeStats_NodeSize(aln));
  }
  printf("\n");
}

/*-------------------------------------------------------------------------*/

/*
 *                       Recording Maximum Usage
 *                       -----------------------
 */

/*
 * Most of the data structures used in tabling are persistent: table
 * info frames, subgoal frames, and the trie nodes and hash tables
 * (basic and time-stamped).  Therefore, the maximum usage of each of
 * these structures at any particular time during an evaluation is the
 * same as their current usage.  However, there are some structures
 * whose usage fluctuates during an evaluation.  In particular,
 * TimeStamp Indices and answer list nodes are reclaimed upon
 * completion.  Therefore, we explicitly track their maximum usage at
 * those times just prior to reclamation.  The following data
 * structure maintains the maximum usage for each of these structures.
 * And as the (ultimate) maximum usage of a particular structure is
 * independent of that of other structures, we also explicitly track
 * the maximum utilization of table space.
 */

struct {
  counter tsi;                   /* TS Index Nodes */
  counter alns;                  /* Answer List Nodes */
  unsigned long  total_bytes;    /* total tablespace in bytes */
} maxTableSpaceUsage = {0,0,0};

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

void reset_maximum_tablespace_stats(void)
{
  maxTableSpaceUsage.tsi = maxTableSpaceUsage.alns = 0;
  maxTableSpaceUsage.total_bytes = 0;
}

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

/*
 * To be used to facilitate statistical recordings, but in situations
 * where one is not currently inspecting results, e.g., when reclaiming
 * table space at completion.  This procedure queries the current state
 * of usage of all table components, and tests whether this is a new
 * maximum.
 */

void compute_maximum_tablespace_stats(void)
{
  NodeStats btn, aln;
  SubgStats sf;
  HashStats btht;

  btn = node_statistics(&smTableBTN);
  btht = hash_statistics(&smTableBTHT);
  sf = subgoal_statistics();
  aln = node_statistics(&smALN);

  update_maximum_tablespace_stats(&btn,&btht,&sf,&aln);
}

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

/*
 * To be used to facilitate statistical reporting at the moment -- when
 * one is currently inspecting results.  Presumably one would want to
 * independently query the current state of usage of all table
 * components for such reporting, so we use these results rather than
 * recompute them.
 */

void update_maximum_tablespace_stats(NodeStats *btn, HashStats *btht,
				     SubgStats *sf, NodeStats *aln)
{
  unsigned long  byte_size;

  byte_size =
    CurrentTotalTableSpaceUsed(*btn,*btht,*sf,*aln);
  if ( byte_size > maxTableSpaceUsage.total_bytes )
    maxTableSpaceUsage.total_bytes = byte_size;
  if ( NodeStats_NumUsedNodes(*aln) > maxTableSpaceUsage.alns )
    maxTableSpaceUsage.alns = NodeStats_NumUsedNodes(*aln);
}

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

counter maximum_answer_list_nodes()
{
  return (maxTableSpaceUsage.alns);
}

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

unsigned long maximum_total_tablespace_usage()
{
  return (maxTableSpaceUsage.total_bytes);
}

/*==========================================================================*/
