/* File:      table_stats.h
** Author(s): Ernie Johnson
** Contact:   xsb-contact@cs.sunysb.edu
** 
** Copyright (C) The Research Foundation of SUNY, 1986, 1993-1998
** 
** XSB is free software; you can redistribute it and/or modify it under the
** terms of the GNU Library General Public License as published by the Free
** Software Foundation; either version 2 of the License, or (at your option)
** any later version.
** 
** XSB is distributed in the hope that it will be useful, but WITHOUT ANY
** WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
** FOR A PARTICULAR PURPOSE.  See the GNU Library General Public License for
** more details.
** 
** You should have received a copy of the GNU Library General Public License
** along with XSB; if not, write to the Free Software Foundation,
** Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
**
** $Id: table_stats.h,v 1.1.1.1 2001/06/13 16:41:13 kostis Exp $
** 
*/


#ifndef TABLE_STATISTICS

#define TABLE_STATISTICS

#include "struct_manager.h"

/*=========================================================================*/

/*
 *       Interface for Reporting Tabling-Component Usage Statistics
 *       ==========================================================
 */


/*
 *                       Recording Current Usage
 *                       -----------------------
 */


/* For Generic Nodes
   ----------------- */
typedef struct {
  counter nBlocks;      /* num blocks of nodes allocated (if applicable) */
  counter nAlloced;     /* total number of nodes allocated */
  counter nFree;        /* number of allocated nodes not currently in use */
  counter size;         /* size of your particular node */
} NodeStats;

#define NodeStats_NumBlocks(NS)		( (NS).nBlocks )

#define NodeStats_NumAllocNodes(NS)	( (NS).nAlloced )
#define NodeStats_NumFreeNodes(NS)	( (NS).nFree )
#define NodeStats_NumUsedNodes(NS)	( (NS).nAlloced - (NS).nFree )

#define NodeStats_NodeSize(NS)		( (NS).size )

#define NodeStats_SizeAllocNodes(NS)	( NodeStats_NumAllocNodes(NS)	\
					  * NodeStats_NodeSize(NS) )
#define NodeStats_SizeFreeNodes(NS)	( NodeStats_NumFreeNodes(NS)	\
					  * NodeStats_NodeSize(NS) )
#define NodeStats_SizeUsedNodes(NS)	( NodeStats_NumUsedNodes(NS)	\
					  * NodeStats_NodeSize(NS) )

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

/* For Subgoal Frames
   ------------------ */
typedef struct {
  NodeStats sf;          /* general stats about Subgoal Frames */
  counter nProducers;    /* number of producer subgoal frames */
  counter nConsumers;    /* subgoal frames for consuming (subsumed) subgoals */
} SubgStats;

#define SubgStats_NumBlocks(SS)		NodeStats_NumBlocks((SS).sf)

#define SubgStats_NumAllocFrames(SS)	NodeStats_NumAllocNodes((SS).sf)
#define SubgStats_NumFreeFrames(SS)	NodeStats_NumFreeNodes((SS).sf)
#define SubgStats_NumUsedFrames(SS)	NodeStats_NumUsedNodes((SS).sf)
#define SubgStats_NumProducers(SS)	( (SS).nProducers )
#define SubgStats_NumConsumers(SS)	( (SS).nConsumers )

#define SubgStats_FrameSize(SS)		NodeStats_NodeSize((SS).sf)
#define SubgStats_SizeAllocFrames(SS)	NodeStats_SizeAllocNodes((SS).sf)
#define SubgStats_SizeFreeFrames(SS)	NodeStats_SizeFreeNodes((SS).sf)
#define SubgStats_SizeUsedFrames(SS)	NodeStats_SizeUsedNodes((SS).sf)

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

/* For Hash Tables
   --------------- */
typedef struct {
  NodeStats hdr;        /* contains stats about the headers */
  counter ttlBkts;      /* total number of buckets allocated */
  counter ttlUsedBkts;  /* number of nonempty buckets */
  counter occupancy;    /* number of objects maintained by the hash table(s) */
  counter bktSize;      /* size of each bucket (usually just a pointer) */
} HashStats;

#define HashStats_NumBlocks(HS)		NodeStats_NumBlocks((HS).hdr)

#define HashStats_NumAllocHeaders(HS)	NodeStats_NumAllocNodes((HS).hdr)
#define HashStats_NumFreeHeaders(HS)	NodeStats_NumFreeNodes((HS).hdr)
#define HashStats_NumUsedHeaders(HS)	NodeStats_NumUsedNodes((HS).hdr)

#define HashStats_HeaderSize(HS)	NodeStats_NodeSize((HS).hdr)

#define HashStats_SizeAllocHeaders(HS)	NodeStats_SizeAllocNodes((HS).hdr)
#define HashStats_SizeUsedHeaders(HS)	NodeStats_SizeUsedNodes((HS).hdr)
#define HashStats_SizeFreeHeaders(HS)	NodeStats_SizeFreeNodes((HS).hdr)

#define HashStats_NumBuckets(HS)	( (HS).ttlBkts )
#define HashStats_NonEmptyBuckets(HS)	( (HS).ttlUsedBkts )
#define HashStats_EmptyBuckets(HS)	( (HS).ttlBkts - (HS).ttlUsedBkts )
#define HashStats_TotalOccupancy(HS)	( (HS).occupancy )

#define HashStats_BucketSize(HS)	( (HS).bktSize )

#define HashStats_SizeAllocBuckets(HS)	( HashStats_NumBuckets(HS) *	\
					  HashStats_BucketSize(HS) )

#define HashStats_SizeAllocTotal(HS)	( HashStats_SizeAllocHeaders(HS) + \
					  HashStats_SizeAllocBuckets(HS) )
#define HashStats_SizeUsedTotal(HS)	( HashStats_SizeUsedHeaders(HS) +  \
					  HashStats_SizeAllocBuckets(HS) )
#define HashStats_SizeFreeTotal(HS)	HashStats_SizeFreeHeaders(HS)

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

/* Collection Routines
   ------------------- */
SubgStats subgoal_statistics(void);
NodeStats node_statistics(Structure_Manager *);
HashStats hash_statistics(Structure_Manager *);

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

/* Helpful Totaling Routines
   ------------------------- */
#define CurrentTotalTableSpaceAlloc(BTN,BTHT,SF,ALN)	   \
   ( NodeStats_SizeAllocNodes(BTN)  +  HashStats_SizeAllocTotal(BTHT)  +   \
     SubgStats_SizeAllocFrames(SF)  +  NodeStats_SizeAllocNodes(ALN) )

#define CurrentTotalTableSpaceUsed(BTN,BTHT,SF,ALN)	 \
   ( NodeStats_SizeUsedNodes(BTN)  +  HashStats_SizeUsedTotal(BTHT)  +	 \
     SubgStats_SizeUsedFrames(SF)  +  NodeStats_SizeUsedNodes(ALN) )

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

/* Printing Routines
   ----------------- */
extern void print_detailed_tablespace_stats(void);

/*-------------------------------------------------------------------------*/

/*
 *                       Recording Maximum Usage
 *                       -----------------------
 */

/* Reset Usage
   ----------- */
extern void reset_maximum_tablespace_stats(void);

/* Poll Current Usage for New Maximum
   ---------------------------------- */
extern void compute_maximum_tablespace_stats(void);
extern void update_maximum_tablespace_stats(NodeStats *btn, HashStats *btht,
					    SubgStats *sf, NodeStats *aln);

/* Read Currently Recorded Maximum Values
   -------------------------------------- */
/* One should first force a check of the maximum usage before querying
   for these values. */
extern counter maximum_answer_list_nodes(void);
extern unsigned long  maximum_total_tablespace_usage(void);

/*=========================================================================*/


#endif
