/* File:      tables.c
** Author(s): Ernie Johnson
** Contact:   xsb-contact@cs.sunysb.edu
** 
** Copyright (C) The Research Foundation of SUNY, 1986, 1993-1998
** 
** XSB is free software; you can redistribute it and/or modify it under the
** terms of the GNU Library General Public License as published by the Free
** Software Foundation; either version 2 of the License, or (at your option)
** any later version.
** 
** XSB is distributed in the hope that it will be useful, but WITHOUT ANY
** WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
** FOR A PARTICULAR PURPOSE.  See the GNU Library General Public License for
** more details.
** 
** You should have received a copy of the GNU Library General Public License
** along with XSB; if not, write to the Free Software Foundation,
** Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
**
** $Id: tables.c,v 1.15 2006/09/18 14:26:02 kostis Exp $
** 
*/


#include "xxx_config.h"

#include <stdio.h>
#include <stdlib.h>

#include "auxlry.h"
#include "cell_xsb.h"
#include "heap_xsb.h"
#include "memory_xsb.h"
#include "psc_xsb.h"
#include "table_stats.h"
#include "macro_xsb.h"
#include "register.h"
#include "binding.h"
#include "error_xsb.h"
#include "flags_xsb.h"
#include "inst_xsb.h"
#include "tables.h"

/*-------------------------------------------------------------------------*/

/* Manager Structures
   ------------------ */
Structure_Manager smSF  = SM_InitDecl(struct subgoal_frame,
				      SUBGOAL_FRAMES_PER_BLOCK,
				      "Subgoal Frame");
Structure_Manager smALN = SM_InitDecl(AnsListNode, ALNs_PER_BLOCK,
				      "Answer List Node");

/*-------------------------------------------------------------------------*/

/*
 * Performs the Call Check/Insert operation.
 * Upon exit, CallLUR_VarVector(*results) points to the size of the
 * answer template on the CPS, unless a producer was found and we're
 * running in CHAT node.  In that case, this value is the same as
 * CallInfo_VarVector(*call_info), and the AT is sitting on the heap,   
 * its size pointed to by (hreg - 1).  See slginsts_xsb_i.h for answer
 * template layout.
 */

void table_call_search(TabledCallInfo *call_info, CallLookupResults *results)
{
  variant_call_search(call_info,results);

  if ( IsNULL(CallLUR_Subsumer(*results)) ) {
    /*
     * A New Producer: Move answer template from CPS to Heap.  The
     * arrangement of this vector is similar to that in the CPS: the
     * size of the vector is now at the high end, but the components
     * are still arranged from high mem (first) to low (last).
     */
    CPtr tmplt_component, tmplt_var_addr, h_addr;
    int size, j;

    tmplt_component = CallLUR_VarVector(*results);
    size = int_val(*tmplt_component) & 0xffff;

    for ( j = size - 1, tmplt_component = tmplt_component + size;
	  j >= 0;
	  j--, tmplt_component-- ) {
      tmplt_var_addr = (CPtr)*tmplt_component;
      /* Now we are sure that tmplt_var_addr is a var on the heap */
      h_addr = hreg+j;
      bld_copy(h_addr, (Cell)(tmplt_var_addr));
    }
    hreg += size;
    bld_copy(hreg, cell(CallLUR_VarVector(*results)));
    hreg++;
    /* orig version in tries.c had VarPosReg pointing at Var_{m} */
    CallLUR_VarVector(*results) = CallLUR_VarVector(*results) + size + 1;
  }
}

/*-------------------------------------------------------------------------*/

/*
 * Template is a pointer to the first term in the vector, with the
 * elements arranged from high to low memory.
 */
BTNptr table_answer_search(SGFrame producer, int size,
#if defined(ATTR_VARS) && !defined(CST_SUPPORT)
			   int attv_num,
#elif defined(CST_SUPPORT)
			   CST *constraintTree,
#endif
			   CPtr template, bool *is_new)
{
  BTNptr answer;

  /*
   * We want to save the substitution factor of the answer in the
   * heap, so we have to change variant_answer_search().
   */
  bool wasFound = TRUE;
  
#ifndef IGNORE_DELAYVAR
  ans_var_pos_reg = hreg++;	/* Leave a cell for functor ret/n */
#endif /* IGNORE_DELAYVAR */

  answer = variant_answer_search(size,
				 template,
				 producer,
#if defined(CST_SUPPORT)
				 constraintTree,
#elif defined(ATTR_VARS)
				 attv_num,
#endif
				 &wasFound);

#ifdef DEBUG_DELAYVAR
#ifndef IGNORE_DELAYVAR
  fprintf(stddbg, ">>>> ans_var_pos_reg = ");
  if (isinteger(cell(ans_var_pos_reg)))
    fprintf(stddbg, "\"ret\"\n");
  else 
    fprintf(stddbg, "%s/%d\n", get_name((Psc)(cell(ans_var_pos_reg))),
	    get_arity((Psc)(cell(ans_var_pos_reg))));
#endif /* IGNORE_DELAYVAR */
#endif /* DEBUG_DELAYVAR */
  
  do_delay_stuff(
#if defined(CST_SUPPORT)
      *constraintTree,
#else
      answer,       
#endif
      producer, 
      wasFound);

#ifndef IGNORE_DELAYVAR
  undo_answer_bindings();
#endif /* IGNORE_DELAYVAR */

  *is_new = ! wasFound;
  return ( answer );
}

/*-------------------------------------------------------------------------*/

void table_consume_answer( 
#if defined(CST_SUPPORT)
			  CST answer,
#else
			  BTNptr answer,
#endif
			  int size, 
			  int attv_num,
			  CPtr template)
{
  if (size == 0) {
    num_heap_term_vars = 0;	/* ensure it is initialized in all paths */
    if ( ! 
#if defined(CST_SUPPORT)
	 IsEscapeNode(CST_owning_trie_node(answer))
#else
	 IsEscapeNode(answer)
#endif
	)
      xxx_abort("Size of answer template is 0 but producer contains an "
		"answer\nwith a non-empty substitution!\n");
  }
  else {
    /* this also tracks variables created during unification */
    load_solution_trie(size,attv_num,template,answer);
  }
}

/*-------------------------------------------------------------------------*/

/*
 *                     Table Structure Reclamation
 *                     ===========================
 */

/*
 *  Deallocate all the data structures which become superfluous once the
 *  table has completed.  Currently, this includes the answer list nodes
 *  and the TSIs from the producer, and the answer list nodes from the
 *  subsumed subgoals.  For the producers, the engine requires that the
 *  dummy answer-list node remain, and that its 'next' field be set to
 *  either the constant CON_ANSWERS or UNCOND_ANSWERS depending on whether
 *  there were any conditional answers in the answer list.  For the
 *  subsumed (consumer) subgoals, the entire answer list, including the
 *  dummy, is reclaimed.
 *
 *  For statistical purposes, we check whether the current usage of
 *  these incomplete-table structures are new maximums.
 *
 *  Currently, timestamps from the TSIs are copied back to the TSTNs.
 *  Although not necessary, this method has some advantages.  Foremost,
 *  this field will never contain garbage values, and so we avoid dangling
 *  pointer problems.  Also, maintaining the time stamp values is
 *  beneficial for post-completion analysis of the TSTs and
 *  characteristics of the query evaluation.
 *
 *  Multiple calls to this function is avoided by checking a flag in the
 *  subgoal frame.  See macro reclaim_incomplete_table_structs() which
 *  contains the only reference to this function.
 */

void table_complete_entry(SGFrame producerSF)
{
  ALNptr pRealAnsList, pALN;
  ALNptr tag = NO_ANSWERS;

  if (flags[TRACE_STA]) compute_maximum_tablespace_stats();

  if ( has_answer_code(producerSF) ) {
    /* Reclaim Producer's Answer List
       ------------------------------ */
    if ( has_answers(producerSF) ) {
      pALN = pRealAnsList = subg_answers(producerSF);
      tag = UNCOND_ANSWERS;
      do {
	if ( 
#if defined(CST_SUPPORT)
	    is_conditional_answer(ALN_Constraint(pALN))
#else
	    is_conditional_answer(ALN_Answer(pALN))
#endif
	    ) {
	  tag = COND_ANSWERS;
	  break;
	}
	pALN = ALN_Next(pALN);
      } while ( IsNonNULL(pALN) );
      subg_answers(producerSF) = tag;

#ifdef DEBUG
      if ( IsNULL(subg_ans_list_tail(producerSF)) ||
	   IsNonNULL(ALN_Next(subg_ans_list_tail(producerSF))) )
	xxx_abort("Answer-List exception: Tail is incorrectly maintained");
#endif
      SM_DeallocateStructList(smALN,pRealAnsList,
			      subg_ans_list_tail(producerSF));
      subg_ans_list_tail(producerSF) = NULL;
    }
  }
}

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

/*
 * Frees all the tabling space resources.
 */

void release_all_tabling_resources(void)
{
  SM_ReleaseResources(smTableBTN);
  TrieHT_FreeAllocatedBuckets(smTableBTHT);
  SM_ReleaseResources(smTableBTHT);
  SM_ReleaseResources(smALN);
  SM_ReleaseResources(smSF);
#if defined(CST_SUPPORT)
  CST_release_resources();
#endif
}

/*-------------------------------------------------------------------------*/
