/* File:      assert.P
** Author(s): David S. Warren, Kostis Sagonas
** Contact:   xsb-contact@cs.sunysb.edu
** 
** Copyright (C) The Research Foundation of SUNY, 1993-1998
** 
** XSB is free software; you can redistribute it and/or modify it under the
** terms of the GNU Library General Public License as published by the Free
** Software Foundation; either version 2 of the License, or (at your option)
** any later version.
** 
** XSB is distributed in the hope that it will be useful, but WITHOUT ANY
** WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
** FOR A PARTICULAR PURPOSE.  See the GNU Library General Public License for
** more details.
** 
** You should have received a copy of the GNU Library General Public License
** along with XSB; if not, write to the Free Software Foundation,
** Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
**
** $Id: assert.P,v 1.6 2004/08/22 16:25:05 kostis Exp $
** 
*/

:- compiler_options([xpp_on]).

#include "builtin_tab.h"
#include "sig_xsb.h"
#include "psc_defs.h"
#include "flag_defs_xsb.h"

/*======================================================================*/
/* Builtin support used only in this part of the system and not inlined */
/*======================================================================*/

db_retract_abolish(_Call,_Reclaim) :- '_$builtin'(DB_RETRACT_ABOLISH).
db_retract_clause(_Clause,_Reclaim) :- '_$builtin'(DB_RETRACT_CLAUSE).
db_get_clause(_InitClause,_CI,_CIL,_Head,_Failed,_ClRef,_Entry) :- 
	'_$builtin'(DB_GET_CLAUSE).
db_convert_to_dynamic(_Head,_Tabled) :- '_$builtin'(DB_CONVERT_TO_DYNAMIC).
db_assert_code_to_buff(_Clause) :- '_$builtin'(DB_ASSERT_CODE_TO_BUFF).
db_assert_buff_to_clref(_Head, _AZ, _Indices, _HashTabSize) :-
	'_$builtin'(DB_ASSERT_BUFF_TO_CLREF).

trie_assert(_Clause,_Tabled,_Flag) :- '_$builtin'(TRIE_ASSERT).
trie_retract(_Clref) :- '_$builtin'(TRIE_RETRACT).
trie_retract_nr :- '_$builtin'(TRIE_RETRACT_SAFE).
is_trie_asserted(_Clref) :- '_$builtin'(IS_TRIE_ASSERTED).

/*======================================================================*/
/* Predicates to add dynamic clauses in Prolog's database:		*/
/*  - assert(+Clause)							*/
/*	same as assert(Clause, 1, 1)					*/
/*  - asserta(+Clause)							*/
/*	same as assert(Clause, 0, 1)					*/
/*  - assertz(+Clause)							*/
/*	same as assert(Clause, 1, 1)					*/
/*  - asserti(+Clause, +Index)						*/
/*	same as assert(Clause, 1, Index)				*/
/* where:								*/
/*    assert(+Clause, +AorZ, +Index)            			*/
/*	Asserts a clause.		                                */
/*	AorZ:  0 for insertion as the first clause, 1 for as the last.	*/
/*	Index: the number of the argument on which to index; 		*/
/*		0 for no indexing. 					*/
/*======================================================================*/

/*======================================================================*/
/* Formats for dynamic code:						*/
/* PSC rec point to:							*/
/*	PrRef:								*/
/*		0: BC instruction: fail (if empty),			*/
/*			jump and save breg (if nonempty)		*/
/*		4: Addr of first Clref on ALL chain			*/
/*		8: Addr of last Clref on ALL chain			*/
/*									*/
/* PrRef's point to chain of ClRef's (one of 3 types):			*/
/* (the -8 location stores length of buff + flag indicating ClRef type	*/
/*	ClRef2 (for compiled code):					*/
/*		-8: length of buffer (+2)				*/
/*		-4: Addr of previous ClRef (or PrRef)			*/
/*		0: Try-type instruction, for chain			*/
/*		4: (cont) Addr of next ClRef on chain			*/
/*		8: jump							*/
/*		12: Addr of compiled code				*/
/*	ClRef0 (for unindexed asserted code):				*/
/*		-8: length of buffer (+0)				*/
/*		-4: Addr of previous ClRef (or PrRef)			*/
/*		0: Try-type instruction, for chain			*/
/*		4: (cont) Addr of next ClRef on chain			*/
/*		8+: BC for asserted clause				*/
/*	ClRef1 (for group of indexed clauses, aka SOB record):		*/
/*		-8: length of buffer (+1)				*/
/*		-4: Addr of previous ClRef (or PrRef)			*/
/*		0: Try-type instruction, for chain			*/
/*		4: (cont) Addr of next ClRef on chain			*/
/*		8: BC switch-on-bound instruction (drop thru if var)	*/
/*		11: (cont) arg to index on				*/
/*		12: (cont) address of Hash Table			*/
/*		16: (cont) size of Hash Table				*/
/*		20: BC jump to	(or fail if empty)			*/
/*		24: (cont) Addr of first ClRefI on all subchain		*/
/*		28: Addr of last ClRefI on all subchain			*/
/*		32: Number of clauses in hash table			*/
/*		36+: Hash Table						*/
/*									*/
/* ClRef1's point to indexed clauses, each represented by a ClRefI:	*/
/*	ClRefI (for an indexed clause):					*/
/*		-8: length of buffer (+3)				*/
/*		-4: Addr of previous ClRefI on all chain		*/
/*		0: Try-type instruction, for all subchain		*/
/*		4: (cont) Addr of next ClRefI on all subchain		*/
/*		8: BC noop(6) to skip next 12 bytes			*/
/*		12: Addr of previous ClRefI on bucket chain		*/
/*		16: Try-type instruction, for hash bucket subchain	*/
/*		20: (cont) Addr of next ClRefI in bucket		*/
/* and for each additional index:					*/
/*		24: BC noop(6) to skip next 12 bytes			*/
/*		28: Addr of previous ClRefI on bucket chain		*/
/*		32: Try-type instruction, for hash bucket subchain	*/
/*		36: (cont) Addr of next ClRefI in bucket		*/
/*		40+: BC for asserted code				*/
/*									*/
/*======================================================================*/


% NOTE: Predicate assert0/1 is a quicker assert/1 with no error
%	checking, provided for system usage only.

assert0(Clause) :- assert(Clause,1,1).


assert(Clause)  :-
	check_assert(Clause, 'assert/1'),
	call_assert_hook(Clause),
	assert(Clause,1,1).
asserta(Clause) :-
	check_assert(Clause,'asserta/1'),
	call_assert_hook(Clause),
	assert(Clause,0,1).
assertz(Clause) :-
	check_assert(Clause,'assertz/1'),
	call_assert_hook(Clause),
	assert(Clause,1,1).

call_assert_hook(Clause) :-
	term_psc(assert_hook(_),Psc),
	psc_type(Psc,Type),
	(Type > 0
	 ->	call_c(assert_hook(Clause))
	 ;	true
	).

% NOTE: Predicates asserti[a,z]/2 though exported, are not available to
%	the user (in the intepreter) and so no error checking is done.
%	Should they become available in the interpreter error checking
%	must be added to them.

asserti(Clause, Index) :- assert(Clause,1,Index).
assertai(Clause,Index) :- assert(Clause,0,Index).
assertzi(Clause,Index) :- assert(Clause,1,Index).


assert(Clause, AZ, Index1) :-
	( Clause = (Head:-_) -> term_psc(Head, PSC)
	; Clause = Head -> term_psc(Head, PSC)
	),
	psc_arity(PSC, Arity),
	( '_$trie_asserted'(Clause) -> t_assert_1(Clause, _Flag)
	; ( '_$index'(Head,Index,HashTabSize) -> true
	  ; integer(Index1) ->
	      (	Index1 > Arity -> Index = 0, default_assert_size(HashTabSize)
	      ;	Index1 < 0 -> Index = 1, default_assert_size(HashTabSize)
	      ;	Index = Index1, default_assert_size(HashTabSize)
	      )
	  ; Index = Index1, default_assert_size(HashTabSize)
	  ),
	  convert_to_dynamic(Head),
	  cut_transform(Clause,Nclause),
	  code_to_buff(Nclause),
	  db_assert_buff_to_clref(Head,AZ,Index,HashTabSize)
	).

/*======================================================================*/
/* check_assert(+Clause, +Predicate_Called)				*/
/*	Given the Clause to be asserted in the database and the flavor	*/
/*	of assert used, checks the Clause for errors so  that Prolog's	*/
/*	database remains consistent. As usual, if an error is detected	*/
/*	the predicate aborts.						*/
/*	The errors that can occur are:					*/
/*	  - instantiation error: If Head of Clause is uninstantiated.	*/
/*	  - type error: If Head of Clause is not a callable term or	*/
/*		if Body contains a non-callable goal.			*/
/*	  - permission error: If the predicate corresponding to Head	*/
/*		is one of XSB's built-in predicates (Database error).	*/
/*======================================================================*/

check_assert((Head :- Body), Pred) :- 
	!,			% This captures Clause being variable
	( var(Head) ->
		abort(('Uninstantiated argument 1 of ', Pred))
	; check_assert_fact(Head, Pred), 
	  ( check_body(Body) -> true
	  ; abort(('Wrong type in argument 1 of ', Pred,
			   ' (callable term expected)'))
	  )
	).
check_assert(Head, Pred) :- check_assert_fact(Head, Pred).

check_assert_fact(Head, Pred) :- 	% asserting a fact
	( callable(Head) -> 
	    ( functor(Head, F, A), standard_symbol(F, A, _Prop) ->
		abort(('Not permitted to assert to a built-in predicate: ',F/A))
	    ;	true
	    )
	; abort(('Wrong type in argument 1 of ', Pred, 
			 ' (callable term expected)'))
	).

check_body(X) :- var(X), !.
check_body((X,Y)) :- !, check_body(X), check_body(Y).
check_body(X) :- callable(X).

/*======================================================================*/
/* cut_transform(+Clause, -NewClause)					*/
/*	Transform cuts to '_$cutto'.					*/
/*======================================================================*/

cut_transform((Head:-Body),(Nhead:-Nbody)) :- !,
	Head =.. Hlist,
	append(Hlist,[Cutpoint],Nhlist),
	Nhead =.. Nhlist,
	goal_cut_trans(Body,Nbody,Cutpoint).
cut_transform(Head,Head). /* leave unchanged, Arity is one less */

/*======================================================================*/
/* convert_to_dynamic(+PredHead)					*/
/*	Convert the predicate to T_DYNA, if not already.		*/
/*======================================================================*/

dynamic((P1,P2)) :- !, dynamic(P1), dynamic(P2).
dynamic(P/A) :- !, functor(Head,P,A), convert_to_dynamic(Head).
dynamic(Head) :- convert_to_dynamic(Head).

convert_to_dynamic(PHead) :-
	( '_$table'(PHead) -> Tabled = 1 ; Tabled = 0 ),
	db_convert_to_dynamic(PHead,Tabled).

/*======================================================================*/
/* retract(+Fact)							*/
/*	this routine retracts facts. It does so by running the chain of */
/*	buffers, explicitly. When it finds a fact that unifies, it 	*/
/*	overwrites the first instruction in the buffer (after the 	*/
/*	retrymeelse instruction) to be a fail. This is somewhat of a 	*/
/*	kludge but is easy. Besides you shouldn't be using retract 	*/
/*	anyway. 							*/
/*	(dsw 4/21/92) It's also incorrect. It deletes ALL clauses in	*/
/*	the group COMPILED with it. This is particularly unpleasant in	*/
/*	op/3, where redefining a 'builtin' operator, deletes them ALL.	*/
/*======================================================================*/


retract(Rule) :-	% changed for multifile predicates (2/24/98)
	(Rule = (Head :- Body)
	 ->	true
	 ;	Head = Rule, Body = true
	),
	call_retract_hook(Head,Body),
	(nonvar(Head), '_$multifile_retract'(Head, _, _)
	 % argument 1 fo retract/1 must be instantiated
	 ->	do_multifile_retract(Head, Body)
	 ;	retract0((Head :- Body))
	).

call_retract_hook(Head,Body) :-
	term_psc(retract_hook(_,_),Psc),
	psc_type(Psc,Type),
	(Type > 0
	 ->	call_c(retract_hook(Head,Body))
	 ;	true
	).

do_multifile_retract(Head, Body) :-
	'_$multifile_retract'(Head, _, New_Head),
	check_retract(New_Head, retract/1),
	term_psc(New_Head, PSC),
	psc_type(PSC,Type), 
	( Type =:= T_PRED ->	/* is compiled */
	        functor(Head,F,A),
	        abort(['Not permitted to retract from static predicate', F/A])
	; db_retract(New_Head, Body)
	).


retract0(Rule) :-	% original retract/1, is used in standard.P
	(Rule = (Head :- Body)
	 ->	true
	 ;	Head = Rule, Body = true
	),
	check_retract(Head, retract/1),
	term_psc(Head, PSC),
	psc_type(PSC,Type),
	( Type =:= T_PRED ->	/* is compiled */
	        functor(Head,F,A),
	        abort(['Not permitted to retract from static predicate', F/A])
	; Type =:= T_DYNA,
	        db_retract(Head, Body)
	).

retractall(Fact) :-
	call_retract_hook(Fact,true),
	(nonvar(Fact), '_$multifile_retract'(Fact, _, _)
	 ->	do_multifile_retractall(Fact)
	 ;	retractall0(Fact)
	).

do_multifile_retractall(Fact) :-
	'_$multifile_retract'(Fact, _, New_Fact),
	retractall0(New_Fact),
	fail.
do_multifile_retractall(_).

retractall0(Fact) :-
	check_retract(Fact, retractall/1),
	term_psc(Fact, PSC),
	psc_type(PSC,Type), 
	( Type =:= T_PRED ->	/* is compiled */
		functor(Fact,F,A),
	        abort(['Not permitted to retract from static predicate', F/A])
	; Type =:= T_DYNA ->
	        db_retractall(Fact)
	; true
	).

retract_nr(Rule) :-
	(Rule = (Head :- Body)
	 ->	true
	 ;	Head = Rule, Body = true
	),
	check_retract(Head, retract_nr/1),
	term_psc(Head, PSC),
	psc_type(PSC,Type), 
	( Type =:= T_PRED ->	/* is compiled */
		functor(Head,F,A),
	        abort(['Not permitted to retract from static predicate', F/A])
	; Type =:= T_DYNA,
	        db_retract_nr(Head, Body)
	).

reclaim_space(Fact) :-
	check_retract(Fact, reclaim_space/1),
	term_psc(Fact,PSC),
	psc_type(PSC,Type), 
	( Type =:= T_PRED ->	/* is compiled */
		functor(Fact,F,A),
	        abort(['Not permitted to retract from static predicate', F/A])
	; Type =:= T_DYNA ->
	    db_reclaim_space(Fact)
	; true
	).

/*======================================================================*/
/* check_retract(+Head_of_Clause, +Predicate_Called)			*/
/*	Given the Head of the Clause to be retracted from the database	*/
/*	checks it for exceptions so that Prolog's database remains	*/
/*	consistent. As usual, if an error is detected the predicate	*/
/*	aborts.								*/
/*	The errors that can occur are:					*/
/*	  - instantiation error: If Head_of_Clause is uninstantiated.	*/
/*	  - type error: If Head_of_Clause is not a callable term.	*/
/*	  - permission error: If the predicate corresponding to Head	*/
/*		is one of XSB's built-in predicates (Database error).	*/
/*	Note that it is not a type error to call the retract(s) with	*/
/*	Head a functor of a non-dynamic procedure, or with a body	*/
/*	instantiated to a non-valid Prolog body. In these cases		*/
/*	predicate retract or retractall simply fail.			*/
/*======================================================================*/

check_retract(Head, Pred) :-
	( var(Head) ->
		abort(('Uninstantiated argument 1 of ', Pred))
	; callable(Head) -> 
		( functor(Head, F, A), standard_symbol(F, A, _Prop) ->
			abort(('Not permitted to retract clauses of built-in predicate ', F/A))
		; true
		)
	; abort(('Wrong type in argument 1 of ', Pred,
			 ' (callable term expected)'))
	).


/*======================================================================*/
/* abolish(+Pred/+Arity) or abolish(+PredSpec)				*/
/*	Removes all procedures specified from the Prolog database.	*/
/*	After this command is executed the current program functions	*/
/*	as if the named procedures had never existed. Predicate		*/
/*	abolish/[1,2] removes all clauses of the specified procedure	*/
/*	regardless of whether they are dynamic or compiled, but it	*/
/*	cannot abolish built-in predicates or predicates in unloaded	*/
/*	modules.							*/
/*	As it ought to, abolish/[1,2] does not complain when trying to	*/
/*	abolish an undefined predicate.					*/
/*======================================================================*/

abolish(Spec) :-
	( var(Spec) -> abort('Uninstantiated argument of abolish/1')
	; Spec = (Pred/Arity)
	 ->	(var(Pred)
		 ->	abort('Uninstantiated argument of abolish/1')
		 ; arity_integer(Arity)
		 ->	functor(Nspec,Pred,Arity)
		 ;	abort('Incorrect arity in abolish/1')
		)
	; Nspec = Spec, functor(Nspec,Pred,Arity)
	),
	(standard_symbol(Pred,Arity,_Prop)
	 ->	abort('Not permitted to abolish built-in predicates')
	 ;	call_retract_hook(Nspec,true),
		abolish1(Nspec)
	).

abolish1(Term) :-
	retract_index(Term),
	retract_trie_index(Term),
	db_retract_abolish(Term,1/*reclaim*/).

/*======================================================================*/
/* abolish(+Pred, +Arity)						*/
/*	Predicate abolish/2 is an obsolete special case of abolish/1	*/
/*	maintained here to provide compatibility with DEC-10 Prolog,	*/
/*	Quintus Prolog, C Prolog and earlier versions of SB and		*/
/*	PSB-Prolog.							*/
/*======================================================================*/

abolish(Pred,Arity) :- abolish(Pred/Arity).

/*======================================================================*/
/* clause(+Head, ?Body)							*/
/*	Predicate clause/2 searches the database for a clause whose	*/
/*	head matches the given Head and whose body matches Body. So	*/
/*	Head must be instantiated. This procedure is non-deterministic.	*/
/*	it can be used to backtrack through all the clauses matching	*/
/*	a given Head and Body. It fails when there are no (or no	*/
/*	further) matching clauses in the database.			*/
/*	In the case of unit-clauses, Body is unified with 'true'.	*/
/*======================================================================*/

clause(Head, Body) :-
	check_clause(Head),
	term_psc(Head, PSC),
	psc_type(PSC, Type), Type = T_DYNA,
	db_clause(Head, Body0),
	goal_cut_untrans(Body0, Body).

/*======================================================================*/
/* check_clause(+Head_of_Clause)					*/
/*	Given the Head of the Clause to be asserted in the database	*/
/*	checks this Head for errors. As usual, if an error is detected	*/
/*	the predicate aborts.						*/
/*	The errors that can occur are:					*/
/*	  - instantiation error: If Head of Clause is uninstantiated.	*/
/*	  - type error: If Head of Clause is not a callable term.	*/
/*	  - permission error: If the predicate corresponding to Head	*/
/*		is one of XSB's built-in predicates (Database error).	*/
/*	Note that it is not a type error to call clause/2 with Head a	*/
/*	functor of a non-dynamic procedure, or with body instantiated	*/
/*	to a non-valid Prolog body. In these cases predicate clause/2	*/
/*	simply fails.							*/
/*======================================================================*/

check_clause(Head) :-
	( var(Head) ->
		abort('Uninstantiated argument 1 of clause/2')
	; callable(Head) -> 
		( functor(Head, F, A), standard_symbol(F, A, _Prop) ->
			abort(('Not permitted to access clauses of built-in predicate ', F/A))
		; true
		)
	; abort(('Wrong type in argument 1 of clause/2', 
			 ' (callable term expected)'))
	).

/*======================================================================*/
/*======================================================================*/

t_assert_exit(Clause) :- abort(('Illegal Term in trie_assert:',Clause)).

t_assert(Clause, Flag) :-
	check_assert(Clause,'trie_assert/1'),
	t_assert_1(Clause, Flag).

t_assert_1((Head :- Body), _Flag) :- !, t_assert_exit((Head :- Body)).
t_assert_1(Term, Flag) :- 
	( '_$table'(Term) -> Tabled = 1 ; Tabled = 0 ),
	trie_assert(Term,Tabled,Flag).

/*======================================================================*/
/* Module dbclause defines routines to step through clauses and 	*/
/* return them, and also to delete clauses for retracting.  Space is	*/
/* reclaimed.  See above for formats of data structures.		*/
/*======================================================================*/

% db_retract(+Head,+Body) deletes clauses and reclaims space

db_retract(Head,Body) :-
	goal_cut_trans_a(Body,Tbody,_),
	db_get_clauses(Head,0,Clref,EntryP),
	( is_trie_asserted(Clref) -> 
		call_trie_retract(Head,Body,Clref)
	;   db_clause_match(Head,Tbody,EntryP),
	    db_retract_clause(Clref,1/*reclaim space*/)
	).

%% A version that does not reclaim space
db_retract_nr(Head,Body) :-
	goal_cut_trans_a(Body,Tbody,_),
	db_get_clauses(Head,0,Clref,EntryP),
	( is_trie_asserted(Clref) ->
		call_trie_retract_nr(Head,Body)
	;   db_clause_match(Head,Tbody,EntryP),
	    db_retract_clause(Clref,0) % 0 -> Don''t reclaim space! *****
	).

% transform a goal to turn cuts (!) into '_$cutto's. Differs from
% goal_cut_trans in that variables are not wrapped with call/1,
% but are left unchanged.

goal_cut_trans_a(X,X,_) :- var(X),!.
goal_cut_trans_a([X|Y],Z,_) :- !, hack_consult([X|Y],Z). %  a HACK!
goal_cut_trans_a(!,'_$cutto'(Cutpoint),Cutpoint) :- !.
goal_cut_trans_a('_$load_undef'(X),Y,_) :- !, Y=X.
goal_cut_trans_a((A,B),(Na,Nb),Cutpoint) :- !,
	goal_cut_trans_a(A,Na,Cutpoint),
	goal_cut_trans_a(B,Nb,Cutpoint).
goal_cut_trans_a((A;B),(Na;Nb),Cutpoint) :- !,
	goal_cut_trans_a(A,Na,Cutpoint),
	goal_cut_trans_a(B,Nb,Cutpoint).
goal_cut_trans_a((A->B),(A->Nb),Cutpoint) :- !,
	goal_cut_trans_a(B,Nb,Cutpoint).
goal_cut_trans_a(X,X,_).

hack_consult([X],consult(X)) :- !.
hack_consult([X|Y],(consult(X),Z)) :- !,hack_consult(Y,Z).


% db_retractall(+Call) deletes all clauses matching a pattern

db_retractall(Call) :-
	( is_most_general_term(Call) ->
	    db_retract_abolish(Call,0/*don't reclaim*/)
	; db_retract(Call,_)
	),
	fail.
db_retractall(_).

% db_clause(#Head,#Body) retrieves a clause.

db_clause(Head,Body) :-
	goal_cut_trans_a(Body,Tbody,_),
	db_get_clauses(Head,0,_Clref,EntryP),
	db_clause_match(Head,Tbody,EntryP).

% db_clause_match(#Call,#Body,+Entry)
%	Determines whether Clref contains a clause whose head matches Call.
%	Type indicates the kind	of Clref (indexed or not). 

db_clause_match(Head,Body,Entry) :-
	db_body_save(Buff),
	buff_set_cell(Buff,true),	% set default
	(db_do_call(Entry,Head)
	 ->	stat_set_flag(CLAUSE_INT,0),
	 	db_body_save(Buff),
		buff_cell(Buff,Body0),
		goal_cut_trans_a(Body0, Body1, _),
		( atom(Body1) ->
		    conname(Body1,Body2), Body = Body2
		; Body = Body1
		)
	 ;	stat_set_flag(CLAUSE_INT,0),
	 	fail
	).

% db_do_call(+CodeEntry, +Head)
%	Makes the call to Head on the heap after setting the interrupt flag
%	that will retrieve the code from the heap.  By this time we finally
%	know that Head is "dynamic" (psc_type = 1) so there is no need to
%	retrieve the type.  Under NO circumstances, should code be put
%	between stat_set_flag/2 and code_call/3, because the hack of building
%	the body of the call on the heap will not work!!.

db_do_call(Entry, Head) :-
	stat_set_flag(CLAUSE_INT, 1),	% set clause interrupt on
	code_call(Entry, Head, T_DYNA).	% dynamic predicate.

/*======================================================================*/

:- set_inthandler('_$clause_int'(_), MYSIG_CLAUSE).

:- dynamic(db_body_save(_)).
:- buff_alloc(8,Buff),asserti(db_body_save(Buff),0).

'_$clause_int'(Call) :-
	stat_set_flag(CLAUSE_INT, 0),	% turn clause intercept off
 	db_body_save(Buff),
	buff_set_var(Buff, 8, X),
	X = Call.

/*======================================================================*/

%% db_get_clauses(+Term, +Failed, -Clref, -EntryPoint) returns Clrefs
%%      for each clause whose head might unify with Term. It uses the SOB 
%%      index blocks to efficiently find them. It may return clauses whose
%%	head won''t unify with Term.
%%	Failed means fetch retract(ed)_nr clauses.

db_get_clauses(Head, Failed, ClRef, EntryP) :-
	db_get_clauses0(0, _, _, Head, Failed, ClRef, EntryP).

%% db_get_clauses0 calls the builtin, db_get_clause, to get the next ClRef.
%% 	It uses db_get_clauses1 to do a one clause look-ahead, so that a 
%%	deterministic retract doesn''t lay down a choice point.  This makes
%%	the semantics of the new C retract more consistent with the old.

db_get_clauses0(CurrCls0,CurrInd0,CurrILvl0,Head,Failed,ClRef,EntryP) :-
	db_get_clause(CurrCls0,CurrInd0,CurrILvl0,Head,Failed,ClRef0,EntryP0),
	( ClRef0 =:= 0 -> !, fail
	; db_get_clauses1(ClRef0,CurrInd0,CurrILvl0,Head,Failed,
			  ClRef,EntryP,EntryP0)
	).

db_get_clauses1(ClRef0,CurrInd0,CurrILvl0,Head,Failed,ClRef,EntryP,EntryP0) :-
	db_get_clause(ClRef0,CurrInd0,CurrILvl0,Head,Failed,ClRef1,EntryP1),
	( ClRef1 =:= 0 -> !, ClRef = ClRef0, EntryP = EntryP0
	; ( ClRef = ClRef0, EntryP = EntryP0
	  ; db_get_clauses1(ClRef1,CurrInd0,CurrILvl0,Head,Failed,
			    ClRef,EntryP,EntryP1)
	  )
	).

/************************************************************************/
/* The following routine cleans up by reclaiming space for all clauses	*/
/* that were erased without reclaiming their space.			*/
/************************************************************************/

db_reclaim_space(Head) :-
	db_get_clauses(Head,1,Clref,_),
	db_retract_clause(Clref,1/*reclaim space*/),
	fail.
db_reclaim_space(_).

%==================================================
% Trie Assert related code
%==================================================

call_trie_retract(Head,true,Clref) :-
	repeat,
	( trie_retract(0), % Make Last Node Save to 0
	  call(Head) ->
	    trie_retract(Clref)
	; !, fail
	).

% not safe
% if call(Head) results in Backtracking through program clauses
% which may call trie code disaster will result.

call_trie_retract_nr(Head,true) :-
	'_$savecp'(Breg),
	repeat,
	( trie_retract(0), % Zero out Last_Nod_Sav
	  call(Head) ->
	    ( trie_retract_nr -> true ; '_$cutto'(Breg),fail )
	; !, fail	
	).	

/*======================================================================*/
/* Supporting code for load_dyn and friends for "fast" reading in of    */
/* dynamic predicates.							*/
/*======================================================================*/

/* Must fail after code is put into buffer,
   to reset variables bound in the process */
code_to_buff(Clause) :- db_assert_code_to_buff(Clause), fail.
code_to_buff(_).

buff_to_clref(Head, AZ, Indices, HashTabSize) :-
	db_assert_buff_to_clref(Head, AZ, Indices, HashTabSize).
